package com.starbuds.server.service.finance.cloud.client;

import cn.hutool.core.convert.Convert;
import cn.hutool.core.util.RandomUtil;
import cn.hutool.http.ContentType;
import cn.hutool.http.HttpRequest;
import cn.hutool.http.HttpResponse;
import com.alibaba.fastjson.JSON;
import com.starbuds.server.service.finance.cloud.request.CloudAccountRequest;
import com.starbuds.server.service.finance.cloud.response.CloudAccountResponse;
import com.starbuds.server.service.finance.util.CloudAccountUtil;
import lombok.extern.slf4j.Slf4j;

import java.util.HashMap;
import java.util.Map;

/**
 * 云账户客户端
 */
@Slf4j
public abstract class AbstractCloudAccountClient implements CloudAccountClient {
    private String dealerId;
    private String brokerId;
    private String desKey;
    private String appKey;

    public AbstractCloudAccountClient(String dealerId, String brokerId, String desKey, String appKey) {
        this.dealerId = dealerId;
        this.brokerId = brokerId;
        this.desKey = desKey;
        this.appKey = appKey;
    }

    /**
     * 执行请求
     *
     * @param request
     * @return
     */
    @Override
    public <T extends CloudAccountResponse> T execute(CloudAccountRequest<T> request) {

        long timestamp = System.currentTimeMillis() / 1000;

        log.info("原始请求入参:\n{}=deskey={}", JSON.toJSONString(request, true),desKey);

        // data
        String data = CloudAccountUtil.getData(JSON.toJSONString(request), desKey);
        // 随机数
        String mess = RandomUtil.randomNumbers(32);

        // 请求头
        Map<String, String> headerMap = new HashMap<>(3);
        headerMap.put("dealer-id", dealerId);
        headerMap.put("request-id", Convert.toStr(System.currentTimeMillis()));

        // 表单
        Map<String, Object> formMap = new HashMap<>(7);
        formMap.put("data", data);
        formMap.put("mess", mess);
        formMap.put("timestamp", timestamp);
        formMap.put("sign", CloudAccountUtil.getSign(data, mess, timestamp, appKey));
        formMap.put("sign_type", "sha256");

        log.info("请求路径:{}", request.getApiUrl());
        log.info("请求头:{}", headerMap);
        log.info("请求参数:{}", formMap);

        // 构造基本请求头
        HttpResponse response = new HttpRequest(request.getApiUrl())
                .contentType(ContentType.FORM_URLENCODED.toString())
                .method(request.getMethod())
                .headerMap(headerMap, true)
                .form(formMap)
                .timeout(30 * 1000)
                .execute();

        if (response.isOk()) {
            log.info("请求结果:\n{}", JSON.toJSONString(response.body(), true));
            return JSON.parseObject(response.body(), request.getResponseClass());
        }
        log.info("请求出错，状态码:{}", response.getStatus());
        return null;
    }

    private static final char[] HEX_DIGITS = {'0', '1', '2', '3', '4', '5',
            '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};

    private static String getFormattedText(byte[] bytes) {
        int len = bytes.length;
        StringBuilder buf = new StringBuilder(len * 2);
        for (int j = 0; j < len; j++) {
            buf.append(HEX_DIGITS[(bytes[j] >> 4) & 0x0f]);
            buf.append(HEX_DIGITS[bytes[j] & 0x0f]);
        }
        return buf.toString();
    }
}
