package com.starbuds.server.service.finance.impl;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.starbuds.server.common.pojo.daoplus.QueryWrapper;
import com.starbuds.server.common.pojo.daoplus.UpdateWrapper;
import com.starbuds.server.common.pojo.finance.CaiyuCashoutNotify;
import com.starbuds.server.common.pojo.finance.CaiyuContractDO;
import com.starbuds.server.common.pojo.finance.bo.WithdrawRequestBO;
import com.starbuds.server.common.pojo.finance.bo.WithdrawRequestRecordBO;
import com.starbuds.server.common.service.finance.CaiyuService;
import com.starbuds.server.service.finance.provider.CaiyuContractProvider;
import com.starbuds.server.service.finance.util.CaiyuUtil;
import com.starbuds.server.service.finance.util.HttpUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.beans.factory.annotation.Value;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.HashMap;
import java.util.Map;

import static com.starbuds.server.service.finance.util.SpringContextUtil.isDev;

/**
 * 财鱼服务
 *
 * @author aoe
 * @date 2021/1/19
 */
@Slf4j
@Service(version = "1.0.0")
public class CaiyuServiceImpl implements CaiyuService {

    @Value("${caiyu.baseUrl}")
    private String baseUrl;

    /**
     * 商户号
     */
    @Value("${caiyu.partnerCode}")
    private String partnerCode;

    /**
     * MD5签名密钥
     */
    @Value("${caiyu.secretKey}")
    private String secretKey;

    /**
     * 免实名电签地址
     */
    private String URL_SIGN_CONTRACT_NO_BANK;

    /**
     * 申请提现地址
     */
    private String URL_APPLY_WITHDRAWAL;

    /**
     * 查询账户余额
     */
    private String URL_BALANCE;

    private BigDecimal HUNDRED = new BigDecimal("100");

    @Resource
    private CaiyuContractProvider caiyuContractProvider;

    @PostConstruct
    private void init() {
        URL_SIGN_CONTRACT_NO_BANK = baseUrl + "/noRealName/signContractNoBank";
        URL_APPLY_WITHDRAWAL = baseUrl + "/cashout/apply/api";
        URL_BALANCE = baseUrl + "/cashout/balance";
    }

    @Override
    public WithdrawRequestRecordBO doWithdraw(WithdrawRequestBO wr) {
        String timiOrderId = wr.getUserWalletWithdrawHisId() + "";
        String idCard = wr.getIdCard();
        String phone = wr.getPhone();
        // 元 换算成 分
        Long cash = wr.getAmount().multiply(HUNDRED).longValue();
        String bankCard = wr.getReceivingAccount();

        Map<String, String> paramsMap = new HashMap<>();
        // 需MD5签名参数
        // 商户号
        paramsMap.put("cashoutPartnerCode", partnerCode);
        // 订单号：提现方唯一订单号，最大32位
        paramsMap.put("partnerOrderId", "" + timiOrderId);
        // 身份证号
        paramsMap.put("idCard", idCard);
        // 手机号
        paramsMap.put("phone", phone);
        // 提现金额（整数）：单位分
        paramsMap.put("cash", "" + cash);
        // 时间戳
        paramsMap.put("timestamp", String.valueOf(System.currentTimeMillis()));

        WithdrawRequestRecordBO wrd = new WithdrawRequestRecordBO();
        wrd.setUrl(URL_APPLY_WITHDRAWAL);
        wrd.setIsSuccess(false);

        // MD5签名
        String sign;
        try {
            sign = CaiyuUtil.paramsMD5Sign(paramsMap, secretKey);
        } catch (Exception e) {
            log.error("财鱼 申请提现 签名失败 " + JSON.toJSONString(paramsMap), e);
            wrd.setErrMsg("财鱼 申请提现 签名失败");
            return wrd;
        }
        paramsMap.put("sign", sign);

        // 可选参数
        // 账户类型：1，银行卡；2，支付宝（账号传值时必填）
        paramsMap.put("accountType", "1");
        // 账号：账号和账户类型是成对出现的
        paramsMap.put("accountCard", bankCard);
        wrd.setRequest(JSON.toJSONString(paramsMap));

        // 发送POST请求
        long startTime = System.currentTimeMillis();
        String res = doPost(URL_APPLY_WITHDRAWAL, paramsMap, "申请提现");
        wrd.setSpendTime(startTime);
        wrd.setResponse(res);
        if (!isSuccess(res)) {
            wrd.setErrMsg(getErrMsg(res));
            return wrd;
        }

        wrd.setIsSuccess(true);
        return wrd;
    }

    /**
     * 查询余额 （单位：元）
     *
     * @return null:查询失败
     */
    @Override
    public BigDecimal getBalance() {
        Map<String, String> params = new HashMap<>();
        params.put("cashoutPartnerCode", partnerCode);
        long timestamp = System.currentTimeMillis();
        params.put("timestamp", "" + timestamp);
        String url = null;
        String res;
        try {
            url = URL_BALANCE + "?cashoutPartnerCode=" + partnerCode
                    + "&timestamp=" + timestamp
                    + "&sign=" + CaiyuUtil.paramsMD5Sign(params, secretKey);
            res = HttpUtil.get(url);
        } catch (Exception e) {
            log.error("查询财鱼银行卡账户余额失败 " + url, e);
            return null;
        }
        if (isDev()) {
            log.info("财鱼账户余额 \n{}", res);
        }
        if (!JSONObject.isValidObject(res)) {
            return null;
        }
        JSONObject json = JSONObject.parseObject(res);
        if (!"S000".equals(json.getString("status"))) {
            return null;
        }
        JSONArray data = json.getJSONArray("data");
        if (data == null) {
            return null;
        }
        for (int i = 0; i < data.size(); i++) {
            JSONObject d = data.getJSONObject(i);
            if (d == null) {
                return null;
            }
            // 账户类型：1，银行卡；2，支付宝
            Integer accountType = d.getInteger("accountType");
            if (accountType == null) {
                continue;
            }
            if (1 == accountType) {
                // 账户余额，单位分
                Long balance = d.getLong("balance");
                if (balance == null) {
                    return null;
                }
                return new BigDecimal("" + balance).multiply(HUNDRED);
            }
        }
        return null;
    }

    /**
     * 提现
     *
     * @param timiOrderId Timi订单号
     * @param idCard      提现用户身份证号
     * @param phone       提现用户手机号
     * @param cash        提现金额：单位分
     * @param bankCard    银行卡号
     * @return null:成功；错误信息
     */
    @Override
    public String cashout(String timiOrderId, String idCard, String phone, Long cash, String bankCard) {
        Map<String, String> paramsMap = new HashMap<>();
        // 需MD5签名参数
        // 商户号
        paramsMap.put("cashoutPartnerCode", partnerCode);
        // 订单号：提现方唯一订单号，最大32位
        paramsMap.put("partnerOrderId", "" + timiOrderId);
        // 身份证号
        paramsMap.put("idCard", idCard);
        // 手机号
        paramsMap.put("phone", phone);
        // 提现金额（整数）：单位分
        paramsMap.put("cash", "" + cash);
        // 时间戳
        paramsMap.put("timestamp", String.valueOf(System.currentTimeMillis()));

        // MD5签名
        String sign;
        try {
            sign = CaiyuUtil.paramsMD5Sign(paramsMap, secretKey);
        } catch (Exception e) {
            log.error("财鱼 申请提现 签名失败 " + JSON.toJSONString(paramsMap), e);
            return "财鱼 申请提现 签名失败";
        }
        paramsMap.put("sign", sign);

        // 可选参数
        // 账户类型：1，银行卡；2，支付宝（账号传值时必填）
        paramsMap.put("accountType", "1");
        // 账号：账号和账户类型是成对出现的
        paramsMap.put("accountCard", bankCard);

        // 发送POST请求
        String res = doPost(URL_APPLY_WITHDRAWAL, paramsMap, "申请提现");
        if (!isSuccess(res)) {
            return getErrMsg(res);
        }

        return null;
    }

    /**
     * 免实名电签
     *
     * @param name              用户姓名
     * @param phone             手机号
     * @param idCard            身份证号
     * @param idCardImgFrontUrl 身份证正面照
     * @param idCardImgBackUrl  身份证反面照
     * @return null:成功；错误信息
     */
    @Override
    public String signContractNoBank(Long userId, String name, String phone, String idCard, String idCardImgFrontUrl, String idCardImgBackUrl) {
        Map<String, String> paramsMap = new HashMap<>();
        // 需MD5签名参数
        // 商户号
        paramsMap.put("partnerCode", partnerCode);
        // 姓名
        paramsMap.put("name", name);
        // 手机号
        paramsMap.put("phone", phone);
        // 身份证号
        paramsMap.put("idCard", idCard);

        // MD5签名
        String sign;
        try {
            sign = CaiyuUtil.paramsMD5Sign(paramsMap, secretKey);
        } catch (Exception e) {
            log.error("财鱼 免实名电签 签名失败 " + JSON.toJSONString(paramsMap), e);
            return "财鱼 免实名电签 签名失败";
        }
        paramsMap.put("sign", sign);

        // 发送POST请求
        String paramJsonString = JSON.toJSONString(paramsMap);
        String res = doPost(URL_SIGN_CONTRACT_NO_BANK, paramJsonString, "免实名电签");
        if (!isSuccess(res)) {
            return getErrMsg(res);
        }

        String pdfFilePath = getValueByData(res, "pdfFilePath");
        if (StringUtils.isBlank(pdfFilePath)) {
            log.error(String.format("财鱼 免实名电签 失败（合同地址不存在） %s", paramJsonString));
            return "无法获得合同地址";
        }

        // 存储合同
        CaiyuContractDO po = new CaiyuContractDO();
        po.setUserId(userId);
        po.setContractUrl(pdfFilePath);
        po.setDetail(res);

        if (caiyuContractProvider.exist(userId)) {
            UpdateWrapper uw = new UpdateWrapper();

            QueryWrapper qw = new QueryWrapper();
            qw.eq("userId", userId);
            uw.setQw(qw);

            po.setUpdateTime(System.currentTimeMillis());
            uw.setPo(po);
            return caiyuContractProvider.update(uw) ? null : "更新签约信息失败";
        }

        return caiyuContractProvider.create(po) == null ? "存储财鱼用户签约合同失败" : null;
    }

    @Override
    public boolean checkSignCallback(CaiyuCashoutNotify notify) {
        if (notify == null
                || StringUtils.isBlank(notify.getCashoutPartnerCode())
                || StringUtils.isBlank(notify.getPartnerOrderId())
                || StringUtils.isBlank(notify.getOrderId())
                || StringUtils.isBlank(notify.getResult())
                || StringUtils.isBlank(notify.getSign())
        ) {
            return false;
        }
        Map<String, String> map = new HashMap<>();
        map.put("cashoutPartnerCode", notify.getCashoutPartnerCode());
        map.put("partnerOrderId", notify.getPartnerOrderId());
        map.put("orderId", notify.getOrderId());
        map.put("result", notify.getResult());

        try {
            String sign = CaiyuUtil.paramsMD5Sign(map, secretKey);
            return notify.getSign().equals(sign);
        } catch (Exception e) {
            return false;
        }
    }

    /**
     * 执行Post
     *
     * @param url
     * @param json
     * @param bizStr 业务描述
     * @return
     */
    private String doPost(String url, String json, String bizStr) {
        long startTime = System.currentTimeMillis();
        log.info("财鱼 {} 开始 {} {} {}", bizStr, startTime, url, json);
        String res;
        try {
            res = HttpUtil.postMaxTimout(url, json);
            log.info("财鱼 {} 返回结果 {}", bizStr, res);
        } catch (Exception e) {
            log.error(String.format("财鱼 %s %d 请求失败 耗时 %d", bizStr, startTime, System.currentTimeMillis() - startTime), e);
            return String.format("财鱼 %s 请求失败", bizStr);
        }
        log.info("财鱼 {} {} 结束 耗时 {}", bizStr, startTime, System.currentTimeMillis() - startTime);
        return res;
    }

    /**
     * 执行Post
     *
     * @param url
     * @param params
     * @param bizStr 业务描述
     * @return
     */
    private String doPost(String url, Map<String, String> params, String bizStr) {
        String json = JSON.toJSONString(params);
        long startTime = System.currentTimeMillis();
        log.info("财鱼 {} 开始 {} {} {}", bizStr, startTime, url, json);
        String res;
        try {
            res = HttpUtil.postMaxTimout(url, params);
            log.info("财鱼 {} 返回结果 {}", bizStr, res);
        } catch (Exception e) {
            log.error(String.format("财鱼 %s %d 请求失败 耗时 %d", bizStr, startTime, System.currentTimeMillis() - startTime), e);
            return String.format("财鱼 %s 请求失败", bizStr);
        }
        log.info("财鱼 {} {} 结束 耗时 {}", bizStr, startTime, System.currentTimeMillis() - startTime);
        return res;
    }

    /**
     * 请求结果是否成功
     *
     * @param res
     * @return true 操作成功
     */
    private boolean isSuccess(String res) {
        if (!JSON.isValidObject(res)) {
            return false;
        }
        JSONObject json = JSON.parseObject(res);
        if ("S000".equals(json.getString("status"))) {
            return true;
        }
        return false;
    }

    private String getErrMsg(String res) {
        if (!JSON.isValidObject(res)) {
            return res;
        }
        JSONObject json = JSON.parseObject(res);
        String msg = json.getString("msg");
        if (StringUtils.isBlank(msg)) {
            return "未知错误，操作失败";
        }
        return msg;
    }

    /**
     * 从返回的data中获取数据
     *
     * @param res
     * @param key
     * @return
     */
    private String getValueByData(String res, String key) {
        JSONObject json = JSON.parseObject(res);
        JSONObject data = json.getJSONObject("data");
        if (data == null) {
            return null;
        }
        return data.getString(key);
    }
}
