package com.starbuds.server.service.finance.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import cn.hutool.http.*;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.starbuds.server.common.constant.finance.UserVirtualCoinAccountHisStatusEnum;
import com.starbuds.server.common.pojo.api.PageList;
import com.starbuds.server.common.pojo.db.UserVirtualCoinAccountDO;
import com.starbuds.server.common.pojo.db.UserVirtualCoinAccountHisDO;
import com.starbuds.server.common.pojo.db.VirtualCoinDO;
import com.starbuds.server.common.pojo.finance.ChainNotify;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.finance.ChainService;
import com.starbuds.server.service.finance.dao.UserVirtualCoinAccountDAO;
import com.starbuds.server.service.finance.dao.UserVirtualCoinAccountHisDAO;
import com.starbuds.server.service.finance.dao.VirtualCoinDAO;
import com.starbuds.server.service.finance.provider.WalletProvider;
import com.starbuds.server.service.finance.util.ChainSignUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * 链支付实现
 *
 * @author Lay
 * @date 2020/3/13
 */
@Service(version = "1.0.0")
@Slf4j
public class ChainServiceImpl implements ChainService {

    @Reference(version = "1.0.0")
    IdentityService identityService;

    /**
     * 充值
     */
    private static final String TRADE_TYPE_RECHARGE = "1";

    /**
     * 提币
     */
    private static final String TRADE_TYPE_WITHDRAW = "2";

    /**
     * 交易成功
     */
    private static final String TRADE_STATUS_SUCCESS = "1";

    /**
     * 交易失败
     */
    private static final String TRADE_STATUS_FAIL = "2";

    @Value("${chain.appId}")
    private String chainAppId;

    @Value("${chain.appKey}")
    private String chainAppKey;

    @Value("${chain.apiUrl}")
    private String apiUrl;

    @Resource
    UserVirtualCoinAccountDAO userVirtualCoinAccountDAO;

    @Resource
    UserVirtualCoinAccountHisDAO userVirtualCoinAccountHisDAO;

    @Resource
    VirtualCoinDAO virtualCoinDAO;

    @Resource
    WalletProvider walletProvider;

    /**
     * 通知验签
     *
     * @param notifyJson
     * @return
     */
    @Override
    public boolean checkNotifySign(String notifyJson) {
        try {
            JSONObject jsonObject = JSON.parseObject(notifyJson);
            return ChainSignUtil.checkSign(jsonObject, chainAppKey, jsonObject.getString(ChainSignUtil.SIGN));
        } catch (Exception e) {
            log.error("链支付验签失败", e);
        }
        return false;
    }

    /**
     * 异步通知处理
     *
     * @param chainNotify
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public boolean handleNotify(ChainNotify chainNotify) {
        try {
            // 交易类型
            Long curTime = System.currentTimeMillis();
            if (TRADE_TYPE_RECHARGE.equals(chainNotify.getTransType())) {
                // 充值
                VirtualCoinDO virtualCoin = virtualCoinDAO.selectByContractAddress(chainNotify.getTokenAddress());
                if (virtualCoin == null) {
                    return true;
                }

                // 查询用户账户
                UserVirtualCoinAccountDO userVirtualCoinAccount = userVirtualCoinAccountDAO.selectByAccountAddress(chainNotify.getToAddress());
                if (userVirtualCoinAccount == null) {
                    return true;
                }

                Long amount = new BigDecimal(chainNotify.getAmount()).multiply(BigDecimal.valueOf(virtualCoin.getCoinMultiple())).setScale(0, RoundingMode.DOWN).longValue();
                Long hisId = identityService.getId();

                // 加币
                walletProvider.addBalance(virtualCoin.getCoinId(), userVirtualCoinAccount.getUserId(), amount, "用户充值", hisId);

                // 创建记录
                UserVirtualCoinAccountHisDO userVirtualCoinAccountHisDO = new UserVirtualCoinAccountHisDO();
                userVirtualCoinAccountHisDO.setHisId(hisId);
                userVirtualCoinAccountHisDO.setUserId(userVirtualCoinAccount.getUserId());
                userVirtualCoinAccountHisDO.setCoinId(virtualCoin.getCoinId());
                userVirtualCoinAccountHisDO.setFromAddress(chainNotify.getFromAddress());
                userVirtualCoinAccountHisDO.setToAddress(chainNotify.getToAddress());
                userVirtualCoinAccountHisDO.setTokenAddress(chainNotify.getTokenAddress());
                userVirtualCoinAccountHisDO.setHisAmount(amount);
                userVirtualCoinAccountHisDO.setTransType(Integer.valueOf(TRADE_TYPE_RECHARGE));
                userVirtualCoinAccountHisDO.setHisStatus(UserVirtualCoinAccountHisStatusEnum.SUCCESS.getValue());
                userVirtualCoinAccountHisDO.setOrderId(String.valueOf(hisId));
                userVirtualCoinAccountHisDO.setTxId(chainNotify.getTxid());
                userVirtualCoinAccountHisDO.setHisCreateTime(curTime);
                userVirtualCoinAccountHisDO.setHisNotifyTime(curTime);
                userVirtualCoinAccountHisDAO.insert(userVirtualCoinAccountHisDO);

            } else if (TRADE_TYPE_WITHDRAW.equals(chainNotify.getTransType())) {
                // 提币

                // 查询订单记录
                UserVirtualCoinAccountHisDO accountHis = userVirtualCoinAccountHisDAO.selectById(Convert.toLong(chainNotify.getOrderNo()));
                if (accountHis == null || UserVirtualCoinAccountHisStatusEnum.GOING != UserVirtualCoinAccountHisStatusEnum.toEnum(accountHis.getHisStatus())) {
                    return true;
                }

                // 交易失败
                if (TRADE_STATUS_FAIL.equals(chainNotify.getStatus())) {
                    // 返还币
                    walletProvider.addBalance(accountHis.getCoinId(), accountHis.getUserId(), accountHis.getHisAmount(), "提币失败返还", accountHis.getHisId());
                    accountHis.setHisStatus(UserVirtualCoinAccountHisStatusEnum.FAIL.getValue());
                } else if (TRADE_STATUS_SUCCESS.equals(chainNotify.getStatus())) {
                    accountHis.setHisStatus(UserVirtualCoinAccountHisStatusEnum.SUCCESS.getValue());
                }

                // 更新记录状态
                accountHis.setHisNotifyTime(curTime);
                userVirtualCoinAccountHisDAO.update(accountHis);
            }
        } catch (Exception e) {
            log.error("链支付异步通知处理失败", e);
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            return false;
        }
        return true;
    }

    /**
     * 用户用户虚拟币账户
     *
     * @param userId
     * @return
     */
    @Override
    public List<UserVirtualCoinAccountDO> getAccounts(Long userId) {
        return userVirtualCoinAccountDAO.selectByUserId(userId);
    }

    /**
     * 根据主网查询
     *
     * @param userId
     * @param coinNet
     * @return
     */
    @Override
    public UserVirtualCoinAccountDO getAccountByCoinNet(Long userId, String coinNet) {
        return userVirtualCoinAccountDAO.selectById(userId, coinNet);
    }

    /**
     * 获取虚拟币列表
     *
     * @return
     */
    @Override
    public List<VirtualCoinDO> getVirtualCoinList() {
        return virtualCoinDAO.selectAll();
    }

    /**
     * 获取账户记录列表
     *
     * @param pageIndex
     * @param pageSize
     * @param userId
     * @param coinId
     * @return
     */
    @Override
    public PageList<UserVirtualCoinAccountHisDO> getAccountHisList(int pageIndex, int pageSize, Long userId, Integer coinId, Integer transType) {
        PageList<UserVirtualCoinAccountHisDO> pageList = new PageList<>(pageIndex, pageSize);
        pageList.setList(userVirtualCoinAccountHisDAO.selectAll(pageList.getPageCursor(), pageSize, userId, coinId, transType));
        pageList.setTotal(userVirtualCoinAccountHisDAO.count(userId, coinId, transType));
        return pageList;
    }

    /**
     * 创建账户
     *
     * @param userId
     * @param coinNet
     * @return
     */
    @Override
    public String createAccount(Long userId, String coinNet) {
        try {
            // 调用远程创建账户
            Map<String, Object> params = new HashMap<>();
            params.put("member", userId);
            params.put("chaincode", coinNet);
            params.put("note", "");
            params.put("token_address", "");
            String sign = ChainSignUtil.md5Sign(params, chainAppKey);

            JSONObject jsonObject = new JSONObject();
            jsonObject.put("version", "1.0");
            jsonObject.put("method", "api.createAccount");
            jsonObject.put("appid", chainAppId);
            jsonObject.put("params", params);
            jsonObject.put("timestamp", System.currentTimeMillis());
            jsonObject.put("sign", sign);
            log.info(jsonObject.toJSONString());
            HttpRequest httpRequest = HttpUtil.createRequest(Method.POST, apiUrl);
            httpRequest.body(jsonObject.toJSONString(), ContentType.JSON.toString());
            HttpResponse response = httpRequest.execute();
            log.info(response.body());
            if (response.isOk()) {
                JSONObject jsonRes = JSON.parseObject(response.body());
                if (jsonRes != null && jsonRes.getBoolean("success")) {
                    String accountAddress = jsonRes.getString("msg");
                    // 创建记录
                    UserVirtualCoinAccountDO userVirtualCoinAccount = new UserVirtualCoinAccountDO();
                    userVirtualCoinAccount.setUserId(userId);
                    userVirtualCoinAccount.setCoinNet(coinNet);
                    userVirtualCoinAccount.setAccountAddress(accountAddress);
                    userVirtualCoinAccount.setAccountCreateTime(System.currentTimeMillis());
                    userVirtualCoinAccountDAO.insert(userVirtualCoinAccount);
                    return accountAddress;
                }
            }
        } catch (Exception e) {
            log.error("账户创建失败", e);
        }
        return null;
    }

    /**
     * 创建提币申请
     *
     * @param userId
     * @param coinId
     * @param amount
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public boolean applyWithdraw(Long userId, Integer coinId, Long amount, String toAddress) {
        try {
            Long hisId = identityService.getId();

            // 查询账户
            VirtualCoinDO virtualCoin = virtualCoinDAO.selectById(coinId);
            if (virtualCoin == null) {
                return false;
            }

            UserVirtualCoinAccountDO userVirtualCoinAccount = userVirtualCoinAccountDAO.selectById(userId, virtualCoin.getCoinNet());
            if (userVirtualCoinAccount == null) {
                return false;
            }

            // 扣币
            Long balance = walletProvider.reduceBalance(coinId, userId, amount, "用户提币", hisId);
            if (balance < 0) {
                return false;
            }

            // 写入记录
            Long curTime = System.currentTimeMillis();
            UserVirtualCoinAccountHisDO userVirtualCoinAccountHisDO = new UserVirtualCoinAccountHisDO();
            userVirtualCoinAccountHisDO.setHisId(hisId);
            userVirtualCoinAccountHisDO.setUserId(userId);
            userVirtualCoinAccountHisDO.setCoinId(coinId);
            userVirtualCoinAccountHisDO.setFromAddress(userVirtualCoinAccount.getAccountAddress());
            userVirtualCoinAccountHisDO.setToAddress(toAddress);
            userVirtualCoinAccountHisDO.setTokenAddress(virtualCoin.getContractAddress());
            userVirtualCoinAccountHisDO.setHisAmount(amount);
            userVirtualCoinAccountHisDO.setTransType(Integer.valueOf(TRADE_TYPE_WITHDRAW));
            userVirtualCoinAccountHisDO.setHisStatus(UserVirtualCoinAccountHisStatusEnum.GOING.getValue());
            userVirtualCoinAccountHisDO.setOrderId(String.valueOf(hisId));
            userVirtualCoinAccountHisDO.setHisCreateTime(curTime);
            userVirtualCoinAccountHisDAO.insert(userVirtualCoinAccountHisDO);

            // 请求提币
            String realAmount = BigDecimal.valueOf(amount).divide(BigDecimal.valueOf(virtualCoin.getCoinMultiple())).toPlainString();
            Map<String, Object> params = new HashMap<>();
            params.put("member", userId);
            params.put("chaincode", virtualCoin.getCoinNet());
            params.put("from_address", userVirtualCoinAccount.getAccountAddress());
            params.put("to_address", toAddress);
            params.put("token_address", virtualCoin.getContractAddress());
            params.put("amount", realAmount);
            params.put("orderno", hisId);
            params.put("note", "");
            String sign = ChainSignUtil.md5Sign(params, chainAppKey);

            JSONObject jsonObject = new JSONObject();
            jsonObject.put("version", "1.0");
            jsonObject.put("method", "api.sendTransaction");
            jsonObject.put("appid", chainAppId);
            jsonObject.put("params", params);
            jsonObject.put("timestamp", System.currentTimeMillis());
            jsonObject.put("sign", sign);

            String body = jsonObject.toJSONString();
            log.info(body);
            HttpRequest httpRequest = HttpUtil.createRequest(Method.POST, apiUrl);
            httpRequest.body(body, ContentType.JSON.toString());
            HttpResponse response = httpRequest.execute();
            boolean success = false;
            if (response.isOk()) {
                JSONObject jsonRes = JSON.parseObject(response.body());
                if (jsonRes != null) {
                    if (jsonRes.getBoolean("success")) {
                        success = true;
                    } else {
                        throw new RuntimeException(jsonRes.getString("msg"));
                    }
                }
            }

            if (!success) {
                throw new RuntimeException("链支付处理交易失败");
            }

            return true;
        } catch (Exception e) {
            log.error("提币申请创建失败", e);
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
        }
        return false;
    }

    /**
     * 判断币种是否存在
     *
     * @param coinId
     * @return
     */
    @Override
    public boolean isVirtualCoinExists(Integer coinId) {
        return virtualCoinDAO.countById(coinId) > 0;
    }

    /**
     * 判断主网是否存在
     *
     * @param coinNet
     * @return
     */
    @Override
    public boolean isCoinNetExists(String coinNet) {
        return virtualCoinDAO.countByCoinNet(coinNet) > 0;
    }

    /**
     * 获取虚拟币 id 名称哈希表
     *
     * @return
     */
    @Override
    public Map<Integer, String> getCoinIdNameMap() {
        List<VirtualCoinDO> list = getVirtualCoinList();
        if (CollectionUtil.isNotEmpty(list)) {
            return list.stream().collect(Collectors.toMap(VirtualCoinDO::getCoinId, VirtualCoinDO::getCoinName));
        }
        return null;
    }
}
