package com.starbuds.server.service.finance.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.date.DateField;
import cn.hutool.core.date.DateTime;
import cn.hutool.core.date.DateUtil;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.pojo.api.RecordList;
import com.starbuds.server.common.pojo.finance.ChannelExchangeOrder;
import com.starbuds.server.common.pojo.finance.ChannelWalletHisVO;
import com.starbuds.server.common.pojo.finance.ChannelWalletVO;
import com.starbuds.server.common.pojo.util.TimeSpan;
import com.starbuds.server.common.service.finance.ChannelWalletService;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.service.finance.provider.ChannelWalletProvider;
import com.starbuds.server.service.finance.provider.WalletProvider;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.List;
import java.util.Optional;

/**
 * 渠道钱包服务实现
 *
 * @author Lay
 * @date 2020/8/18
 */
@Service(version = "1.0.0")
@Slf4j
public class ChannelWalletServiceImpl implements ChannelWalletService {

    @Reference(version = "1.0.0")
    StatPlatformService statPlatformService;

    @Resource
    ChannelWalletProvider channelWalletProvider;

    @Resource
    WalletProvider walletProvider;

    /**
     * 星钻红包
     *
     * @param channelId
     * @param hisId
     * @param exchangeOrders
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public Long exchangeDiamond(Long channelId, Long hisId, List<ChannelExchangeOrder> exchangeOrders) {
        Long totalAmount = exchangeOrders.stream().mapToLong(ChannelExchangeOrder::getAmount).sum();

        Long balance = null;
        try {
            // 扣钱
            balance = channelWalletProvider.reduceBalance(WalletTypeEnum.STAR_DIAMOND, channelId, totalAmount, "发放星钻红包", hisId);
            if (balance != null) {
                if (balance < 0L) {
                    return balance;
                }

                // 批量加钱
                for (ChannelExchangeOrder exchangeOrder : exchangeOrders) {
                    walletProvider.addBalance(WalletTypeEnum.STAR_DIAMOND, exchangeOrder.getUserId(), exchangeOrder.getAmount(), "星钻红包", hisId);
                }

            }

            // 统计
            statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.STAR_DIAMOND_FROM_REDPACKET.getValue(), BigDecimal.valueOf(totalAmount));

            return balance;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("星钻红包发放失败", e);
        }

        return balance;
    }

    /**
     * 查询渠道钱包
     *
     * @param channelId
     * @return
     */
    @Override
    public ChannelWalletVO getChannelWallet(Long channelId) {
        return channelWalletProvider.getChannelWallet(channelId);
    }

    /**
     * 查询钱包余额
     *
     * @param channelId
     * @param walletType
     * @return
     */
    @Override
    public Long getBalance(Long channelId, WalletTypeEnum walletType) {
        return channelWalletProvider.getBalance(channelId, walletType);
    }

    /**
     * 查询钱包余额
     *
     * @param channelId
     * @param walletType
     * @return
     */
    @Override
    public Long getBalance(Long channelId, Integer walletType) {
        return channelWalletProvider.getBalance(channelId, walletType);
    }

    /**
     * 创建渠道钱包
     *
     * @param channelId
     * @return
     */
    @Override
    public boolean createChannelWallet(Long channelId) {
        return channelWalletProvider.createChannelWallet(channelId);
    }

    /**
     * 获取渠道钱包明细
     *
     * @param lastId
     * @param pageSize
     * @param channelId
     * @param walletType
     * @param yearMonth
     * @param isIncome
     * @return
     */
    @Override
    public RecordList<ChannelWalletHisVO, Long> getChannelWalletHisListByTypeAndMonth(Long lastId, int pageSize, Long channelId, WalletTypeEnum walletType, Integer yearMonth, Integer isIncome) {
        return getChannelWalletHisListByTypeAndMonth(lastId, pageSize, channelId, walletType.getValue(), yearMonth, isIncome);
    }

    /**
     * 获取渠道钱包明细
     *
     * @param lastId
     * @param pageSize
     * @param channelId
     * @param walletType
     * @param yearMonth
     * @param isIncome
     * @return
     */
    @Override
    public RecordList<ChannelWalletHisVO, Long> getChannelWalletHisListByTypeAndMonth(Long lastId, int pageSize, Long channelId, Integer walletType, Integer yearMonth, Integer isIncome) {
        RecordList<ChannelWalletHisVO, Long> recordList = new RecordList<>(pageSize);

        // 处理时间
        TimeSpan ts = getTimeSpanByYearMonth(yearMonth);
        List<ChannelWalletHisVO> walletHisVOList = channelWalletProvider.getChannelWalletHisList(channelId, lastId, pageSize, walletType, isIncome, ts.getBeginTime(), ts.getEndTime());

        if (CollectionUtil.isNotEmpty(walletHisVOList)) {
            lastId = walletHisVOList.get(walletHisVOList.size() - 1).getHisId();
        }

        recordList.setList(walletHisVOList);
        recordList.setLastId(lastId);
        return recordList;
    }

    /**
     * 获取渠道钱包明细
     *
     * @param lastId
     * @param pageSize
     * @param channelId
     * @param walletType
     * @param beginTime
     * @param endTime
     * @param isIncome
     * @return
     */
    @Override
    public RecordList<ChannelWalletHisVO, Long> getChannelWalletHisListByTypeAndTime(Long lastId, int pageSize, Long channelId, Integer walletType, Long beginTime, Long endTime, Integer isIncome) {
        RecordList<ChannelWalletHisVO, Long> recordList = new RecordList<>(pageSize);

        // 处理时间
        List<ChannelWalletHisVO> walletHisVOList = channelWalletProvider.getChannelWalletHisList(channelId, lastId, pageSize, walletType, isIncome, beginTime, endTime);

        if (CollectionUtil.isNotEmpty(walletHisVOList)) {
            lastId = walletHisVOList.get(walletHisVOList.size() - 1).getHisId();
        }

        recordList.setList(walletHisVOList);
        recordList.setLastId(lastId);
        return recordList;
    }

    /**
     * 获取渠道钱包求和值
     *
     * @param channelId
     * @param walletType
     * @param yearMonth
     * @param isIncome
     * @return
     */
    @Override
    public long getChannelWalletHisSumByTypeAndMonth(Long channelId, WalletTypeEnum walletType, Integer yearMonth, Integer isIncome) {
        return getChannelWalletHisSumByTypeAndMonth(channelId, walletType.getValue(), yearMonth, isIncome);
    }

    /**
     * 获取渠道钱包求和值
     *
     * @param channelId
     * @param walletType
     * @param yearMonth
     * @param isIncome
     * @return
     */
    @Override
    public long getChannelWalletHisSumByTypeAndMonth(Long channelId, Integer walletType, Integer yearMonth, Integer isIncome) {
        TimeSpan ts = getTimeSpanByYearMonth(yearMonth);
        return Optional.ofNullable(channelWalletProvider.sumByTypeAndTime(channelId, walletType, isIncome, ts.getBeginTime(), ts.getEndTime())).orElse(0L);
    }

    /**
     * 获取渠道钱包求和值
     *
     * @param channelId
     * @param walletType
     * @param beginTime
     * @param endTime
     * @param isIncome
     * @return
     */
    @Override
    public long getChannelWalletHisSumByTypeAndTime(Long channelId, Integer walletType, Long beginTime, Long endTime, Integer isIncome) {
        return Optional.ofNullable(channelWalletProvider.sumByTypeAndTime(channelId, walletType, isIncome, beginTime, endTime)).orElse(0L);
    }

    private static TimeSpan getTimeSpanByYearMonth(Integer yearMonth) {
        TimeSpan ts = new TimeSpan();
        if (yearMonth != null) {
            DateTime dateTime = DateUtil.parse(String.valueOf(yearMonth), "yyyyMM");
            dateTime.setField(DateField.DAY_OF_MONTH, 1);
            dateTime.setField(DateField.HOUR_OF_DAY, 0);
            dateTime.setField(DateField.MINUTE, 0);
            dateTime.setField(DateField.SECOND, 0);
            dateTime.setField(DateField.MILLISECOND, 0);
            ts.setBeginTime(dateTime.getTime());
            dateTime.offset(DateField.MONTH, 1);
            dateTime.offset(DateField.MILLISECOND, -1);
            ts.setEndTime(dateTime.getTime());
        }
        return ts;
    }
}
