package com.starbuds.server.service.finance.impl;

import cn.hutool.core.convert.Convert;
import cn.hutool.crypto.SecureUtil;
import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.cache.RedisCacheKeys;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.common.CommonStatus;
import com.starbuds.server.common.constant.finance.PayStatusEnum;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.pojo.db.CloudAccountTransferHisDO;
import com.starbuds.server.common.pojo.finance.*;
import com.starbuds.server.common.pojo.mq.QueueTransferUserMsg;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.finance.CloudAccountService;
import com.starbuds.server.service.finance.cloud.provider.CloudAccountProvider;
import com.starbuds.server.service.finance.cloud.response.*;
import com.starbuds.server.service.finance.dao.CloudAccountTransferHisDAO;
import com.starbuds.server.service.finance.dao.UserBankcardDAO;
import com.starbuds.server.service.finance.rocket.RocketProducer;
import com.starbuds.server.service.finance.util.SpringContextUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.concurrent.TimeUnit;

import static java.math.BigDecimal.ROUND_HALF_UP;

/**
 * @Author jiangxiaowei
 * @DateTime 2020/7/24 18:53
 * @Description 云账户
 */
@Slf4j
@Service(version = "1.0.0")
public class CloudAccountServiceImpl implements CloudAccountService {

    @Resource
    CloudAccountProvider cloudAccountProvider;

    @Reference(version = "1.0.0")
    RedisService redisService;

    @Resource
    UserBankcardDAO userBankcardDAO;

    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Resource
    CloudAccountTransferHisDAO cloudAccountTransferHisDAO;

    @Resource
    RocketProducer rocketProducer;

    @Override
    public boolean verifyId(String idCard, String realName) {
        VerifyIdResponse response = cloudAccountProvider.verifyId(idCard, realName);
        if (response == null) {
            return false;
        }
        return response.isSuccess();
    }

    @Override
    public VerifyRequestDTO verifyCard(String cardNo, String idCard, String realName, String mobile) {
        // 发送鉴权请求
        VerifyResponse response = cloudAccountProvider.verifyRequest(cardNo, idCard, realName, mobile);
        // 结果包装
        VerifyRequestDTO requestDTO = new VerifyRequestDTO();
        if (response == null) {
            requestDTO.setSuccess(false);
            requestDTO.setMessage("服务无响应");
            return requestDTO;
        }
        if (response.isSuccess()) {
            requestDTO.setSuccess(true);
            String data = response.getData();
            VerifyResponse.Data responseData = JSON.parseObject(data, VerifyResponse.Data.class);
            requestDTO.setRef(responseData.getRef());
        } else {
            requestDTO.setSuccess(false);
            requestDTO.setMessage(response.getMessage());
        }
        return requestDTO;
    }

    @Override
    public VerifyConfirmRequestDTO verifyConfirm(Long userId, String cardNo, String idCard, String realName, String mobile, String captcha, String ref) {
        VerifyConfirmResponse response = cloudAccountProvider.verifyConfirm(cardNo, idCard, realName, mobile, captcha, ref);
        VerifyConfirmRequestDTO requestDTO = new VerifyConfirmRequestDTO();
        if (response == null) {
            requestDTO.setSuccess(false);
            requestDTO.setMessage("服务无响应");
        }
        if (response.isSuccess()) {
            String redisValue = redisService.get(RedisCacheKeys.getBankcardKey(cardNo));
            if (StringUtils.isNotEmpty(redisValue)) {
                CardInfo cardInfo = JSON.parseObject(redisValue, CardInfo.class);
                // 绑卡
                if (cardInfo != null) {
                    UserBankcardDO bankcard = new UserBankcardDO();
                    bankcard.setCardId(getCardId(userId, cardNo));
                    bankcard.setUserId(userId);
                    bankcard.setCardNo(cardNo);
                    bankcard.setBankCode(cardInfo.getBankCode());
                    bankcard.setBankName(cardInfo.getBankName());
                    bankcard.setCardType(cardInfo.getCardType());
                    bankcard.setMobile(mobile);
                    bankcard.setIsSupportCloudAccount(cardInfo.getIsSupport() != null && cardInfo.getIsSupport() ? BoolIntVal.True : BoolIntVal.False);
                    bankcard.setUpdateTime(System.currentTimeMillis());
                    bankcard.setStatus(CommonStatus.Normal);
                    userBankcardDAO.saveOrUpdate(bankcard);
                }
            }
            requestDTO.setSuccess(true);

        } else {
            requestDTO.setSuccess(false);
            requestDTO.setMessage(response.getMessage());
        }
        return requestDTO;
    }

    @Override
    public CardInfo queryCard(String cardNo, String bankName) {
        CardQueryResponse response = cloudAccountProvider.queryCard(cardNo, bankName);
        if (response == null) {
            return null;
        }
        if (response.isSuccess()) {
            CardInfo cardInfo = new CardInfo();
            cardInfo.setBankCode(response.getData().getBankCode());
            cardInfo.setBankName(response.getData().getBankName());
            cardInfo.setCardType(response.getData().getCardType());
            cardInfo.setIsSupport(response.getData().getIsSupport());
            redisService.set(RedisCacheKeys.getBankcardKey(cardNo), JSON.toJSONString(cardInfo), 24, TimeUnit.HOURS);
            return cardInfo;
        }
        return null;
    }

    @Override
    public PaymentOrderDTO payToUser(Long userId, String realName, String cardNo, String idCard, String amount, String payRemark, String bizId) {

        String orderAmount = amount;
        // 测试环境
        if (SpringContextUtil.isDev()) {
            amount = "0.01";
        }
        Long orderId = identityService.getId();
        PaymentOrderResponse response = cloudAccountProvider.payOrder(orderId, realName, cardNo, idCard, amount, payRemark);
        PaymentOrderDTO orderDTO = new PaymentOrderDTO();
        if (response.isSuccess()) {
            long curTime = System.currentTimeMillis();

            CloudAccountTransferHisDO hisDO = new CloudAccountTransferHisDO();
            hisDO.setOrderId(Convert.toStr(orderId));
            hisDO.setUserId(userId);
            hisDO.setCardNo(cardNo);
            hisDO.setIdCard(idCard);
            hisDO.setPayAmount(orderAmount);
            hisDO.setPayRemark(payRemark);
            hisDO.setPayStatus(PayStatusEnum.WAITTING.getValue());
            hisDO.setCreateTime(curTime);
            hisDO.setUpdateTime(curTime);
            hisDO.setBizId(bizId);
            hisDO.setRequestId(response.getRequestId());
            hisDO.setOutId(response.getData().getRef());
            cloudAccountTransferHisDAO.insert(hisDO);

            orderDTO.setSuccess(true);
            orderDTO.setOrderId(response.getData().getOrderId());
            orderDTO.setPay(response.getData().getPay());
            orderDTO.setRef(response.getData().getRef());
        } else {
            orderDTO.setSuccess(false);
            orderDTO.setMessage(response.getMessage());
        }
        return orderDTO;
    }

    @Override
    public boolean handleOrderNotify(CloudAccountOrderNotify notify) {
        log.info("notify:{}", JSON.toJSONString(notify));
        String orderId = notify.getData().getOrderId();
        String status = notify.getData().getStatus();
        long curTime = System.currentTimeMillis();
        CloudAccountTransferHisDO transferHisDO = cloudAccountTransferHisDAO.selectByOrderId(orderId);
        log.info("transferHis:{}", transferHisDO);
        if (transferHisDO == null || PayStatusEnum.WAITTING.getValue() != transferHisDO.getPayStatus()) {
            log.info("订单已处理");
            return true;
        }
        // 成功
        if ("1".equals(status)) {
            // 生产环境校验金额
            if (!SpringContextUtil.isDev()) {
                BigDecimal pay = new BigDecimal(transferHisDO.getPayAmount()).divide(BigDecimal.ONE, 2, ROUND_HALF_UP);
                BigDecimal broker = new BigDecimal(notify.getData().getBrokerAmount()).divide(BigDecimal.ONE, 2, ROUND_HALF_UP);
                if (pay.compareTo(broker) != 0) {
                    log.info("订单金额错误 {} {} compare:{}", notify.getData().getBrokerAmount(), transferHisDO.getPayAmount());
                    return false;
                }
            }
            cloudAccountTransferHisDAO.updateResult(orderId, PayStatusEnum.SUCCESSED.getValue(), JSON.toJSONString(notify), curTime);

            QueueTransferUserMsg transferMsg = new QueueTransferUserMsg();
            transferMsg.setUserId(transferHisDO.getUserId());
            transferMsg.setAmount(transferHisDO.getPayAmount());
            transferMsg.setSuccess(true);
            transferMsg.setRef("星尘转出");
            rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.TRANSFER_TO_USER, transferMsg.serialize());

            log.info("回调处理成功");
            return true;
        } else {
            cloudAccountTransferHisDAO.updateResult(orderId, PayStatusEnum.FAILED.getValue(), JSON.toJSONString(notify), curTime);

            log.info("回调处理失败");
            return false;
        }
    }

    @Override
    public String decodeData(String encryptStr) {
        return cloudAccountProvider.decodeData(encryptStr);
    }

    @Override
    public String getSign(String encryptStr, String mess, Long timestamp) {
        return cloudAccountProvider.getSign(encryptStr, mess, timestamp);
    }

    /**
     * 获取卡号id
     *
     * @param userId
     * @param cardNo
     * @return
     */
    public static String getCardId(Long userId, String cardNo) {
        return SecureUtil.md5(String.format("%d:%s", userId, cardNo));
    }
}
