package com.starbuds.server.service.finance.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import com.starbuds.server.common.constant.finance.BagItemTypeEnum;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.message.im.ContentType;
import com.starbuds.server.common.constant.message.im.ImActionType;
import com.starbuds.server.common.constant.message.push.ModuleTypeEnum;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.pojo.api.PageList;
import com.starbuds.server.common.pojo.data.CoinGoodsListItemVO;
import com.starbuds.server.common.pojo.db.CoinGoodsOrderDO;
import com.starbuds.server.common.pojo.finance.CoinGoodsOrderVO;
import com.starbuds.server.common.pojo.message.im.biz.SystemMsg;
import com.starbuds.server.common.pojo.user.UserBasicInfo;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.data.CoinGoodsService;
import com.starbuds.server.common.service.finance.CoinGoodsOrderService;
import com.starbuds.server.common.service.message.ImService;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.finance.dao.CoinGoodsOrderDAO;
import com.starbuds.server.service.finance.provider.BagProvider;
import com.starbuds.server.service.finance.provider.WalletProvider;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.List;
import java.util.Optional;

import static com.starbuds.server.common.constant.finance.WalletTypeEnum.GOLD_COIN;

/**
 * 金币订单服务实现
 *
 * @author Lay
 * @date 2020/3/30
 */
@Slf4j
@Service(version = "1.0.0")
public class CoinGoodsOrderServiceImpl implements CoinGoodsOrderService {

    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Reference(version = "1.0.0")
    CoinGoodsService coinGoodsService;

    @Reference(version = "1.0.0")
    UserService userService;

    @Reference(version = "1.0.0")
    ImService imService;

    @Reference(version = "1.0.0")
    private StatPlatformService statPlatformService;

    @Resource
    WalletProvider walletProvider;

    @Resource
    CoinGoodsOrderDAO coinGoodsOrderDAO;

    @Resource
    BagProvider bagProvider;

    /**
     * 支付金币商品
     *
     * @param userId
     * @param goodsId
     * @param quantity
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public Long addCoinGoodsOrder(Long userId, Long goodsId, Integer quantity, Long targetUserId) {
        CoinGoodsListItemVO coinGoods = coinGoodsService.getCoinGoodsListItem(goodsId);
        if (coinGoods == null) {
            return null;
        }

        Long amount = coinGoods.getGoodsPrice() * quantity;
        Long orderId = identityService.getId();
        Long curTime = System.currentTimeMillis();

        // 扣钱
        Long balance = walletProvider.reduceBalance(coinGoods.getGoodsPriceType(), userId, amount, String.format("兑换%s", coinGoods.getGoodsName()), orderId);
        if (balance == null || balance < 0) {
            return balance;
        }

        // 增加物品
        bagProvider.addQuantity(targetUserId, coinGoods.getItemId(), BagItemTypeEnum.PROPS.getValue(), quantity, -1L, String.format("%d%s兑换", amount, coinGoods.getGoodsPriceTypeName()), orderId);

        // 创建订单
        CoinGoodsOrderDO coinGoodsOrder = new CoinGoodsOrderDO();
        coinGoodsOrder.setOrderId(orderId);
        coinGoodsOrder.setGoodsId(goodsId);
        coinGoodsOrder.setUserId(userId);
        coinGoodsOrder.setGoodsPrice(coinGoods.getGoodsPrice());
        coinGoodsOrder.setGoodsPriceType(coinGoods.getGoodsPriceType());
        coinGoodsOrder.setOrderAmount(amount);
        coinGoodsOrder.setOrderQuantity(quantity);
        coinGoodsOrder.setOrderCreateTime(curTime);
        coinGoodsOrder.setGoodsName(coinGoods.getGoodsName());
        coinGoodsOrder.setGoodsPic(coinGoods.getGoodsPic());
        coinGoodsOrderDAO.insert(coinGoodsOrder);

        // 赠送推送
        if (!userId.equals(targetUserId)) {
            UserBasicInfo basicInfo = userService.getBasicInfo(userId);
            // 推送
            String pushBody = String.format("%s(id:%d)通过商城向你赠送了【%s】，请前往背包查收！",
                    Optional.ofNullable(basicInfo).map(UserBasicInfo::getUserName).orElse(""),
                    Optional.ofNullable(basicInfo).map(UserBasicInfo::getUserNo).orElse(null),
                    coinGoods.getGoodsName());
            SystemMsg systemMsg = new SystemMsg();
            systemMsg.setContentType(ContentType.TEXT);
            systemMsg.setTitle("赠送通知");
            systemMsg.setBody(pushBody);
            systemMsg.setActionType(ImActionType.JUMP_NATIVE.getValue());
            systemMsg.setModuleType(ModuleTypeEnum.BAG.getDesc());
            systemMsg.setModuleId(Convert.toStr(targetUserId));
            imService.sendCustomImMsg(null, targetUserId, systemMsg, null);
        }

        // 统计 消耗金币-金币商城
        if (GOLD_COIN == WalletTypeEnum.toEnum(coinGoods.getGoodsPriceType())) {
            statPlatformService.addPlatformStat(Optional.ofNullable(userService.getChannelId(userId)).orElse(0L),
                    StatPlatformTypeEnum.GOLD_COIN_REDUCE_COIN_SHOP, new BigDecimal("" + Math.abs(amount)));
        }

        return balance;
    }

    /**
     * 获取金币订单列表
     *
     * @param pageIndex
     * @param pageCursor
     * @param userId
     * @return
     */
    @Override
    public PageList<CoinGoodsOrderVO> getCoinGoodsOrderList(int pageIndex, int pageCursor, Long userId) {
        PageList<CoinGoodsOrderVO> pageList = new PageList<>(pageIndex, pageCursor);
        pageList.setTotal(coinGoodsOrderDAO.count(userId));
        List<CoinGoodsOrderVO> list = coinGoodsOrderDAO.selectAll(pageList.getPageCursor(), pageCursor, userId);
        if (CollectionUtil.isNotEmpty(list)) {
            for (CoinGoodsOrderVO order : list) {
                order.setGoodsPriceTypeName(WalletTypeEnum.toEnum(order.getGoodsPriceType()).getName());
            }
            pageList.setList(list);
        }
        return pageList;
    }
}
