package com.starbuds.server.service.finance.impl;

import cn.hutool.core.convert.Convert;
import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.finance.PayStatusEnum;
import com.starbuds.server.common.constant.finance.WithdrawPayTypeEnum;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.pojo.db.EasyTaxTransferHisDO;
import com.starbuds.server.common.pojo.finance.EasyTaxOrderNotify;
import com.starbuds.server.common.pojo.finance.EasyTaxOrderNotifyData;
import com.starbuds.server.common.pojo.finance.PaymentOrderDTO;
import com.starbuds.server.common.pojo.mq.QueueTransferUserMsg;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.finance.EasyTaxService;
import com.starbuds.server.service.finance.dao.EasyTaxTransferHisDAO;
import com.starbuds.server.service.finance.easyTax.provider.EasyTaxProvider;
import com.starbuds.server.service.finance.easyTax.response.PaymentBankPayResponse;
import com.starbuds.server.service.finance.rocket.RocketProducer;
import com.starbuds.server.service.finance.util.SpringContextUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;

import javax.annotation.Resource;
import java.math.BigDecimal;

import static java.math.BigDecimal.ROUND_HALF_UP;

/**
 * @Author jiangxiaowei
 * @DateTime 2020/11/27 16:36
 * @Description 易纳税
 */
@Slf4j
@Service(version = "1.0.0")
public class EasyTaxServiceImpl implements EasyTaxService {

    @Resource
    EasyTaxProvider easyTaxProvider;

    @Resource
    EasyTaxTransferHisDAO easyTaxTransferHisDAO;

    @Resource
    RocketProducer rocketProducer;

    @Reference(version = "1.0.0")
    IdentityService identityService;

    /**
     * 银⾏卡实时下单
     *
     * @param userId
     * @param realName
     * @param cardNo
     * @param idCard
     * @param amount
     * @param payRemark
     * @param bizId
     * @param payTypeEnum
     */
    @Override
    public PaymentOrderDTO payToUser(Long userId, String realName, String cardNo, String idCard, String amount, String payRemark, String bizId, WithdrawPayTypeEnum payTypeEnum) {
        String orderAmount = amount;
        // 测试环境
        if (SpringContextUtil.isDev()) {
            amount = "0.01";
        }
        Long orderId = identityService.getId();
        PaymentBankPayResponse response = easyTaxProvider.paymentBankPay(orderId, realName, cardNo, idCard, amount, payRemark, payTypeEnum);
        PaymentOrderDTO orderDTO = new PaymentOrderDTO();
        if (response.isSuccess()) {
            long curTime = System.currentTimeMillis();

            EasyTaxTransferHisDO hisDO = new EasyTaxTransferHisDO();
            hisDO.setOrderId(Convert.toStr(orderId));
            hisDO.setUserId(userId);
            hisDO.setCardNo(cardNo);
            hisDO.setIdCard(idCard);
            hisDO.setPayAmount(orderAmount);
            hisDO.setPayRemark(payRemark);
            hisDO.setPayStatus(PayStatusEnum.WAITTING.getValue());
            hisDO.setCreateTime(curTime);
            hisDO.setUpdateTime(curTime);
            hisDO.setBizId(bizId);
            hisDO.setRequestId(response.getRequestId());
            hisDO.setOutId(response.getData().getOrderNo());
            easyTaxTransferHisDAO.insert(hisDO);

            orderDTO.setSuccess(true);
            orderDTO.setOrderId(Convert.toStr(orderId));
            orderDTO.setPay(orderAmount);
            orderDTO.setRef(response.getData().getOrderNo());
        } else {
            orderDTO.setSuccess(false);
            orderDTO.setMessage(response.getMessage());
        }
        return orderDTO;
    }

    /**
     * 处理订单回调通知
     *
     * @param notify
     * @return
     */
    @Override
    public boolean handleOrderNotify(EasyTaxOrderNotify notify) {
        log.info("notify:{}", JSON.toJSONString(notify));
        if (notify == null) {
            log.info("notify为空");
            return false;
        }

        EasyTaxOrderNotifyData notifyData = JSON.parseObject(notify.getData(), EasyTaxOrderNotifyData.class);

        log.info("notifyData:{}", JSON.toJSONString(notifyData));

        String orderId = notifyData.getDealerOrderNo();
        String status = notifyData.getStatus();
        long curTime = System.currentTimeMillis();

        EasyTaxTransferHisDO transferHisDO = easyTaxTransferHisDAO.selectByOrderId(orderId);
        log.info("transferHis:{}", transferHisDO);
        if (transferHisDO == null || PayStatusEnum.WAITTING.getValue() != transferHisDO.getPayStatus()) {
            log.info("订单已处理");
            return true;
        }
        // 成功
        if ("3".equals(status)) {
            // 生产环境校验金额
            if (!SpringContextUtil.isDev()) {
                BigDecimal pay = new BigDecimal(transferHisDO.getPayAmount()).divide(BigDecimal.ONE, 2, ROUND_HALF_UP);
                BigDecimal broker = new BigDecimal(notifyData.getTradeAmount()).divide(BigDecimal.ONE, 2, ROUND_HALF_UP);
                if (pay.compareTo(broker) != 0) {
                    log.info("订单金额错误 {} {} compare:{}", notifyData.getTradeAmount(), transferHisDO.getPayAmount());
                    return false;
                }
            }
            easyTaxTransferHisDAO.updateResult(orderId, PayStatusEnum.SUCCESSED.getValue(), JSON.toJSONString(notify), curTime);

            QueueTransferUserMsg transferMsg = new QueueTransferUserMsg();
            transferMsg.setUserId(transferHisDO.getUserId());
            transferMsg.setAmount(transferHisDO.getPayAmount());
            transferMsg.setSuccess(true);
            transferMsg.setRef("粉钻转出");
            rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.TRANSFER_TO_USER, transferMsg.serialize());

            log.info("回调处理成功");
            return true;
        } else {
            easyTaxTransferHisDAO.updateResult(orderId, PayStatusEnum.FAILED.getValue(), JSON.toJSONString(notify), curTime);

            log.info("回调处理失败");
            return false;
        }
    }

    /**
     * 解密加密字符串
     *
     * @param encryptStr
     * @return
     */
    @Override
    public String decodeData(String encryptStr) {
        return easyTaxProvider.decodeData(encryptStr);
    }

    /**
     * 获取签名
     *
     * @param encryptStr
     * @param timestamp
     * @param requestId
     * @return
     */
    @Override
    public String getSign(String encryptStr, String timestamp, String requestId) {
        return easyTaxProvider.getSign(encryptStr, timestamp, requestId);
    }
}
