package com.starbuds.server.service.finance.impl;

import com.starbuds.server.common.constant.finance.GameCoinOrderApiPlatformEnum;
import com.starbuds.server.common.constant.finance.GameCoinOrderOperateEnum;
import com.starbuds.server.common.constant.finance.GameCoinOrderStatusEnum;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.pojo.daoplus.QueryWrapper;
import com.starbuds.server.common.pojo.daoplus.UpdateWrapper;
import com.starbuds.server.common.pojo.finance.GameCoinOrderDO;
import com.starbuds.server.common.service.finance.GameService;
import com.starbuds.server.common.service.finance.WalletService;
import com.starbuds.server.service.finance.provider.GameCoinOrderProvider;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;

import javax.annotation.Resource;

import static com.starbuds.server.common.constant.finance.GameCoinOrderOperateEnum.OUT;

/**
 * 游戏服务实现
 *
 * @author aoe
 * @date 2020/9/11
 */
@Slf4j
@Service(version = "1.0.0")
public class GameServiceImpl implements GameService {

    @Resource
    private GameCoinOrderProvider gameCoinOrderProvider;

    @Reference(version = "1.0.0")
    private WalletService walletService;

    /**
     * 添加游戏币订单
     *
     * @param userId          用户id
     * @param platform        平台代码
     * @param operate         操作（转入/转出）
     * @param apiPlatformEnum 接口平台
     * @param coin            游戏币数量
     * @return true:成功
     */
    @Override
    public Long addGameCoinOrder(Long userId,
                                 String platform,
                                 GameCoinOrderOperateEnum operate,
                                 GameCoinOrderApiPlatformEnum apiPlatformEnum,
                                 Long coin) {
        GameCoinOrderDO po = new GameCoinOrderDO();
        po.setUserId(userId);
        po.setPlatform(platform);
        po.setOperate(operate.getValue());
        po.setCoin(coin);
        po.setStatus(GameCoinOrderStatusEnum.PROCESSING.getValue());
        po.setApiPlatform(apiPlatformEnum.getValue());
        po = gameCoinOrderProvider.create(po);

        if (po == null) {
            return null;
        }
        return po.getOrderId();
    }

    /**
     * 更新游戏币订单
     *
     * @param orderId   订单id
     * @param isSuccess true:成功 false:失败
     * @param remarks
     * @return
     */
    @Override
    public boolean updateGameCoinOrder(Long orderId, boolean isSuccess, String remarks) {
        boolean exist = gameCoinOrderProvider.exist(orderId);
        if (!exist) {
            log.warn("订单不存在，不能执行加币/减币 orderId={}", orderId);
            return false;
        }
        GameCoinOrderDO po = gameCoinOrderProvider.getById(orderId);
        Integer status = po.getStatus();
        GameCoinOrderStatusEnum statusEnum = GameCoinOrderStatusEnum.toEnum(status);
        if (statusEnum != GameCoinOrderStatusEnum.PROCESSING) {
            log.warn("订单已处理，不能重复处理 orderId={}", orderId);
            return false;
        }

        Integer version = po.getVersion();
        GameCoinOrderStatusEnum updateStatusEnum = GameCoinOrderStatusEnum.FAIL;
        if (!isSuccess) {
            return updateGameCoinOrderStatus(orderId, version, updateStatusEnum, remarks);
        }

        boolean isAdd = OUT == GameCoinOrderOperateEnum.toEnum(po.getOperate());
        return addOrReduceCoin(po.getCoin(), orderId, isAdd, po.getUserId(), version);
    }

    /**
     * 获得游戏币订单状态
     *
     * @param orderId 订单id
     * @return
     */
    @Override
    public Integer getGameCoinOrderStatus(Long orderId) {
        return gameCoinOrderProvider.selectStatus(orderId);
    }

    /**
     * 判断是否存在
     *
     * @param id 主键
     * @return true:存在
     */
    @Override
    public boolean exist(Long id) {
        return gameCoinOrderProvider.exist(id);
    }

    /**
     * 转入在途总金额
     *
     * @param userId
     * @return
     */
    @Override
    public Long getInOnWayTotal(Long userId) {
        Long total = gameCoinOrderProvider.selectInOnWayTotal(userId);
        return total == null ? 0 : total;
    }

    /**
     * 加币/减币
     *
     * @param coin    游戏币的数量
     * @param orderId 游戏币订单id
     * @param isAdd   true:加币; false:减币
     * @return
     */
    private boolean addOrReduceCoin(Long coin, Long orderId, boolean isAdd, Long userId, Integer version) {
        // 更新钱包
        Long balance;
        if (isAdd) {
            balance = walletService.addCoin(userId, WalletTypeEnum.GAME_COIN.getValue(), coin, null, orderId);
        } else {
            balance = walletService.reduceCoin(userId, WalletTypeEnum.GAME_COIN.getValue(), coin, null, orderId);
        }
        GameCoinOrderStatusEnum updateStatusEnum = GameCoinOrderStatusEnum.SUCCESS;
        String remarks = null;
        if (balance == null || balance < 0) {
            log.warn("更新钱包失败，不能执行加币/减币 orderId={} balance={}", orderId, balance);
            updateStatusEnum = GameCoinOrderStatusEnum.FAIL;
            remarks = "加币/减币失败 balance=" + balance;
        }

        // 更新订单状态
        return updateGameCoinOrderStatus(orderId, version, updateStatusEnum, remarks);
    }

    private boolean updateGameCoinOrderStatus(Long orderId, Integer version, GameCoinOrderStatusEnum updateStatusEnum, String remarks) {
        QueryWrapper qw = new QueryWrapper();
        qw.eq("orderId", orderId);
        qw.eq("version", version);
        qw.ne("status", GameCoinOrderStatusEnum.SUCCESS.getValue());

        GameCoinOrderDO updatePo = new GameCoinOrderDO();
        updatePo.setVersion(version + 1);
        updatePo.setUpdateTime(System.currentTimeMillis());
        updatePo.setRemarks(remarks);
        updatePo.setStatus(updateStatusEnum.getValue());

        UpdateWrapper uw = new UpdateWrapper();
        uw.setQw(qw);
        uw.setPo(updatePo);

        boolean flag = gameCoinOrderProvider.update(uw);
        log.info("更新游戏币订单 flag={} update={}", flag, updatePo);
        return flag;
    }
}
