package com.starbuds.server.service.finance.impl;

import cn.hutool.core.convert.Convert;
import com.google.common.base.Preconditions;
import com.starbuds.server.common.constant.finance.IssueGoldCoinOrderTypeEnum;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.pojo.api.PageList;
import com.starbuds.server.common.pojo.db.IssueGoldCoinHisDO;
import com.starbuds.server.common.pojo.finance.qo.IssueGoldCoinHisPageListQO;
import com.starbuds.server.common.service.finance.IssueGoldCoinHisService;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.finance.provider.IssueGoldCoinHisProvider;
import com.starbuds.server.service.finance.provider.WalletProvider;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.math.NumberUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.Collection;
import java.util.Map;
import java.util.Optional;

/**
 * 系统发放金币流水
 */
@Service(version = "1.0.0")
@Slf4j
public class IssueGoldCoinHisServiceImpl implements IssueGoldCoinHisService {

    @Resource
    private IssueGoldCoinHisProvider issueGoldCoinHisProvider;

    @Resource
    private WalletProvider walletProvider;

    @Reference(version = "1.0.0")
    private StatPlatformService statPlatformService;

    @Reference(version = "1.0.0")
    private UserService userService;

    @Override
    public IssueGoldCoinHisDO getById(Long id) {
        return issueGoldCoinHisProvider.getById(id);
    }

    @Override
    public boolean exist(Long id) {
        return issueGoldCoinHisProvider.exist(id);
    }

    @Override
    public PageList<IssueGoldCoinHisDO> getPageList(IssueGoldCoinHisPageListQO qo) {
        return issueGoldCoinHisProvider.getPageList(qo);
    }

    @Override
    public Map<Long, IssueGoldCoinHisDO> getMapByIds(Collection ids) {
        return issueGoldCoinHisProvider.getMapByIds(ids);
    }

    // 手动编码开始

    /**
     * 给用户发放金币
     *
     * @param receiverUserId 获得金币的用户id
     * @param amount         金币数量(必须大于 0)
     * @param orderId        产生金币的订单id
     * @param orderType      订单类型
     * @param remarkPrefix   发放说明前缀
     * @return true 成功
     */
    @Override
    @Transactional(rollbackFor = Exception.class)
    public boolean payGoldCoin(long receiverUserId, int amount, long orderId,
                               IssueGoldCoinOrderTypeEnum orderType, String remarkPrefix) {
        Preconditions.checkNotNull(orderType, "orderType 不能为空");
        Preconditions.checkArgument(amount > 0, "发放金币数量必须大于 0");
        Preconditions.checkNotNull(remarkPrefix, "remarkPrefix 不能为 null");

        IssueGoldCoinHisDO po = new IssueGoldCoinHisDO();
        po.setUserId(receiverUserId);
        po.setAmount(amount);
        po.setOrderId(orderId);
        po.setOrderType(orderType.getValue());
        po = issueGoldCoinHisProvider.create(po);
        if (po == null) {
            log.error("保存[系统发放金币流水]失败 receiverUserId={} amount={} orderId={} orderType={}",
                    receiverUserId, amount, orderId, orderType);
            return false;
        }

        String remark = String.format("[%s]发放%d个金币", remarkPrefix, amount);
        Long addBalance = walletProvider.addBalance(WalletTypeEnum.GOLD_COIN, receiverUserId, amount, remark, orderId);
        if (addBalance == null) {
            log.error("[系统发放金币-增加钱包余额]失败 receiverUserId={} amount={} orderId={} orderType={}",
                    receiverUserId, amount, orderId, orderType);
            return false;
        }

        Long channelId = Optional.ofNullable(userService.getChannelId(receiverUserId)).orElse(0L);
        statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.ISSUE_GOLD_COIN_AMOUNT, BigDecimal.valueOf(amount));

        return true;
    }
}