package com.starbuds.server.service.finance.impl;

import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.finance.*;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.pojo.daoplus.QueryWrapper;
import com.starbuds.server.common.pojo.daoplus.UpdateWrapper;
import com.starbuds.server.common.pojo.db.UserAuthDO;
import com.starbuds.server.common.pojo.finance.ThirdPartyCallbackHisDO;
import com.starbuds.server.common.pojo.finance.ThirdPartyCallbackHisSaveBO;
import com.starbuds.server.common.pojo.finance.ThirdPartyWithdrawHisDO;
import com.starbuds.server.common.pojo.finance.bo.ThirdPartyCallbackBO;
import com.starbuds.server.common.pojo.finance.bo.WithdrawRequestBO;
import com.starbuds.server.common.pojo.finance.bo.WithdrawRequestRecordBO;
import com.starbuds.server.common.pojo.mq.QueueTransferUserMsg;
import com.starbuds.server.common.service.finance.CaiyuService;
import com.starbuds.server.common.service.finance.ThirdPartyService;
import com.starbuds.server.common.service.finance.WithdrawRequestService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.finance.provider.CaiyuContractProvider;
import com.starbuds.server.service.finance.provider.ThirdPartyCallbackHisProvider;
import com.starbuds.server.service.finance.provider.ThirdPartyWithdrawHisProvider;
import com.starbuds.server.service.finance.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.beans.BeanUtils;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;

import java.math.BigDecimal;
import java.util.HashMap;
import java.util.Map;

import static com.starbuds.server.common.constant.finance.ThirdPartyWithdrawStatusEnum.*;
import static com.starbuds.server.common.constant.finance.WithdrawPayTypeEnum.*;

/**
 * 第三方相关服务
 *
 * @author aoe
 * @date 2021/1/20
 */
@Service(version = "1.0.0")
@Slf4j
public class ThirdPartyServiceImpl implements ThirdPartyService {

    @Resource
    private ThirdPartyCallbackHisProvider thirdPartyCallbackHisProvider;

    @Resource
    private ThirdPartyWithdrawHisProvider thirdPartyWithdrawHisProvider;

    @Resource
    private RocketProducer rocketProducer;

    private BigDecimal HUNDRED = new BigDecimal("100");

    @Resource
    private CaiyuContractProvider caiyuContractProvider;

    @Reference(version = "1.0.0")
    private UserService userService;

    @Resource
    private CaiyuService caiyuService;

    @Resource(name = "yunxiaokeServiceImpl")
    private WithdrawRequestService yunxiaokeService;

    /**
     * 充值类型:第三方公司
     */
    private Map<WithdrawPayTypeEnum, ThirdPartyCompanyEnum> companyEnumMap;

    /**
     * 充值类型:第三方公司提现请求服务实例
     */
    private Map<WithdrawPayTypeEnum, WithdrawRequestService> thirdPartyServiceMap;

    @PostConstruct
    private void init() {
        companyEnumMap = new HashMap<>();
        companyEnumMap.put(CAIYU_TRANSFER, ThirdPartyCompanyEnum.CAI_YU);
        companyEnumMap.put(YUNXIAOKE_TRANSFER, ThirdPartyCompanyEnum.YUN_XIAO_KE);

        thirdPartyServiceMap = new HashMap<>();
        thirdPartyServiceMap.put(CAIYU_TRANSFER, caiyuService);
        thirdPartyServiceMap.put(YUNXIAOKE_TRANSFER, yunxiaokeService);
    }

    /**
     * 保存第三方回调记录
     *
     * @param his
     * @return
     */
    @Override
    public boolean saveCallbackHis(ThirdPartyCallbackHisSaveBO his) {
        ThirdPartyCallbackHisDO po = new ThirdPartyCallbackHisDO();
        BeanUtils.copyProperties(his, po);
        ThirdPartyCallbackEnum infoType = his.getInfoType();
        po.setCompanyId(infoType.getCompany().getId());
        po.setBusinessType(infoType.getBusinessType().getValue());
        po.setKeywordDesc(infoType.getKeywordDesc());

        return thirdPartyCallbackHisProvider.create(po) == null ? false : true;
    }

    /**
     * 用户是否需要签约
     *
     * @param userId
     * @param payTypeEnum
     * @return true:需要签约操作; false:不需要签约操作
     */
    @Override
    public boolean isNeedSignContractOperation(Long userId, WithdrawPayTypeEnum payTypeEnum) {
        if (CAIYU_TRANSFER != payTypeEnum) {
            return false;
        }

        // 增加新类型时整理
        return !caiyuContractProvider.exist(userId);
    }

    @Override
    public String signContract(Long userId, WithdrawPayTypeEnum payTypeEnum) {
        if (CAIYU_TRANSFER != payTypeEnum) {
            return "该支付类型不支持签约";
        }
        UserAuthDO user = userService.getUserAuth(userId);
        if (user == null) {
            return "用户没有实名认证";
        }
        if (StringUtils.isBlank(user.getRealName()) || StringUtils.isBlank(user.getIdCardNo())) {
            return "用户实名信息不完整";
        }
        String phone = userService.getPhone(userId);
        if (StringUtils.isBlank(phone)) {
            return "用户手机号不存在";
        }

        String result = caiyuService.signContractNoBank(userId, user.getRealName(), phone, user.getIdCardNo(), null, null);

        return result;
    }

    /**
     * 发送提现请求并返回相关记录
     *
     * @param wr
     * @return 提现请求发送记录
     */
    @Override
    public WithdrawRequestRecordBO sendWithdrawRequest(WithdrawRequestBO wr) {
        WithdrawRequestRecordBO wrd = new WithdrawRequestRecordBO();

        // 验证数据
        WithdrawPayTypeEnum payTypeEnum = wr.getPayTypeEnum();
        ThirdPartyCompanyEnum company = companyEnumMap.get(payTypeEnum);
        if (company == null) {
            wrd.setErrMsg(payTypeEnum.getName() + "没有匹配的支付公司");
            return wrd;
        }
        WithdrawRequestService withdrawRequestService = thirdPartyServiceMap.get(wr.getPayTypeEnum());
        if (withdrawRequestService == null) {
            wrd.setErrMsg(payTypeEnum.getName() + " : " + company.getName() + "没有实现提现逻辑");
            return wrd;
        }

        // 验证余额
        BigDecimal balance = withdrawRequestService.getBalance();
        if (balance == null) {
            wrd.setErrMsg(payTypeEnum.getName() + "查询余额失败");
            return wrd;
        }
        if (wr.getAmount().compareTo(balance) == 1) {
            wrd.setErrMsg(payTypeEnum.getName() + "余额不足！余额:" + balance + "元 提现金额:" + wr.getAmount() + "元");
            return wrd;
        }

        // 执行提现操作
        wrd = withdrawRequestService.doWithdraw(wr);
        ThirdPartyWithdrawHisDO po = new ThirdPartyWithdrawHisDO();
        Long uwwHisId = wr.getUserWalletWithdrawHisId();
        po.setUserWalletWithdrawHisId(uwwHisId);
        po.setUserId(wr.getUserId());
        po.setAmount(wr.getAmount());
        po.setAccountType(wr.getAccountTypeEnum().getValue());
        po.setReceivingAccount(wr.getReceivingAccount());
        po.setCompanyId(company.getId());
        po.setRequestTimes(1);
        po.setRequestData(wrd.getRequest());
        po.setRequestUrl(wrd.getUrl());
        po.setResponseData(wrd.getResponse());
        po.setResponseIsSuccess(wrd.getIsSuccess() ? BoolIntVal.True : BoolIntVal.False);
        po.setResponseErrMsg(wrd.getErrMsg());
        po.setSpendTime(wrd.getSpendTime().intValue());
        po.setWithdrawStatus(HAS_BEEN_SENT.getValue());

        // 云销客的接口只能靠这个关联
        String outId = wrd.getOutId();
        if (StringUtils.isNotBlank(outId)) {
            po.setOutId(outId);
        }

        if (thirdPartyWithdrawHisProvider.exist(uwwHisId)) {
            ThirdPartyWithdrawHisDO tpwPO = thirdPartyWithdrawHisProvider.getByUserWalletWithdrawHisId(uwwHisId);
            Integer requestTimes = tpwPO.getRequestTimes();

            po.setRequestTimes(requestTimes + 1);
            po.setUpdateTime(System.currentTimeMillis());

            QueryWrapper qw = new QueryWrapper();
            qw.eq("userWalletWithdrawHisId", uwwHisId);
            qw.eq("updateTime", tpwPO.getUpdateTime());
            UpdateWrapper uw = new UpdateWrapper();
            uw.setQw(qw);
            uw.setPo(po);
            thirdPartyWithdrawHisProvider.update(uw);
            log.info("提现记录更新 {}", JSON.toJSONString(po, true));
        } else {
            thirdPartyWithdrawHisProvider.create(po);
            log.info("提现记录新增 {}", JSON.toJSONString(po, true));
        }

        return wrd;
    }

    @Override
    public boolean processWithdrawCallback(ThirdPartyCallbackBO callback) {
        if (callback == null) {
            log.warn("第三方回调失败 callback 对象 is null");
            return false;
        }
        Long uwwHisId = callback.getUserWalletWithdrawHisId();
        if (uwwHisId == null) {
            log.warn("第三方回调失败 userWalletWithdrawHisId is null");
            return false;
        }
        if (callback.getIsSuccess() == null) {
            log.warn("第三方回调失败 userWalletWithdrawHisId = {} 回调返回的处理结果是否成功 isSuccess is null", uwwHisId);
            return false;
        }
        // 检查订单是否已经处理
        ThirdPartyWithdrawHisDO po = thirdPartyWithdrawHisProvider.getByUserWalletWithdrawHisId(uwwHisId);
        if (po == null) {
            log.warn("第三方回调失败 userWalletWithdrawHisId = {} 没有提现请求发起记录", uwwHisId);
            return false;
        }
        if (HAS_BEEN_SENT != ThirdPartyWithdrawStatusEnum.toEnum(po.getWithdrawStatus())) {
            log.warn("第三方回调失败 userWalletWithdrawHisId = {} 已处理，不能重复处理", uwwHisId);
            return false;
        }

        // 更新提现记录状态
        QueryWrapper qw = new QueryWrapper();
        qw.eq("userWalletWithdrawHisId", uwwHisId);
        qw.eq("updateTime", po.getUpdateTime());
        UpdateWrapper uw = new UpdateWrapper();
        uw.setQw(qw);

        po.setOutId(callback.getOutId());
        po.setCallbackData(callback.getCallbackData());
        po.setCallbackMsg(callback.getCallbackMsg());
        po.setWithdrawStatus(callback.getIsSuccess() ? CALLBACK_SUSSES.getValue() : CALLBACK_FAIL.getValue());
        po.setUpdateTime(System.currentTimeMillis());
        uw.setPo(po);


//        if (callback.getIsSuccess()) {
//            //withdrawStatus 为新加字段，原有逻辑中没有赋值为null，需要先判断
//            if (callback.getWithdrawStatus() == null) {
//                po.setWithdrawStatus(CALLBACK_SUSSES.getValue());
//            }
//            if (callback.getWithdrawStatus() == CALLBACK_TICREF){
//                //如果原来不是成功状态则设置为：退票状态；成功状态则设置为：成功后再退票 这个状态
//                boolean isSuccessed = po.getWithdrawStatus().compareTo(CALLBACK_SUSSES.getValue()) == 0;
//                po.setWithdrawStatus(isSuccessed ? CALLBACK_SUSSES_TICREF.getValue() : CALLBACK_TICREF.getValue());
//            }
//            else {
//                po.setWithdrawStatus(CALLBACK_SUSSES.getValue());
//            }
//        }
//        else {
//            po.setWithdrawStatus(CALLBACK_FAIL.getValue());
//        }

        boolean flag = thirdPartyWithdrawHisProvider.update(uw);

        // 发送消息：粉钻转出
        if (flag) {
            QueueTransferUserMsg transferMsg = new QueueTransferUserMsg();
            transferMsg.setUserId(po.getUserId());
            // 换算粉钻数量
            Integer pinkAmount = po.getAmount().multiply(HUNDRED).intValue();
            transferMsg.setAmount(pinkAmount.toString());
            transferMsg.setSuccess(true);
            transferMsg.setRef("粉钻转出");
            rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.TRANSFER_TO_USER, transferMsg.serialize());
            log.info("第三方回调成功 userWalletWithdrawHisId = {} 发送消息：粉钻转出 {}", uwwHisId, transferMsg);
        } else {
            log.warn("第三方回调失败 更新提现记录状态失败 {}", uw);
        }
        return flag;
    }

    @Override
    public ThirdPartyWithdrawHisDO getByOutId(String outId, ThirdPartyCompanyEnum companyEnum) {
        return thirdPartyWithdrawHisProvider.getByOutId(outId, companyEnum);
    }


}
