package com.starbuds.server.service.finance.impl;

import com.starbuds.server.common.constant.finance.PayStatusEnum;
import com.starbuds.server.common.constant.finance.PayTypeEnum;
import com.starbuds.server.common.pojo.db.PayOrderDO;
import com.starbuds.server.common.pojo.db.UserVipDO;
import com.starbuds.server.common.pojo.db.VipItemDO;
import com.starbuds.server.common.pojo.db.VipOrderDO;
import com.starbuds.server.common.pojo.finance.VipItemVO;
import com.starbuds.server.common.pojo.finance.VipOrderDTO;
import com.starbuds.server.common.pojo.user.UserCardVO;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.finance.UserVipVO;
import com.starbuds.server.common.service.finance.VipService;
import com.starbuds.server.service.finance.dao.PayOrderDAO;
import com.starbuds.server.service.finance.dao.UserVipDAO;
import com.starbuds.server.service.finance.dao.VipItemDAO;
import com.starbuds.server.service.finance.dao.VipOrderDAO;
import com.starbuds.server.service.finance.provider.VipProvider;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.beans.BeanUtils;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.util.List;

/**
 * VIP 服务实现
 *
 * @author Lay
 * @date 2020/4/4
 */
@Service(version = "1.0.0")
@Slf4j
public class VipServiceImpl implements VipService {

    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Resource
    VipProvider vipProvider;

    @Resource
    VipItemDAO vipItemDAO;

    @Resource
    VipOrderDAO vipOrderDAO;

    @Resource
    PayOrderDAO payOrderDAO;

    @Resource
    UserVipDAO userVipDAO;

    /**
     * 获取 vip 项目
     *
     * @return
     */
    @Override
    public List<VipItemVO> getVipItems() {
        return vipItemDAO.selectAll();
    }

    /**
     * 获取 vip 项目
     *
     * @param itemId
     * @return
     */
    @Override
    public VipItemDO getVipItem(Long itemId) {
        return vipItemDAO.selectById(itemId);
    }

    /**
     * 创建会员订单
     *
     * @param vipOrderDTO
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public Long addVipOrder(VipOrderDTO vipOrderDTO) {
        VipItemDO vipItem = vipItemDAO.selectById(vipOrderDTO.getItemId());
        if (vipItem != null) {
            try {
                Long orderId = identityService.getId();
                Long curTime = System.currentTimeMillis();

                // 创建会员订单
                VipOrderDO vipOrder = new VipOrderDO();
                vipOrder.setOrderId(orderId);
                vipOrder.setUserId(vipOrderDTO.getUserId());
                vipOrder.setItemId(vipOrderDTO.getItemId());
                vipOrder.setItemName(vipItem.getItemName());
                vipOrder.setVipPrice(vipItem.getVipPrice());
                vipOrder.setVipDays(vipItem.getVipDays());
                vipOrder.setOrderCreateTime(curTime);
                vipOrder.setOrderUpdateTime(curTime);
                vipOrder.setSingleChatMinutes(vipItem.getSingleChatMinutes());
                vipOrder.setSingleChatMinutesPerDay(vipItem.getSingleChatMinutesPerDay());
                vipOrder.setVipLevel(vipItem.getVipLevel());
                vipOrder.setChannelId(vipOrderDTO.getChannelId());
                vipOrderDAO.insert(vipOrder);

                // 创建支付订单
                String subject = String.format("购买%s", vipItem.getItemName());
                PayOrderDO payOrderDO = new PayOrderDO();
                payOrderDO.setOrderId(orderId);
                payOrderDO.setUserId(vipOrderDTO.getUserId());
                payOrderDO.setChannelId(vipOrderDTO.getChannelId());
                payOrderDO.setPayType(PayTypeEnum.VIP.getValue());
                payOrderDO.setPayWay(vipOrderDTO.getPayWay());
                payOrderDO.setOrderAmount(vipItem.getVipPrice());
                payOrderDO.setPayStatus(PayStatusEnum.WAITTING.getValue());
                payOrderDO.setPayOrderCreateTime(curTime);
                payOrderDO.setCreateIp(vipOrderDTO.getClientIp());
                payOrderDO.setPaySubject(subject);
                payOrderDO.setPayBody(subject);
                payOrderDAO.insert(payOrderDO);

                return orderId;
            } catch (Exception e) {
                TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
                log.error("会员订单创建失败", e);
            }
        }
        return null;
    }

    /**
     * 获取用户会员信息
     *
     * @param userId
     * @return
     */
    @Override
    public UserVipVO getUserVip(Long userId) {
        UserVipDO userVipDO = userVipDAO.selectById(userId);
        if (userVipDO != null) {
            UserVipVO userVip = new UserVipVO();
            BeanUtils.copyProperties(userVipDO, userVip);
            return userVip;
        }
        return null;
    }

    /**
     * 用户是否会员
     *
     * @param userId
     * @return
     */
    @Override
    public boolean isVip(Long userId) {
        return userVipDAO.countById(userId) > 0;
    }

    /**
     * 领取每日免费时长
     *
     * @param userId
     * @return
     */
    @Override
    public boolean getDayFreeMinutes(Long userId) {
        return vipProvider.getDayFreeMinutes(userId);
    }

    /**
     * 查询最近开通的 vip 列表
     *
     * @return
     */
    @Override
    public List<UserCardVO> getLatestVip() {
        return userVipDAO.selectLatest();
    }
}
