package com.starbuds.server.service.finance.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import cn.hutool.core.date.DateField;
import cn.hutool.core.date.DateTime;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.lang.TypeReference;
import cn.hutool.core.util.StrUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.serializer.SerializerFeature;
import com.starbuds.server.common.constant.cache.RedisFinanceKeys;
import com.starbuds.server.common.constant.common.CommonStatus;
import com.starbuds.server.common.constant.finance.RefTypeEnum;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.finance.WalletWithdrawHisApplyStatus;
import com.starbuds.server.common.constant.finance.WithdrawPayTypeEnum;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.pojo.api.PageList;
import com.starbuds.server.common.pojo.api.RecordList;
import com.starbuds.server.common.pojo.data.PropsVO;
import com.starbuds.server.common.pojo.db.*;
import com.starbuds.server.common.pojo.finance.*;
import com.starbuds.server.common.pojo.finance.bo.WithdrawRequestBO;
import com.starbuds.server.common.pojo.finance.bo.WithdrawRequestRecordBO;
import com.starbuds.server.common.pojo.mq.QueueAnchorWithdrawMsg;
import com.starbuds.server.common.pojo.mq.QueueUserWithdrawMsg;
import com.starbuds.server.common.pojo.user.AnchorSignedRatio;
import com.starbuds.server.common.pojo.user.UserAlipayVO;
import com.starbuds.server.common.pojo.util.TimeSpan;
import com.starbuds.server.common.service.core.CacheService;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.finance.*;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.common.service.user.AgentService;
import com.starbuds.server.common.service.user.ThirdAuthService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.common.utils.GameUtil;
import com.starbuds.server.service.finance.dao.*;
import com.starbuds.server.service.finance.provider.*;
import com.starbuds.server.service.finance.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;
import org.springframework.util.CollectionUtils;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.stream.Collectors;

import static com.starbuds.server.common.constant.finance.ThirdPartyReceivingAccountTypeEnum.BANK_CARD;
import static com.starbuds.server.common.constant.finance.WithdrawPayTypeEnum.*;
import static com.starbuds.server.service.finance.util.SpringContextUtil.isDev;

/**
 * 钱包服务实现
 *
 * @author Lay
 * @date 2019/12/21
 */
@Service(version = "1.0.0")
@Slf4j
public class WalletServiceImpl implements WalletService {

    @Reference(version = "1.0.0")
    ThirdAuthService thirdAuthService;

    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Reference(version = "1.0.0")
    AgentService agentService;

    @Reference(version = "1.0.0")
    CacheService cacheService;

    @Reference(version = "1.0.0")
    PayService payService;

    @Reference(version = "1.0.0")
    CloudAccountService cloudAccountService;

    @Reference(version = "1.0.0")
    EasyTaxService easyTaxService;

    @Reference(version = "1.0.0")
    UserService userService;

    @Reference(version = "1.0.0")
    StatPlatformService statPlatformService;

    @Resource
    SailtobService sailtobService;

    @Resource
    BagProvider bagProvider;

    @Resource
    IWalletProvider walletProvider;

    @Resource
    AgentWalletProvider agentWalletProvider;

    @Resource
    ChatTimeWalletProvider chatTimeWalletProvider;

    @Resource
    private WalletTransDustToPinkHisProvider walletTransDustToPinkHisProvider;

    @Resource
    UserWalletWithdrawHisDAO userWalletWithdrawHisDAO;

    @Resource
    VirtualCoinDAO virtualCoinDAO;

    @Resource
    UserBankcardDAO bankcardDAO;

    @Resource
    UserWalletHisDAO userWalletHisDAO;

    @Resource
    UserWalletTransferHisDAO userWalletTransferHisDAO;

    @Resource
    RocketProducer rocketProducer;

    @Resource
    private ThirdPartyService thirdPartyService;

    /**
     * 个税比例
     */
    @Value("${userTaxRatio}")
    private String userTaxRatioStr;

    /**
     * 粉->钻手续费比率
     */
    @Value("${transDustToDiamondRatio}")
    private String transDustToDiamondRatioStr;

    @Value("${name.diamond}")
    private String nameDiamond;

    /**
     * 个税比例
     */
    private double userTaxRatio;

    /**
     * 粉->钻手续费比率
     */
    private double transDustToDiamondRatio;

    @PostConstruct
    private void init() {
        userTaxRatio = NumberUtils.toDouble(userTaxRatioStr, 0.08);
        transDustToDiamondRatio = NumberUtils.toDouble(transDustToDiamondRatioStr, 0.02);
        log.info("个税比例 {} 粉->钻手续费比率 {}", userTaxRatio, transDustToDiamondRatio);
    }

    /**
     * 获取钱包类型名称哈希表
     *
     * @return
     */
    @Override
    public Map<Integer, String> getWalletTypeNameMap() {
        String typeNameStr = cacheService.get(RedisFinanceKeys.WALLET_TYPE_NAME);
        if (StringUtils.isBlank(typeNameStr)) {
            return refreshWalletTypeNameCache();
        }
        Map<Integer, String> map = JSON.parseObject(typeNameStr, new TypeReference<Map<Integer, String>>() {
        });
        return Optional.ofNullable(map).orElse(new HashMap<>());
    }

    /**
     * 刷新缓存
     *
     * @return
     */
    @Override
    public Map<Integer, String> refreshWalletTypeNameCache() {
        try {
            Map<Integer, String> typeNameMap = new HashMap<>();
            // 默认类型
            for (WalletTypeEnum walletType : WalletTypeEnum.values()) {
                typeNameMap.put(walletType.getValue(), walletType.getName());
            }
            List<VirtualCoinDO> list = virtualCoinDAO.selectAll();
            if (CollectionUtil.isNotEmpty(list)) {
                typeNameMap.putAll(list.stream().collect(Collectors.toMap(VirtualCoinDO::getCoinId, VirtualCoinDO::getCoinName)));
            }
            cacheService.set(RedisFinanceKeys.WALLET_TYPE_NAME, JSON.toJSONString(typeNameMap, SerializerFeature.WriteNonStringKeyAsString));
            return typeNameMap;
        } catch (Exception e) {
            log.error("缓存刷新失败", e);
        }
        return new HashMap<>();
    }

    /**
     * 查询用户钱包
     *
     * @param userId
     * @return
     */
    @Override
    public UserWalletVO getUserWallet(Long userId) {
        return walletProvider.getUserWallet(userId);
    }

    /**
     * 查询用户虚拟币钱包
     *
     * @param userId
     * @return
     */
    @Override
    public UserWalletVO getUserVirtualWallet(Long userId) {
        return walletProvider.getUserVirtualWallet(userId);
    }

    /**
     * 查询钱包余额
     *
     * @param userId
     * @param walletType
     * @return
     */
    @Override
    public Long getBalance(Long userId, WalletTypeEnum walletType) {
        return walletProvider.getBalance(userId, walletType);
    }

    /**
     * 查询钱包余额
     *
     * @param userId
     * @param walletType
     * @return
     */
    @Override
    public Long getBalance(Long userId, Integer walletType) {
        return walletProvider.getBalance(userId, walletType);
    }

    /**
     * 创建用户钱包
     *
     * @param userId
     * @return
     */
    @Override
    public boolean createUserWallet(Long userId) {
        return walletProvider.createUserWallet(userId);
    }

    /**
     * 获取用户钱包明细
     *
     * @param lastId
     * @param pageSize
     * @param userId
     * @param walletType
     * @param yearMonth
     * @param isIncome
     * @return
     */
    @Override
    public RecordList<UserWalletHisVO, Long> getUserWalletHisListByTypeAndMonth(Long lastId, int pageSize, Long userId, WalletTypeEnum walletType, Integer yearMonth, Integer isIncome) {
        return getUserWalletHisListByTypeAndMonth(lastId, pageSize, userId, walletType.getValue(), yearMonth, isIncome);
    }

    /**
     * 获取用户钱包明细
     *
     * @param lastId
     * @param pageSize
     * @param userId
     * @param walletType
     * @param yearMonth
     * @param isIncome
     * @return
     */
    @Override
    public RecordList<UserWalletHisVO, Long> getUserWalletHisListByTypeAndMonth(Long lastId, int pageSize, Long userId, Integer walletType, Integer yearMonth, Integer isIncome) {
        RecordList<UserWalletHisVO, Long> recordList = new RecordList<>(pageSize);

        // 处理时间
        TimeSpan ts = getTimeSpanByYearMonth(yearMonth);
        List<UserWalletHisVO> walletHisVOList = walletProvider.getUserWalletHisList(userId, lastId, pageSize, walletType, isIncome, ts.getBeginTime(), ts.getEndTime());

        if (CollectionUtil.isNotEmpty(walletHisVOList)) {
            lastId = walletHisVOList.get(walletHisVOList.size() - 1).getHisId();
        }

        recordList.setList(walletHisVOList);
        recordList.setLastId(lastId);
        return recordList;
    }

    /**
     * 获取用户钱包明细
     *
     * @param lastId
     * @param pageSize
     * @param userId
     * @param walletType
     * @param beginTime
     * @param endTime
     * @param isIncome
     * @return
     */
    @Override
    public RecordList<UserWalletHisVO, Long> getUserWalletHisListByTypeAndTime(Long lastId, int pageSize, Long userId, Integer walletType, Long beginTime, Long endTime, Integer isIncome) {
        RecordList<UserWalletHisVO, Long> recordList = new RecordList<>(pageSize);

        // 处理时间
        List<UserWalletHisVO> walletHisVOList = walletProvider.getUserWalletHisList(userId, lastId, pageSize, walletType, isIncome, beginTime, endTime);

        if (CollectionUtil.isNotEmpty(walletHisVOList)) {
            lastId = walletHisVOList.get(walletHisVOList.size() - 1).getHisId();
        }

        recordList.setList(walletHisVOList);
        recordList.setLastId(lastId);
        return recordList;
    }

    /**
     * 获取用户钱包求和值
     *
     * @param userId
     * @param walletType
     * @param yearMonth
     * @param isIncome
     * @return
     */
    @Override
    public long getUserWalletHisSumByTypeAndMonth(Long userId, WalletTypeEnum walletType, Integer yearMonth, Integer isIncome) {
        return getUserWalletHisSumByTypeAndMonth(userId, walletType.getValue(), yearMonth, isIncome);
    }

    /**
     * 获取用户钱包求和值
     *
     * @param userId
     * @param walletType
     * @param yearMonth
     * @param isIncome
     * @return
     */
    @Override
    public long getUserWalletHisSumByTypeAndMonth(Long userId, Integer walletType, Integer yearMonth, Integer isIncome) {
        TimeSpan ts = getTimeSpanByYearMonth(yearMonth);
        return Optional.ofNullable(walletProvider.sumByTypeAndTime(userId, walletType, isIncome, ts.getBeginTime(), ts.getEndTime())).orElse(0L);
    }

    /**
     * 获取用户钱包求和值
     *
     * @param userId
     * @param walletType
     * @param beginTime
     * @param endTime
     * @param isIncome
     * @return
     */
    @Override
    public long getUserWalletHisSumByTypeAndTime(Long userId, Integer walletType, Long beginTime, Long endTime, Integer isIncome) {
        return Optional.ofNullable(walletProvider.sumByTypeAndTime(userId, walletType, isIncome, beginTime, endTime)).orElse(0L);
    }

    /**
     * 加币
     *
     * @param userId
     * @param walletType
     * @param amount
     * @param remarks
     * @param orderId
     * @return
     */
    @Override
    public Long addCoin(Long userId, Integer walletType, Long amount, String remarks, Long orderId) {
        return walletProvider.addBalance(WalletTypeEnum.toEnum(walletType), userId, amount, remarks, orderId);
    }

    /**
     * 扣币
     *
     * @param userId
     * @param walletType
     * @param amount
     * @param remarks
     * @param orderId
     * @return
     */
    @Override
    public Long reduceCoin(Long userId, Integer walletType, Long amount, String remarks, Long orderId) {
        return walletProvider.reduceBalance(WalletTypeEnum.toEnum(walletType), userId, amount, remarks, orderId);
    }

    /**
     * 创建提现申请
     *
     * @param userId
     * @param walletType
     * @param amount
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public Long addWithdrawApply(Long userId, WalletTypeEnum walletType, Long amount) {
        UserAlipayVO userAlipayVO = thirdAuthService.getAlipay(userId);
        if (userAlipayVO == null) {
            return null;
        }

        try {
            Long hisId = identityService.getId();
            // 扣钱
            Long newBalance = walletProvider.reduceBalance(walletType, userId, amount, "提现转出", hisId);

            // 创建记录
            if (newBalance >= 0) {
                UserWalletWithdrawHisDO userWalletWithdrawHisDO = new UserWalletWithdrawHisDO();

                // 读取工会和主播比例
                BigDecimal actualAmount;
                double userTaxRatio = 0.1d;
                AnchorSignedRatio anchorSignedRatio = agentService.getAnchorSignedRatio(userId);
                if (anchorSignedRatio != null) {
                    userWalletWithdrawHisDO.setAgentId(anchorSignedRatio.getAgentId());
                    userWalletWithdrawHisDO.setAgentRatio(anchorSignedRatio.getAgentRatio());
                    userWalletWithdrawHisDO.setUserRatio(anchorSignedRatio.getAnchorRatio());
                    BigDecimal agentAmount = BigDecimal.valueOf(amount).multiply(new BigDecimal(anchorSignedRatio.getAgentRatio())).divide(BigDecimal.valueOf(100), 2, RoundingMode.DOWN);
                    // 主播跟工会分账
                    actualAmount = agentAmount.multiply(new BigDecimal(anchorSignedRatio.getAnchorRatio())).setScale(2, RoundingMode.DOWN);
                    userWalletWithdrawHisDO.setAgentAmount(agentAmount.subtract(actualAmount));
                } else {
                    userWalletWithdrawHisDO.setUserRatio(0.5d);
                    actualAmount = BigDecimal.valueOf(amount).multiply(new BigDecimal("0.5")).divide(BigDecimal.valueOf(100), 2, RoundingMode.DOWN);
                }

                // 扣税
                BigDecimal taxAmount = actualAmount.multiply(new BigDecimal(String.valueOf(userTaxRatio))).setScale(2, RoundingMode.HALF_UP);
                actualAmount = actualAmount.subtract(taxAmount).setScale(2, RoundingMode.DOWN);

                userWalletWithdrawHisDO.setHisId(hisId);
                userWalletWithdrawHisDO.setWalletId(WalletProvider.getWalletId(userId, walletType.getValue()));
                userWalletWithdrawHisDO.setAmount(amount);
                userWalletWithdrawHisDO.setApplyStatus(WalletWithdrawHisApplyStatus.WAIT.getValue());
                userWalletWithdrawHisDO.setHisCreateTime(System.currentTimeMillis());
                userWalletWithdrawHisDO.setAlipayUserId(userAlipayVO.getAlipayUserId());
                userWalletWithdrawHisDO.setAlipayAccount(userAlipayVO.getAlipayAccount());
                userWalletWithdrawHisDO.setActualAmount(actualAmount);
                userWalletWithdrawHisDO.setUserTaxRatio(userTaxRatio);
                userWalletWithdrawHisDO.setTaxAmount(taxAmount);
                userWalletWithdrawHisDO.setRefType(RefTypeEnum.ANCHOR_WITHDRAW.getValue());

                userWalletWithdrawHisDAO.insert(userWalletWithdrawHisDO);
            }

            return newBalance;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("提现记录创建失败", e);
        }
        return null;
    }

    /**
     * 创建提现申请
     *
     * @param userId
     * @param walletType
     * @param amount
     * @param cardNo
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public WithdrawApplyDTO addWithdrawApplyByStarDust(Long userId, WalletTypeEnum walletType, Long amount, String cardNo) {

        try {
            Long hisId = identityService.getId();
            WithdrawApplyDTO applyDTO = new WithdrawApplyDTO();
            // 扣钱
            Long newBalance = walletProvider.reduceBalance(walletType, userId, amount, "提现转出", hisId);

            // 创建记录
            if (newBalance >= 0) {
                UserWalletWithdrawHisDO userWalletWithdrawHisDO = new UserWalletWithdrawHisDO();
                // 实际金额
                BigDecimal actualAmount = new BigDecimal(amount).divide(new BigDecimal(100), 2, RoundingMode.DOWN);
                // 扣税
                BigDecimal taxAmount = actualAmount.multiply(new BigDecimal(String.valueOf(userTaxRatio))).setScale(2, RoundingMode.HALF_UP);
                actualAmount = actualAmount.subtract(taxAmount).setScale(2, RoundingMode.DOWN);

                userWalletWithdrawHisDO.setHisId(hisId);
                userWalletWithdrawHisDO.setWalletId(WalletProvider.getWalletId(userId, walletType.getValue()));
                userWalletWithdrawHisDO.setAmount(amount);
                userWalletWithdrawHisDO.setApplyStatus(WalletWithdrawHisApplyStatus.WAIT.getValue());
                userWalletWithdrawHisDO.setHisCreateTime(System.currentTimeMillis());
                userWalletWithdrawHisDO.setActualAmount(actualAmount);
                userWalletWithdrawHisDO.setUserTaxRatio(userTaxRatio);
                userWalletWithdrawHisDO.setTaxAmount(taxAmount);
                userWalletWithdrawHisDO.setCardNo(cardNo);
                userWalletWithdrawHisDO.setRefType(RefTypeEnum.USER_WITHDRAW.getValue());

                applyDTO.setActualAmount(actualAmount);
                applyDTO.setNewBalance(Convert.toBigDecimal(newBalance));
                applyDTO.setTaxAmount(taxAmount);

                userWalletWithdrawHisDAO.insert(userWalletWithdrawHisDO);
            }

            return applyDTO;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("提现记录创建失败", e);
        }
        return null;
    }

    /**
     * 获取提现记录列表
     *
     * @param pageIndex
     * @param pageSize
     * @param userId
     * @param walletType
     * @param applyStatus
     * @param yearMonth
     * @return
     */
    @Override
    public PageList<UserWalletWithdrawHisVO> getWithdrawHisList(int pageIndex, int pageSize, Long userId, WalletTypeEnum walletType, Integer applyStatus, Integer yearMonth) {
        TimeSpan ts = getTimeSpanByYearMonth(yearMonth);
        PageList<UserWalletWithdrawHisVO> pageList = new PageList<>(pageIndex, pageSize);
        String walletId = WalletProvider.getWalletId(userId, walletType.getValue());
        pageList.setList(userWalletWithdrawHisDAO.selectAll(pageList.getPageCursor(), pageSize, walletId, applyStatus, ts.getBeginTime(), ts.getEndTime()));
        pageList.setTotal(userWalletWithdrawHisDAO.countAll(walletId, applyStatus, ts.getBeginTime(), ts.getEndTime()));
        return pageList;
    }

    /**
     * 提现申请审核
     *
     * @param hisId
     * @param applyStatus
     * @param remarks
     * @param adminId
     * @return null:处理成功；错误信息
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public String setWithdrawApplyStatus(Long hisId, WalletWithdrawHisApplyStatus applyStatus, String remarks, Long adminId, Integer payType) {
        try {
            Long curTime = System.currentTimeMillis();

            // 审核状态判断
            if (WalletWithdrawHisApplyStatus.REJECT != applyStatus && WalletWithdrawHisApplyStatus.RESOLVE != applyStatus) {
                return "审核状态参数有误";
            }

            // 查询记录
            UserWalletWithdrawHisDO withdrawHis = userWalletWithdrawHisDAO.selectById(hisId);
            if (withdrawHis == null || WalletWithdrawHisApplyStatus.WAIT != WalletWithdrawHisApplyStatus.toEnum(withdrawHis.getApplyStatus())) {
                log.info("{}审核记录不存在，或状态不为待审核", hisId);
                return "审核记录不存在或状态不为待审核";
            }

            // 查询钱包
            UserWalletDO userWallet = walletProvider.getUserWallet(withdrawHis.getWalletId());

            // 更新提现记录
            withdrawHis.setHisApplyTime(curTime);
            withdrawHis.setApplyRemarks(remarks);
            withdrawHis.setApplyAdminId(adminId);
            withdrawHis.setApplyStatus(applyStatus.getValue());
            withdrawHis.setPayType(payType);
            if (userWalletWithdrawHisDAO.updateStatus(withdrawHis, WalletWithdrawHisApplyStatus.WAIT.getValue()) == 0) {
                throw new RuntimeException("提现记录更新失败");
            }

            // 钱包类型
            Integer type = userWallet.getType();
            // 星币提现
            if (WalletTypeEnum.STAR_COIN.getValue() == type) {
                QueueAnchorWithdrawMsg withdrawMsg = new QueueAnchorWithdrawMsg();
                withdrawMsg.setAmount(withdrawHis.getActualAmount());
                withdrawMsg.setUserId(userWallet.getUserId());

                if (WalletWithdrawHisApplyStatus.RESOLVE == applyStatus) {
                    // 审核通过
                    if (withdrawHis.getAgentId() != null) {
                        BigDecimal agentAmount = withdrawHis.getAgentAmount();
                        if (agentAmount != null && agentAmount.compareTo(BigDecimal.ZERO) > 0) {
                            // 给工会增加余额
                            agentWalletProvider.addBalance(withdrawHis.getAgentId(), agentAmount, "主播结算", withdrawHis.getHisId());
                        }
                    }

                    if (withdrawHis.getActualAmount().compareTo(BigDecimal.ZERO) > 0) {
                        // 支付宝转账
                        if (WithdrawPayTypeEnum.ALIPAY_TRANSFER.getValue() == payType) {
                            String error = payService.transferToUser(TransferType.ALIPAY, userWallet.getUserId(), withdrawHis.getActualAmount(), hisId.toString(), "提现通过");
                            if (error != null) {
                                throw new Exception(error);
                            }
                        }
                        withdrawMsg.setSuccess(true);
                    }
                } else {
                    // 审核未通过
                    // 还钱
                    walletProvider.addBalance(WalletTypeEnum.toEnum(userWallet.getType()), userWallet.getUserId(), withdrawHis.getAmount(), "提现审核未通过", withdrawHis.getHisId());
                    withdrawMsg.setSuccess(false);
                    withdrawMsg.setRemark(remarks);
                }
                // 主播提现消息
                rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.ANCHOR_WITHDRAW, withdrawMsg.serialize());
            } else if (WalletTypeEnum.PINK_DIAMOND.getValue() == type) {
                // 粉钻提现（原 星尘提现）
                QueueUserWithdrawMsg withdrawMsg = new QueueUserWithdrawMsg();
                withdrawMsg.setAmount(withdrawHis.getActualAmount());
                withdrawMsg.setUserId(userWallet.getUserId());
                if (WalletWithdrawHisApplyStatus.RESOLVE == applyStatus) {
                    // 审核通过
                    if (withdrawHis.getActualAmount().compareTo(BigDecimal.ZERO) > 0) {
                        // 云账户转账
                        if (WithdrawPayTypeEnum.CLOUD_ACCOUNT_TRANSFER.getValue() == payType) {
                            UserAuthDO userAuth = userService.getUserAuth(userWallet.getUserId());
                            if (userAuth == null) {
                                throw new RuntimeException("没有获取到认证信息");
                            }
                            PaymentOrderDTO orderDTO = cloudAccountService.payToUser(userWallet.getUserId(), userAuth.getRealName(), withdrawHis.getCardNo(), userAuth.getIdCardNo(), Convert.toStr(withdrawHis.getActualAmount()), "审核通过", Convert.toStr(hisId));
                            if (!orderDTO.isSuccess()) {
                                throw new RuntimeException(orderDTO.getMessage());
                            }
                            withdrawMsg.setSuccess(true);
                        } else if (WithdrawPayTypeEnum.EASY_TAX_TRANSFER_1.getValue() == payType) {
                            // 线上易纳税打款[金华云企]
                            UserAuthDO userAuth = userService.getUserAuth(userWallet.getUserId());
                            if (userAuth == null) {
                                throw new RuntimeException("没有获取到认证信息");
                            }
                            PaymentOrderDTO orderDTO = easyTaxService.payToUser(userWallet.getUserId(), userAuth.getRealName(), withdrawHis.getCardNo(), userAuth.getIdCardNo(), Convert.toStr(withdrawHis.getActualAmount()), "审核通过", Convert.toStr(hisId), WithdrawPayTypeEnum.EASY_TAX_TRANSFER_1);
                            if (!orderDTO.isSuccess()) {
                                throw new RuntimeException(orderDTO.getMessage());
                            }
                            withdrawMsg.setSuccess(true);
                        } else if (WithdrawPayTypeEnum.EASY_TAX_TRANSFER_2.getValue() == payType) {
                            // 线上易纳税打款[九江万柳]
                            UserAuthDO userAuth = userService.getUserAuth(userWallet.getUserId());
                            if (userAuth == null) {
                                throw new RuntimeException("没有获取到认证信息");
                            }
                            PaymentOrderDTO orderDTO = easyTaxService.payToUser(userWallet.getUserId(), userAuth.getRealName(), withdrawHis.getCardNo(), userAuth.getIdCardNo(), Convert.toStr(withdrawHis.getActualAmount()), "审核通过", Convert.toStr(hisId), WithdrawPayTypeEnum.EASY_TAX_TRANSFER_2);
                            if (!orderDTO.isSuccess()) {
                                throw new RuntimeException(orderDTO.getMessage());
                            }
                            withdrawMsg.setSuccess(true);
                        } else if (WithdrawPayTypeEnum.EASY_TAX_TRANSFER_3.getValue() == payType) {
                            // 线上易纳税打款[湖南泰美]
                            UserAuthDO userAuth = userService.getUserAuth(userWallet.getUserId());
                            if (userAuth == null) {
                                throw new RuntimeException("没有获取到认证信息");
                            }
                            PaymentOrderDTO orderDTO = easyTaxService.payToUser(userWallet.getUserId(), userAuth.getRealName(), withdrawHis.getCardNo(), userAuth.getIdCardNo(), Convert.toStr(withdrawHis.getActualAmount()), "审核通过", Convert.toStr(hisId), WithdrawPayTypeEnum.EASY_TAX_TRANSFER_3);
                            if (!orderDTO.isSuccess()) {
                                throw new RuntimeException(orderDTO.getMessage());
                            }
                            withdrawMsg.setSuccess(true);
                        } else if (WithdrawPayTypeEnum.EASY_TAX_TRANSFER_4.getValue() == payType) {
                            // 线上易纳税打款[湖北万柳]
                            UserAuthDO userAuth = userService.getUserAuth(userWallet.getUserId());
                            if (userAuth == null) {
                                throw new RuntimeException("没有获取到认证信息");
                            }
                            PaymentOrderDTO orderDTO = easyTaxService.payToUser(userWallet.getUserId(), userAuth.getRealName(), withdrawHis.getCardNo(), userAuth.getIdCardNo(), Convert.toStr(withdrawHis.getActualAmount()), "审核通过", Convert.toStr(hisId), WithdrawPayTypeEnum.EASY_TAX_TRANSFER_4);
                            if (!orderDTO.isSuccess()) {
                                throw new RuntimeException(orderDTO.getMessage());
                            }
                            withdrawMsg.setSuccess(true);
                        } else if (CAIYU_TRANSFER.getValue() == payType) {
                            // 财鱼
                            Long userId = userWallet.getUserId();
                            UserAuthDO userAuth = userService.getUserAuth(userId);
                            if (userAuth == null) {
                                throw new RuntimeException("没有获取到认证信息");
                            }
                            if (thirdPartyService.isNeedSignContractOperation(userId, CAIYU_TRANSFER)) {
                                throw new RuntimeException("需要先执行财鱼签约操作");
                            }

                            String idCard = userAuth.getIdCardNo();
                            String cardNo = withdrawHis.getCardNo();
                            // 单位：元
                            BigDecimal actualAmount = isDev() ? new BigDecimal("1") : withdrawHis.getActualAmount();
                            String phone = userService.getPhone(userId);

                            WithdrawRequestBO wr = new WithdrawRequestBO();
                            wr.setUserWalletWithdrawHisId(hisId);
                            wr.setUserId(userId);
                            wr.setPhone(phone);
                            wr.setAmount(actualAmount);
                            wr.setRealName(userAuth.getRealName());
                            wr.setIdCard(idCard);
                            wr.setPayTypeEnum(CAIYU_TRANSFER);
                            wr.setAccountTypeEnum(BANK_CARD);
                            wr.setReceivingAccount(cardNo);
                            WithdrawRequestRecordBO wrd = thirdPartyService.sendWithdrawRequest(wr);

                            if (!wrd.getIsSuccess()) {
                                throw new RuntimeException("打款失败：" + wrd.getErrMsg());
                            }
                            withdrawMsg.setSuccess(true);
                        } else if (YUNXIAOKE_TRANSFER.getValue() == payType) {
                            // 云销客
                            Long userId = userWallet.getUserId();
                            UserAuthDO userAuth = userService.getUserAuth(userId);
                            if (userAuth == null) {
                                throw new RuntimeException("没有获取到认证信息");
                            }

                            String idCard = userAuth.getIdCardNo();
                            String cardNo = withdrawHis.getCardNo();
                            // 单位：元
                            BigDecimal actualAmount = isDev() ? new BigDecimal("1") : withdrawHis.getActualAmount();
                            String phone = userService.getPhone(userId);

                            WithdrawRequestBO wr = new WithdrawRequestBO();
                            wr.setUserWalletWithdrawHisId(hisId);
                            wr.setUserId(userId);
                            wr.setPhone(phone);
                            wr.setAmount(actualAmount);
                            wr.setRealName(userAuth.getRealName());
                            wr.setIdCard(idCard);
                            wr.setPayTypeEnum(YUNXIAOKE_TRANSFER);
                            wr.setAccountTypeEnum(BANK_CARD);
                            wr.setReceivingAccount(cardNo);
                            WithdrawRequestRecordBO wrd = thirdPartyService.sendWithdrawRequest(wr);

                            if (!wrd.getIsSuccess()) {
                                throw new RuntimeException("打款失败：" + wrd.getErrMsg());
                            }
                            withdrawMsg.setSuccess(true);
                        } else if (SAILTOB_TRANSFER_1.getValue() == payType) {
                            //线上智企行打款;税优地：
                            UserAuthDO userAuth = userService.getUserAuth(userWallet.getUserId());
                            if (userAuth == null) {
                                throw new RuntimeException("没有获取到认证信息");
                            }
                            Long userId = userWallet.getUserId();
                            String idCard = userAuth.getIdCardNo();
                            String cardNo = withdrawHis.getCardNo();
                            // 单位：元
                            BigDecimal actualAmount = isDev() ? new BigDecimal("0.01") : withdrawHis.getActualAmount().setScale(2, RoundingMode.DOWN);
                            String phone = userService.getPhone(userId);

                            Boolean isUserSign = sailtobService.isUserSign(userId, Long.valueOf(payType));
                            if (!isUserSign.booleanValue()) {
                                Boolean signResult = sailtobService.sign(userId, userAuth.getRealName(), phone, cardNo, idCard, Long.valueOf(payType));
                                if (!signResult.booleanValue()) {
                                    throw new RuntimeException("签约信息不成功");
                                }
                                throw new RuntimeException("用户正在签约,请稍后");
                            }

                            List<BankcardVO> bankcardVOList = bankcardDAO.selectByUserId(userId);
                            if (CollectionUtils.isEmpty(bankcardVOList) || bankcardVOList.size() <= 0) {
                                throw new RuntimeException("未找到绑定银行卡");
                            }

                            //银行预留手机号码
                            String bankUserPhone = "";
                            for (BankcardVO cardVO : bankcardVOList) {
                                if (cardNo.equals(cardVO.getCardNo())) {
                                    bankUserPhone = cardVO.getMobile();
                                    break;
                                }
                            }

                            WithdrawRequestBO wr = new WithdrawRequestBO();
                            wr.setUserWalletWithdrawHisId(hisId);
                            wr.setUserId(userId);
                            wr.setPhone(bankUserPhone);
                            wr.setAmount(actualAmount);
                            wr.setRealName(userAuth.getRealName());
                            wr.setIdCard(idCard);
                            wr.setPayTypeEnum(WithdrawPayTypeEnum.toEnum(payType));
                            wr.setAccountTypeEnum(BANK_CARD);
                            wr.setReceivingAccount(cardNo);

                            WithdrawRequestRecordBO wrd = sailtobService.doWithdraw(wr);
                            if (!wrd.getIsSuccess()) {
                                throw new RuntimeException("打款失败：" + wrd.getErrMsg());
                            }

                            withdrawMsg.setSuccess(true);
                        } else if (WithdrawPayTypeEnum.OFFLINE_TRANSFER.getValue() == payType) {
                            log.info("线下打款");
                            withdrawMsg.setSuccess(true);
                        } else {
                            throw new RuntimeException("暂不支持此打款方式");
                        }
                    }
                } else {
                    // 审核未通过
                    // 还钱
                    walletProvider.addBalance(WalletTypeEnum.toEnum(userWallet.getType()), userWallet.getUserId(), withdrawHis.getAmount(), "提现审核未通过", withdrawHis.getHisId());
                    withdrawMsg.setSuccess(false);
                    withdrawMsg.setRemark(remarks);
                }
                // 用户提现消息
                rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.USER_WITHDRAW, withdrawMsg.serialize());
            }

            return null;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("提现申请审核失败", e);
            return e.getMessage();
        }
    }

    /**
     * 增加时间钱包分钟数
     *
     * @param userId
     * @param walletType
     * @param expireTime
     * @param minutes
     * @param hisRemarks
     * @param businessId
     * @return
     */
    @Override
    public boolean addChatTime(Long userId, Integer walletType, Long expireTime, Integer minutes, String hisRemarks, String businessId) {
        return chatTimeWalletProvider.addMinutes(userId, walletType, expireTime, minutes, hisRemarks, businessId);
    }

    /**
     * 获取钱包剩余分钟数
     *
     * @param userId
     * @param walletType
     * @return
     */
    @Override
    public Integer getChatTimeWalletMinutes(Long userId, Integer walletType) {
        return getChatTimeWalletMinutes(userId, walletType, System.currentTimeMillis());
    }

    /**
     * 获取钱包剩余分钟数
     *
     * @param userId
     * @param walletType
     * @param time
     * @return
     */
    @Override
    public Integer getChatTimeWalletMinutes(Long userId, Integer walletType, Long time) {
        return chatTimeWalletProvider.getWalletMinutes(userId, walletType, time);
    }

    /**
     * 获取充值钱包项目
     *
     * @param userId
     * @param walletType
     * @param time
     * @return
     */
    @Override
    public List<ChatTimeWalletItem> getChatTimeWalletItems(Long userId, Integer walletType, Long time) {
        List<ChatTimeWalletItem> list = chatTimeWalletProvider.getChatTimeWalletItems(userId, walletType, time);
        if (CollectionUtil.isNotEmpty(list)) {
            list.sort((o1, o2) -> {
                if (o1.getExpireTime().equals(o2.getExpireTime())) {
                    return 0;
                }

                if (o1.getExpireTime().equals(-1L)) {
                    return 1;
                }

                return o1.getExpireTime() < o2.getExpireTime() ? -1 : 1;
            });
            return list;
        }
        return Collections.emptyList();
    }

    /**
     * 使用道具兑换分钟数
     *
     * @param userId
     * @param bagId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public boolean addChatTimeByProps(Long userId, String bagId) {
        BagItem<PropsVO> bagItem = bagProvider.getProps(bagId, userId);
        if (bagItem == null) {
            return false;
        }

        // 使用道具
        Integer quantity = bagProvider.reduceQuantity(bagId, 1, "使用道具", 0L);
        if (quantity == null || quantity < 0) {
            return false;
        }

        // 增加分钟数
        chatTimeWalletProvider.addMinutes(userId, bagItem.getItem().getPropsExtInt(), System.currentTimeMillis() + bagItem.getItem().getPropsExtLong(), Convert.toInt(bagItem.getItem().getPropsExtDouble()), String.format("使用道具%s兑换", bagItem.getItem().getPropsName()), bagId);

        return true;
    }

    @Override
    public List<BankcardVO> getBankcardList(Long userId) {
        return bankcardDAO.selectByUserId(userId);
    }

    @Override
    public boolean isBankcardBind(Long userId) {
        return bankcardDAO.countByUserId(userId) > 0;
    }

    @Override
    public boolean isCardNoBind(Long userId, String cardNo) {
        return bankcardDAO.countByCardNo(userId, cardNo) > 0;
    }

    @Override
    public boolean isCardIdBind(String cardId) {
        return bankcardDAO.countById(cardId) > 0;
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public boolean unBindCard(String cardId) {
        return bankcardDAO.updateStatus(cardId, CommonStatus.Disabled, System.currentTimeMillis()) > 0;
    }

    @Override
    public Long getUserIdByCardId(String cardId) {
        return bankcardDAO.selectUserIdByCardId(cardId);
    }

    @Override
    public BankcardVO getCard(String cardId) {
        return bankcardDAO.selectCardById(cardId);
    }

    /**
     * 星尘兑换星钻（粉钻-不可提现 转 钻石）
     *
     * @param userId
     * @param dustAmount
     * @param channelId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public Long transDustToDiamond(Long userId, Long dustAmount, Long channelId) {
        Long orderId = identityService.getId();

        if (dustAmount == null || dustAmount < 0) {
            return null;
        }

        // 100 的整数倍
        /*if (dustAmount % 100L > 0) {
            return null;
        }*/

        // 减少星尘
        Long dustBalance = walletProvider.reduceBalance(WalletTypeEnum.STAR_DUST.getValue(), userId, dustAmount, "转出兑换" + nameDiamond, orderId);

        if (dustBalance == null || dustBalance < 0) {
            return dustBalance;
        }

        // 费用计算
        Long fee = BigDecimal.valueOf(dustAmount).multiply(new BigDecimal("" + transDustToDiamondRatio)).setScale(0, RoundingMode.DOWN).longValue();
        Long diamondAmount = dustAmount - fee;

        // 增加星钻
        walletProvider.addBalance(WalletTypeEnum.STAR_DIAMOND.getValue(), userId, diamondAmount, "星尘转入", orderId);

        // 创建记录
        UserWalletTransferHisDO userWalletTransferHis = new UserWalletTransferHisDO();
        userWalletTransferHis.setHisId(orderId);
        userWalletTransferHis.setUserId(userId);
        userWalletTransferHis.setChannelId(channelId);
        userWalletTransferHis.setFromWalletType(WalletTypeEnum.STAR_DUST.getValue());
        userWalletTransferHis.setDestWalletType(WalletTypeEnum.STAR_DIAMOND.getValue());
        userWalletTransferHis.setTransAmount(dustAmount);
        userWalletTransferHis.setTransFee(fee);
        userWalletTransferHis.setDestAmount(diamondAmount);
        userWalletTransferHis.setHisCreateTime(System.currentTimeMillis());
        userWalletTransferHisDAO.insert(userWalletTransferHis);

        // 数据统计
        statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.STAR_DUST_TO_DIAMOND.getValue(), BigDecimal.valueOf(dustAmount));
        statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.STAR_DIAMOND_FROM_DUST.getValue(), BigDecimal.valueOf(diamondAmount).multiply((BigDecimal.ONE.subtract(new BigDecimal("" + transDustToDiamondRatio)))));

        return dustBalance;
    }

    /**
     * 查询转换记录
     *
     * @param lastId
     * @param pageSize
     * @param userId
     * @param fromWalletType
     * @param destWalletType
     * @return
     */
    @Override
    public RecordList<UserWalletTransferHisDO, Long> getWalletTransferHisList(Long lastId, int pageSize, Long userId, int fromWalletType, int destWalletType, Integer yearMonth) {
        lastId = Optional.ofNullable(lastId).orElse(0L);
        if (lastId < 0) {
            lastId = 0L;
        }

        RecordList<UserWalletTransferHisDO, Long> recordList = new RecordList<>(pageSize);
        // 处理时间
        TimeSpan ts = getTimeSpanByYearMonth(yearMonth);
        List<UserWalletTransferHisDO> transferHisList = userWalletTransferHisDAO.selectAll(lastId, pageSize, userId, fromWalletType, destWalletType, ts.getBeginTime(), ts.getEndTime());

        if (CollectionUtil.isNotEmpty(transferHisList)) {
            lastId = transferHisList.get(transferHisList.size() - 1).getHisId();
        }

        recordList.setList(transferHisList);
        recordList.setLastId(lastId);
        return recordList;
    }

    /**
     * 星钻兑换游戏币
     *
     * @param userId
     * @param diamondAmount
     * @param channelId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public Long transDiamondToGameCoin(Long userId, Long diamondAmount, Long channelId) {
        Long orderId = identityService.getId();

        // 减少星钻
        Long balance = walletProvider.reduceBalance(WalletTypeEnum.STAR_DIAMOND, userId, diamondAmount, "兑换游戏币", orderId);
        if (balance == null || balance < 0) {
            return balance;
        }

        // 增加游戏币
        Long gameCoin = GameUtil.transDiamond2GameCoin(diamondAmount);
        walletProvider.addBalance(WalletTypeEnum.GAME_COIN, userId, gameCoin, String.format("从%s兑换", nameDiamond), orderId);

        // 创建记录
        UserWalletTransferHisDO userWalletTransferHis = new UserWalletTransferHisDO();
        userWalletTransferHis.setHisId(orderId);
        userWalletTransferHis.setUserId(userId);
        userWalletTransferHis.setChannelId(channelId);
        userWalletTransferHis.setFromWalletType(WalletTypeEnum.STAR_DIAMOND.getValue());
        userWalletTransferHis.setDestWalletType(WalletTypeEnum.GAME_COIN.getValue());
        userWalletTransferHis.setTransAmount(diamondAmount);
        userWalletTransferHis.setTransFee(0L);
        userWalletTransferHis.setDestAmount(gameCoin);
        userWalletTransferHis.setHisCreateTime(System.currentTimeMillis());
        userWalletTransferHisDAO.insert(userWalletTransferHis);

        return balance;
    }

    /**
     * 游戏币兑换星钻
     *
     * @param userId
     * @param gameCoinAmount
     * @param channelId
     * @return
     */
    @Override
    public Long transGameCoinToDiamond(Long userId, Long gameCoinAmount, Long channelId) {
        Long orderId = identityService.getId();
        // 减少游戏币
        Long balance = walletProvider.reduceBalance(WalletTypeEnum.GAME_COIN, userId, gameCoinAmount, "兑换" + nameDiamond, orderId);
        if (balance == null || balance < 0) {
            log.warn("游戏币余额异常 balance={} gameCoinAmount={}, userId={}", balance, gameCoinAmount, userId);
            return balance;
        }

        // 增加星钻
        Long diamondAmount = GameUtil.transGameCoin2Diamond(gameCoinAmount);
        walletProvider.addBalance(WalletTypeEnum.STAR_DIAMOND, userId, diamondAmount, "从游戏币兑换", orderId);

        // 创建记录
        UserWalletTransferHisDO userWalletTransferHis = new UserWalletTransferHisDO();
        userWalletTransferHis.setHisId(orderId);
        userWalletTransferHis.setUserId(userId);
        userWalletTransferHis.setChannelId(channelId);
        userWalletTransferHis.setFromWalletType(WalletTypeEnum.GAME_COIN.getValue());
        userWalletTransferHis.setDestWalletType(WalletTypeEnum.STAR_DIAMOND.getValue());
        userWalletTransferHis.setTransAmount(gameCoinAmount);
        userWalletTransferHis.setTransFee(0L);
        userWalletTransferHis.setDestAmount(diamondAmount);
        userWalletTransferHis.setHisCreateTime(System.currentTimeMillis());
        userWalletTransferHisDAO.insert(userWalletTransferHis);

        return balance;
    }

    /**
     * 昨天的星尘转粉钻
     *
     * @param userId
     * @param today  今天（格式 2020-07-30）
     * @param ytd    昨天（格式 2020-07-30）
     * @return false 不用转换
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public boolean transYtdStarDustToPinkDiamond(Long userId, String today, String ytd) {
        // 昨天星尘总收入
        Long startDustTotal = userWalletHisDAO.sumDailyBalance(userId, ytd, WalletTypeEnum.STAR_DUST.getValue());
        startDustTotal = startDustTotal == null ? 0 : startDustTotal;
        if (startDustTotal < 1) {
            return false;
        }

        // 今天（截止当前时间）消耗的星尘总数
        Long startDustCostTotal = userWalletHisDAO.sumDailyCost(userId, today, WalletTypeEnum.STAR_DUST.getValue());
        startDustCostTotal = startDustCostTotal == null ? 0 : startDustCostTotal;

        // 转换星尘总数
        Long transTotal = startDustTotal + startDustCostTotal;
        if (transTotal < 1) {
            return false;
        }

        Long channelId = userService.getChannelId(userId);
        if (channelId == null) {
            log.warn("用户没有ChannelId，不能进行星尘转粉钻，userId={}", userId);
            return false;
        }

        // 幂等
        Integer transDate = Integer.parseInt(ytd.replaceAll("-", ""));
        if (walletTransDustToPinkHisProvider.exist(transDate, userId)) {
            return false;
        }
        // 星尘兑换粉钻
        Long balance = transDustToPinkDiamond(userId, transTotal, channelId);
        WalletTransDustToPinkHisDO his = new WalletTransDustToPinkHisDO();
        his.setTransDate(transDate);
        his.setUserId(userId);
        his.setTransAmount(transTotal);
        his.setCreateTime(System.currentTimeMillis());
        his = walletTransDustToPinkHisProvider.create(his);
        if (his == null) {
            log.error("星尘兑换粉钻失败，保存星尘转粉钻记录失败。userId={} transTotal={}", userId, transTotal);
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            return false;
        }
        return balance != null && balance >= 0 ? true : false;
    }

    /**
     * 昨天的星尘转粉钻
     *
     * @return
     */
    @Override
    public String transYtdStarDustToPinkDiamond() {
        log.info("星尘转粉钻开始");
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");
        DateTime today = new DateTime();
        DateTime ytd = DateUtil.offsetDay(today, -1);
        String todayStr = sdf.format(today);
        String ytdStr = sdf.format(ytd);

        int pageNum = 1;
        int pageSize = 1000;
        // 用户总数
        int userTotal = 0;
        // 星尘转粉钻总用户数
        int transTotal = 0;
        while (true) {
            List<Long> userIds = userService.getUserIds(pageNum++, pageSize);
            if (CollectionUtils.isEmpty(userIds)) {
                break;
            }
            for (Long userId : userIds) {
                if (userId == null) {
                    continue;
                }
                ++userTotal;
                boolean flag = transYtdStarDustToPinkDiamond(userId, todayStr, ytdStr);
                if (flag) {
                    ++transTotal;
                }
            }
        }

        String result = StrUtil.format("执行日期 {} 结算日期 {} 系统用户总数 {} 星尘转粉钻用户总数 {}",
                todayStr, ytdStr, userTotal, transTotal);
        log.info("星尘转粉钻结束 {}", result);
        return result;
    }

    /**
     * 粉钻兑换星钻（粉钻-可提现 转 钻石）
     *
     * @param userId
     * @param pinkAmount
     * @param channelId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public Long transPinkDiamondToDiamond(Long userId, Long pinkAmount, Long channelId) {
        Long orderId = identityService.getId();

        if (pinkAmount == null || pinkAmount < 0) {
            return null;
        }

        // 100 的整数倍 api-app/v1/wallet/transDustPinkToDiamond 可以打破这个规则
        /*if (pinkAmount % 100L > 0) {
            return null;
        }*/

        // 减少粉钻
        Long pinkBalance = walletProvider.reduceBalance(WalletTypeEnum.PINK_DIAMOND.getValue(), userId, pinkAmount, "转出兑换" + nameDiamond, orderId);

        if (pinkBalance == null || pinkBalance < 0) {
            return pinkBalance;
        }

        // 费用计算
        Long fee = BigDecimal.valueOf(pinkAmount).multiply(new BigDecimal("" + transDustToDiamondRatio)).setScale(0, RoundingMode.DOWN).longValue();
        Long diamondAmount = pinkAmount - fee;

        // 增加星钻
        walletProvider.addBalance(WalletTypeEnum.STAR_DIAMOND.getValue(), userId, diamondAmount, "粉钻转入", orderId);

        // 创建记录
        UserWalletTransferHisDO userWalletTransferHis = new UserWalletTransferHisDO();
        userWalletTransferHis.setHisId(orderId);
        userWalletTransferHis.setUserId(userId);
        userWalletTransferHis.setChannelId(channelId);
        userWalletTransferHis.setFromWalletType(WalletTypeEnum.PINK_DIAMOND.getValue());
        userWalletTransferHis.setDestWalletType(WalletTypeEnum.STAR_DIAMOND.getValue());
        userWalletTransferHis.setTransAmount(pinkAmount);
        userWalletTransferHis.setTransFee(fee);
        userWalletTransferHis.setDestAmount(diamondAmount);
        userWalletTransferHis.setHisCreateTime(System.currentTimeMillis());
        userWalletTransferHisDAO.insert(userWalletTransferHis);

        // 数据统计
        statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.PINK_DIAMOND_TO_DIAMOND.getValue(), BigDecimal.valueOf(pinkAmount));
        statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.STAR_DIAMOND_FROM_PINK.getValue(), BigDecimal.valueOf(diamondAmount));

        return pinkBalance;
    }

    /**
     * 兑换钻石（粉钻-不可提现、粉钻-可提现 兑换 星钻）
     *
     * @param userId
     * @param total
     * @param channelId
     * @return
     */
    @Override
    @Transactional(rollbackFor = Exception.class)
    public TransDustPinkToDiamondVO transDustPinkToDiamond(Long userId, Long total, Long channelId) {
        UserWalletVO wallet = getUserWallet(userId);
        Long dustBalance = getBalance(wallet, WalletTypeEnum.STAR_DUST);
        Long pinkBalance = getBalance(wallet, WalletTypeEnum.PINK_DIAMOND);
        TransDustPinkToDiamondVO vo = new TransDustPinkToDiamondVO();

        if (dustBalance + pinkBalance - total < 0) {
            vo.setErrMsg("余额不足");
            return vo;
        }

        if (dustBalance >= total) {
            dustBalance = total;
            pinkBalance = 0L;
        } else {
            // 粉钻-不可提现 转 钻石
            if (dustBalance % 100 > 0) {
                dustBalance = 0L;
            }
            pinkBalance = total - dustBalance;
        }

        if (dustBalance > 0) {
            dustBalance = transDustToDiamond(userId, dustBalance, channelId);
        }
        if (pinkBalance > 0) {
            if (pinkBalance % 100 > 0) {
                vo.setErrMsg("粉钻-可提现数量必须是100的整数倍");
                return vo;
            }
            // 粉钻-可提现 转 钻石
            pinkBalance = transPinkDiamondToDiamond(userId, pinkBalance, channelId);
        }

        if (dustBalance == null || pinkBalance == null ||
                dustBalance < 0 || pinkBalance < 0) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.warn("兑换钻石失败，数据回滚 userId={} total={}, channelId={} dustBalance={} pinkBalance={}",
                    userId, total, channelId, dustBalance, pinkBalance);
            vo.setErrMsg("兑换失败");
            return vo;
        }

        vo.setDustBalance(dustBalance);
        vo.setPinkBalance(pinkBalance);
        return vo;
    }

    /**
     * 获取用户粉钻流水
     *
     * @param userId
     * @return
     */
    @Override
    public List<UserWalletHisDO> getUserWalletHisByOrderIds(List<Long> orderIds, Long userId, Integer walletType) {
        return userWalletHisDAO.selectUserWalletHisByOrderIds(orderIds, userId, walletType);
    }

    /**
     * 钱包名称Map
     *
     * @return 钱包类型:钱包类型显示的名称
     */
    @Override
    public Map<WalletTypeEnum, String> getWalletNameMap() {
        Map<WalletTypeEnum, String> map = new HashMap<>();
        for (WalletTypeEnum walletTypeEnum : WalletTypeEnum.values()) {
            String value = walletTypeEnum == WalletTypeEnum.STAR_DIAMOND ? nameDiamond : walletTypeEnum.getName();
            map.put(walletTypeEnum, value);
        }
        return map;
    }

    /**
     * 星尘兑换粉钻
     *
     * @param userId
     * @param dustAmount
     * @param channelId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public Long transDustToPinkDiamond(Long userId, Long dustAmount, Long channelId) {
        Long orderId = identityService.getId();

        if (dustAmount == null || dustAmount < 0) {
            return null;
        }

        // 减少星尘
        Long dustBalance = walletProvider.reduceBalance(WalletTypeEnum.STAR_DUST.getValue(), userId, dustAmount, "转出", orderId);

        if (dustBalance == null || dustBalance < 0) {
            return dustBalance;
        }

        // 费用计算
        Long fee = 0L;
        Long pinkAmount = dustAmount - fee;

        // 增加粉钻
        walletProvider.addBalance(WalletTypeEnum.PINK_DIAMOND.getValue(), userId, pinkAmount, "星尘转入", orderId);

        // 创建记录
        UserWalletTransferHisDO userWalletTransferHis = new UserWalletTransferHisDO();
        userWalletTransferHis.setHisId(orderId);
        userWalletTransferHis.setUserId(userId);
        userWalletTransferHis.setChannelId(channelId);
        userWalletTransferHis.setFromWalletType(WalletTypeEnum.STAR_DUST.getValue());
        userWalletTransferHis.setDestWalletType(WalletTypeEnum.PINK_DIAMOND.getValue());
        userWalletTransferHis.setTransAmount(dustAmount);
        userWalletTransferHis.setTransFee(fee);
        userWalletTransferHis.setDestAmount(pinkAmount);
        userWalletTransferHis.setHisCreateTime(System.currentTimeMillis());
        userWalletTransferHisDAO.insert(userWalletTransferHis);

        // 数据统计
        statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.STAR_DUST_TO_PINK_DIAMOND.getValue(), BigDecimal.valueOf(dustAmount));
        statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.PINK_DIAMOND_FROM_DUST.getValue(), BigDecimal.valueOf(pinkAmount));

        return dustBalance;
    }

    private static TimeSpan getTimeSpanByYearMonth(Integer yearMonth) {
        TimeSpan ts = new TimeSpan();
        if (yearMonth != null) {
            DateTime dateTime = DateUtil.parse(String.valueOf(yearMonth), "yyyyMM");
            dateTime.setField(DateField.DAY_OF_MONTH, 1);
            dateTime.setField(DateField.HOUR_OF_DAY, 0);
            dateTime.setField(DateField.MINUTE, 0);
            dateTime.setField(DateField.SECOND, 0);
            dateTime.setField(DateField.MILLISECOND, 0);
            ts.setBeginTime(dateTime.getTime());
            dateTime.offset(DateField.MONTH, 1);
            dateTime.offset(DateField.MILLISECOND, -1);
            ts.setEndTime(dateTime.getTime());
        }
        return ts;
    }

    /**
     * 获得钱包余额
     *
     * @param walletVO
     * @param typeEnum
     * @return
     */
    private Long getBalance(UserWalletVO walletVO, WalletTypeEnum typeEnum) {
        for (UserWalletVO.Item item : walletVO.getWalletItems()) {
            if (typeEnum.getValue() - item.getWalletType() == 0) {
                return item.getBalance();
            }
        }
        return 0L;
    }

}
