package com.starbuds.server.service.finance.impl.gift;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.finance.BagItemTypeEnum;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.pojo.db.SendGiftSnapshotDO;
import com.starbuds.server.common.pojo.finance.gift.*;
import com.starbuds.server.common.pojo.mq.QueueSendGiftSnapShotMsg;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.finance.SendGiftServiceFromBagService;
import com.starbuds.server.common.service.finance.SendGiftSnapshotMsgService;
import com.starbuds.server.common.utils.LogUtil;
import com.starbuds.server.service.finance.provider.GiftProvider;
import com.starbuds.server.service.finance.provider.IBagProvider;
import com.starbuds.server.service.finance.provider.SendGiftSnapshotProvider;
import com.starbuds.server.service.finance.provider.WalletProvider;
import com.starbuds.server.service.finance.provider.gift.GiftExecute;
import com.starbuds.server.service.finance.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.beans.BeanUtils;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.List;

import static com.starbuds.server.service.finance.util.SpringContextUtil.isDev;

/**
 * 从背包赠送礼物
 *
 * @author aoe
 * @date 2021/11/16
 */
@Service(version = "1.0.0")
@Slf4j
public class SendGiftServiceFromBagImpl extends SendGiftAbstract implements SendGiftServiceFromBagService {

    @Resource
    private GiftProvider giftProvider;

    @Reference(version = "1.0.0")
    private IdentityService identityService;

    @Resource
    private SendGiftSnapshotProvider sendGiftSnapshotProvider;

    @Resource
    private IBagProvider bagProvider;

    @Resource
    private WalletProvider walletProvider;

    @Reference(version = "1.0.0")
    SendGiftSnapshotMsgService sendGiftSnapshotMsgService;

    @Override
    @Transactional(rollbackFor = Exception.class)
    public SendGiftOrder sendGift(SendGiftByFinanceAbstract bo, GiftExecute giftExecute) {
        GiftRate giftRate = giftExecute.giftRate();
        Long orderId = bo.getOrderId();
        final Integer quantity = bo.getQuantity();
        List<SendGiftResult> results = new ArrayList<>(bo.getReceiverUsers().size());

        Long senderUserId = bo.getSenderUserId();
        String senderUserName = bo.getSenderUserName();
        Long giftId = bo.getGiftId();
        String giftName = bo.getGiftName();
        Integer bagRemainder = null;
        List receiverUserIds = new ArrayList<>(bo.getReceiverUsers().size());
        List receiverUserNames = new ArrayList<>(bo.getReceiverUsers().size());
        Long totalPay = 0L;
        for (GiftReceiver receiver : bo.getReceiverUsers()) {
            Long receiverUserId = receiver.getUserId();
            String receiverUserName = receiver.getUserName();

            // 减少礼物发送者的背包中的物品数量
            bagRemainder = bagProvider.reduceQuantity(senderUserId, giftId, BagItemTypeEnum.GIFT.getValue(),
                    quantity,
                    String.format("赠送给(%s)[%d]%d个%s", receiverUserName, receiverUserId, quantity, giftName),
                    orderId);
            if (bagRemainder == null || bagRemainder < 0) {
                // 回滚
                TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
                // 数量不足时终止处理
                return null;
            }
            receiverUserIds.add(receiverUserId);
            receiverUserNames.add(receiver.getUserName());
            Long unitPrice = bo.getGiftPrice() * quantity;
            totalPay += unitPrice;

            // 计算礼物收益比率 💪 封装变化
            BigDecimal rate = giftRate.calculate(bo, receiver);
            // 单个用户获得的礼物价值
            // 接收者收到的金额
            Long receiveAmount = BigDecimal.valueOf(unitPrice).multiply(rate).setScale(0, RoundingMode.DOWN).longValue();
            // 增加礼物接收者的粉钻
            WalletTypeEnum walletType = WalletTypeEnum.STAR_DUST;
            walletProvider.addBalance(walletType, receiverUserId, receiveAmount,
                    String.format("[%s](%d)赠送%d个%s", senderUserName, senderUserId, quantity, giftName),
                    orderId);

            SendGiftResult result = new SendGiftResult();
            result.setRate(rate);
            result.setUnitPrice(unitPrice);
            result.setReceiveAmount(receiveAmount);
            result.setWalletType(walletType);
            result.setBagRemainder(bagRemainder);
            result.setReceiverUserId(receiverUserId);
            result.setReceiverUserName(receiverUserName);
            result.setSenderUserId(senderUserId);
            result.setSenderUserName(senderUserName);
            result.setGiftId(giftId);
            result.setGiftName(giftName);
            result.setQuantity(quantity);
            results.add(result);
        }

        if (CollectionUtil.isNotEmpty(results)) {
            giftExecute.doAfterSendByBag(bo, results);
            // 发送消息
            giftExecute.sendMsg(bo, results);
        }

        SendGiftSnapshotDO sendGiftSnapshot = new SendGiftSnapshotDO();
        sendGiftSnapshot.setOrderId(orderId);
        sendGiftSnapshot.setSenderUserId(senderUserId);
        sendGiftSnapshot.setSenderUserName(senderUserName);
        sendGiftSnapshot.setReceiverUserIds(StringUtils.join(receiverUserIds, ","));
        sendGiftSnapshot.setReceiverUserNames(StringUtils.join(receiverUserNames, ","));
        sendGiftSnapshot.setGiftId(giftId);
        sendGiftSnapshot.setGiftName(giftName);
        sendGiftSnapshot.setGiftType(bo.getGiftType());
        sendGiftSnapshot.setGiftPrice(bo.getGiftPrice());
        sendGiftSnapshot.setGiftPriceType(bo.getGiftPriceType());
        sendGiftSnapshot.setQuantity(quantity);
        sendGiftSnapshot.setIsFromBag(BoolIntVal.True);
        sendGiftSnapshot.setTotalPay(totalPay);
        if (bo instanceof SendGiftRtcUnderMaiBO) {
            SendGiftRtcUnderMaiBO rtcBo = (SendGiftRtcUnderMaiBO) bo;
            sendGiftSnapshot.setRoomId(rtcBo.getRoomId());
            sendGiftSnapshot.setRoomFamilyId(rtcBo.getRoomFamilyId());
        }
        SendGiftSnapshotDO snapshotDO = sendGiftSnapshotProvider.create(sendGiftSnapshot);

        sendGiftSnapshotMsgService.sendMQMsg(snapshotDO);

        SendGiftOrder order = new SendGiftOrder();
        order.setOrderId(orderId);
        order.setBalance(Convert.toLong(bagRemainder));
        return order;
    }

    @Override
    protected GiftProvider giftProvider() {
        return giftProvider;
    }

    @Override
    protected IdentityService identityService() {
        return identityService;
    }
}
