package com.starbuds.server.service.finance.impl.gift;

import cn.hutool.core.collection.CollectionUtil;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.media.GiftTypeEnum;
import com.starbuds.server.common.pojo.db.SendGiftSnapshotDO;
import com.starbuds.server.common.pojo.finance.gift.*;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.finance.SendGiftServicePayDiamondService;
import com.starbuds.server.common.service.finance.SendGiftSnapshotMsgService;
import com.starbuds.server.service.finance.provider.GiftProvider;
import com.starbuds.server.service.finance.provider.SendGiftSnapshotProvider;
import com.starbuds.server.service.finance.provider.WalletProvider;
import com.starbuds.server.service.finance.provider.gift.GiftExecute;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.List;

/**
 * 消耗钻石赠送礼物
 *
 * @author aoe
 * @date 2021/11/16
 */
@Service(version = "1.0.0")
@Slf4j
public class SendGiftServicePayDiamondImpl extends SendGiftAbstract implements SendGiftServicePayDiamondService {

    @Resource
    private GiftProvider giftProvider;

    @Reference(version = "1.0.0")
    private IdentityService identityService;

    @Reference(version = "1.0.0")
    SendGiftSnapshotMsgService sendGiftSnapshotMsgService;

    @Resource
    private SendGiftSnapshotProvider sendGiftSnapshotProvider;

    @Resource
    private WalletProvider walletProvider;

    @Override
    @Transactional(rollbackFor = Exception.class)
    public SendGiftOrder sendGift(SendGiftByFinanceAbstract bo, GiftExecute giftExecute) {
        GiftRate giftRate = giftExecute.giftRate();
        boolean isLotteryGift = GiftTypeEnum.LOTTERY.getValue() == bo.getGiftType();
        WalletTypeEnum walletType = WalletTypeEnum.STAR_DUST;
        Long orderId = bo.getOrderId();
        final Integer quantity = bo.getQuantity();
        List<SendGiftResult> results = new ArrayList<>(bo.getReceiverUsers().size());

        Long balance = null;
        Long senderUserId = bo.getSenderUserId();
        String senderUserName = bo.getSenderUserName();
        Long giftId = bo.getGiftId();
        String giftName = bo.getGiftName();
        List receiverUserIds = new ArrayList<>(bo.getReceiverUsers().size());
        List receiverUserNames = new ArrayList<>(bo.getReceiverUsers().size());
        Long totalPay = 0L;
        for (GiftReceiver receiver : bo.getReceiverUsers()) {
            Long receiverUserId = receiver.getUserId();
            String receiverUserName = receiver.getUserName();

            // 计算礼物收益比率 💪 封装变化
            BigDecimal rate = giftRate.calculate(bo, receiver);
            // 单个用户获得的礼物价值
            Long unitPrice = bo.getGiftPrice() * quantity;
            // 接收者收到的金额
            Long receiveAmount = BigDecimal.valueOf(unitPrice).multiply(rate).setScale(0, RoundingMode.DOWN).longValue();

            // 送礼物事件名称
            String sendAction = "赠送";
            if (bo instanceof SendGiftRewardFeedBO) {
                sendAction = "打赏";
            }

            // 尝试扣费
            balance = walletProvider.reduceBalance(bo.getGiftPriceType(), senderUserId, -unitPrice,
                    String.format("%s给[%s](%d)%d个%s", sendAction, receiver.getUserName(), receiverUserId,
                            quantity, bo.getGiftName()),
                    orderId);
            if (balance == null || balance < 0) {
                // 回滚
                TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
                return null;
            }
            receiverUserIds.add(receiverUserId);
            receiverUserNames.add(receiver.getUserName());
            totalPay += unitPrice;

            if (!isLotteryGift) {
                // 用户结算
                walletProvider.addBalance(walletType, receiverUserId, receiveAmount,
                        String.format("[%s](%d)%s%d个%s", bo.getSenderUserName(), senderUserId, sendAction,
                                quantity, bo.getGiftName()),
                        orderId);
            }

            SendGiftResult result = new SendGiftResult();
            result.setRate(rate);
            result.setUnitPrice(unitPrice);
            result.setReceiveAmount(receiveAmount);
            result.setWalletType(walletType);
            result.setReceiverUserId(receiverUserId);
            result.setReceiverUserName(receiverUserName);
            result.setSenderUserId(senderUserId);
            result.setSenderUserName(senderUserName);
            result.setGiftId(giftId);
            result.setGiftName(giftName);
            result.setQuantity(quantity);
            result.setBalance(balance);
            results.add(result);
        }

        if (CollectionUtil.isNotEmpty(results)) {
            // 家族收益
            giftExecute.doAfterSendByDiamond(bo, results);
            // 发送消息
            giftExecute.sendMsg(bo, results);
        }

        SendGiftSnapshotDO sendGiftSnapshot = new SendGiftSnapshotDO();
        sendGiftSnapshot.setOrderId(orderId);
        sendGiftSnapshot.setSenderUserId(senderUserId);
        sendGiftSnapshot.setSenderUserName(senderUserName);
        sendGiftSnapshot.setReceiverUserIds(StringUtils.join(receiverUserIds, ","));
        sendGiftSnapshot.setReceiverUserNames(StringUtils.join(receiverUserNames, ","));
        sendGiftSnapshot.setGiftId(giftId);
        sendGiftSnapshot.setGiftName(giftName);
        sendGiftSnapshot.setGiftType(bo.getGiftType());
        sendGiftSnapshot.setGiftPrice(bo.getGiftPrice());
        sendGiftSnapshot.setGiftPriceType(bo.getGiftPriceType());
        sendGiftSnapshot.setQuantity(quantity);
        sendGiftSnapshot.setIsFromBag(BoolIntVal.False);
        sendGiftSnapshot.setTotalPay(totalPay);
        if (bo instanceof SendGiftRtcUnderMaiBO) {
            SendGiftRtcUnderMaiBO rtcBo = (SendGiftRtcUnderMaiBO) bo;
            sendGiftSnapshot.setRoomId(rtcBo.getRoomId());
            sendGiftSnapshot.setRoomFamilyId(rtcBo.getRoomFamilyId());
        }

        SendGiftSnapshotDO snapshotDO = sendGiftSnapshotProvider.create(sendGiftSnapshot);

        sendGiftSnapshotMsgService.sendMQMsg(snapshotDO);

        SendGiftOrder order = new SendGiftOrder();
        order.setOrderId(orderId);
        order.setBalance(balance);
        order.setWalletType(walletType);
        return order;
    }

    @Override
    protected GiftProvider giftProvider() {
        return giftProvider;
    }

    @Override
    protected IdentityService identityService() {
        return identityService;
    }
}
