package com.starbuds.server.service.finance.impl.lottery;

import cn.hutool.core.lang.WeightRandom;
import cn.hutool.core.util.RandomUtil;
import com.alibaba.fastjson.JSON;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.starbuds.server.common.constant.activity.LotteryGameCostTypeEnum;
import com.starbuds.server.common.constant.cache.RedisActivityKeys;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.pojo.activity.params.LotteryGameCostItemParam;
import com.starbuds.server.common.pojo.db.LotteryGameUserJoinHisDO;
import com.starbuds.server.common.pojo.db.LotteryGameUserQualificationHisDO;
import com.starbuds.server.common.pojo.finance.params.SaveLotteryGameQualificationParam;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.finance.SaveLotteryGameQualificationService;
import com.starbuds.server.service.finance.provider.lottery.LotteryGameStatisticsProvider;
import com.starbuds.server.service.finance.provider.lottery.LotteryGameUserJoinHisProvider;
import com.starbuds.server.service.finance.provider.lottery.cost.LotteryGameCostItem;
import com.starbuds.server.service.finance.provider.lottery.cost.LotteryGameUserQualificationHisProvider;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.math.NumberUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static com.starbuds.server.service.finance.util.SpringContextUtil.isDev;
import static java.lang.Boolean.FALSE;
import static java.lang.Boolean.TRUE;

/**
 * @author aoe
 * @date 2021/12/7
 */
@Service(version = "1.0.0")
@Slf4j
public class SaveLotteryGameQualificationServiceImpl implements SaveLotteryGameQualificationService {

    @Resource(name = "lotteryGameCostNone")
    private LotteryGameCostItem lotteryGameCostNone;

    @Resource(name = "lotteryGameCostCoin")
    private LotteryGameCostItem lotteryGameCostCoin;

    @Resource(name = "lotteryGameCostProps")
    private LotteryGameCostItem lotteryGameCostProps;

    private Map<LotteryGameCostTypeEnum, LotteryGameCostItem> costItemMap;

    @Resource
    private LotteryGameUserQualificationHisProvider lotteryGameUserQualificationHisProvider;

    @Reference(version = "1.0.0")
    private RedisService redisService;

    @Reference(version = "1.0.0")
    private IdentityService identityService;

    @Resource
    private LotteryGameUserJoinHisProvider lotteryGameUserJoinHisProvider;

    @Resource
    private LotteryGameStatisticsProvider lotteryGameStatisticsProvider;

    @PostConstruct
    private void init() {
        costItemMap = ImmutableMap.of(LotteryGameCostTypeEnum.NONE, lotteryGameCostNone,
                LotteryGameCostTypeEnum.COIN, lotteryGameCostCoin,
                LotteryGameCostTypeEnum.PROPS, lotteryGameCostProps);
    }

    /**
     * 保存抽奖资格
     * 1. 存储用户抽奖资格
     * 2. 保存用户参与记录
     * 3. 扣除消耗物品
     * 4. 扣除奖池库存
     *
     * @param param
     * @return null: 失败
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public List<LotteryGameUserQualificationHisDO> costItemAndSaveQualification(SaveLotteryGameQualificationParam param) {
        // 消耗物品
        LotteryGameCostItemParam costItemParam = costItem(param);

        Long snapId = param.getSnapId();
        Long userId = param.getUserId();
        Integer reduceStockQuantity = param.reduceStockQuantity();
        List<LotteryGameUserQualificationHisDO> qualifications = IntStream.range(0, reduceStockQuantity)
                .mapToObj(i -> {
                    LotteryGameUserQualificationHisDO his = getQualificationHis(snapId, userId, param.getClickRate());
                    checkNotNull(his);
                    return his;
                })
                .collect(Collectors.toList());
        lotteryGameUserQualificationHisProvider.createBatch(qualifications);

        // 保存用户参与记录
        saveUserJoinHis(userId, snapId, param.getRoomId(), qualifications.size());

        // 扣减库存
        reduceRewardCount(snapId, reduceStockQuantity);
        log.info("保存抽奖资格成功 param={} his={}", param, qualifications);
        lotteryGameStatisticsProvider.sendCostForkMsg(param.getCostType(), costItemParam);
        return qualifications;
    }

    /**
     * 保存用户参与记录
     */
    @Transactional(rollbackFor = Exception.class)
    public void saveUserJoinHis(Long userId, Long snapId, Long roomId, int count) {
        if (count < 1) {
            return;
        }
        long current = System.currentTimeMillis();
        List<LotteryGameUserJoinHisDO> hisList = IntStream.range(0, count)
                .mapToObj(i -> {
                    LotteryGameUserJoinHisDO his = new LotteryGameUserJoinHisDO();
                    his.setUserId(userId);
                    his.setSnapId(snapId);
                    his.setRoomId(roomId);
                    his.setCreateTime(current);
                    return his;
                })
                .collect(Collectors.toList());

        boolean batchInsert = lotteryGameUserJoinHisProvider.createBatch(hisList);
        checkState(batchInsert, "保存抽奖资格失败，保存用户参与记录失败 his=" + JSON.toJSONString(hisList));
    }

    /**
     * 消耗物品
     *
     * @param param
     */
    @Transactional(rollbackFor = Exception.class)
    public LotteryGameCostItemParam costItem(SaveLotteryGameQualificationParam param) {
        LotteryGameCostItem lotteryGameCostItem = costItemMap.get(param.getCostType());
        LotteryGameCostItemParam costItemParam = getCostItem(identityService.getId(), param);
        boolean costResult = lotteryGameCostItem.costOneItem(costItemParam);
        Preconditions.checkState(costResult,
                String.format("保存抽奖资格失败，消耗物品失败 lotteryGameCostItem=%s param=%s",
                        lotteryGameCostItem, param));
        return costItemParam;
    }

    private LotteryGameUserQualificationHisDO getQualificationHis(Long snapId, Long userId, Double clickRate) {
        LotteryGameUserQualificationHisDO po = new LotteryGameUserQualificationHisDO();
        po.setSnapId(snapId);
        po.setUserId(userId);
        po.setIsEffective(effectiveClick(clickRate));
        po.setCreateTime(System.currentTimeMillis());
        return po;
    }

    private LotteryGameCostItemParam getCostItem(Long orderId, SaveLotteryGameQualificationParam param) {
        int quantity = 1;
        Integer costQuantity = param.getCostQuantity();
        if (costQuantity != null && costQuantity > 0) {
            quantity = costQuantity;
        }
        return LotteryGameCostItemParam.builder()
                .orderId(orderId)
                .userId(param.getUserId())
                .itemId(param.getCostItemId())
                .quantity(quantity)
                .lotteryGamePlayType(param.getLotteryGamePlayType())
                .build();
    }

    /**
     * 减少奖项总数量
     *
     * @param snapId
     * @param quantity
     * @return
     */
    private void reduceRewardCount(long snapId, int quantity) {
        String key = RedisActivityKeys.getLotteryGameRewardCount(snapId);
        if (isDev()) {
            String value = redisService.get(key);
            int stock = NumberUtils.toInt(redisService.get(key));
            log.info("减少奖项总数量 snapId={} quantity={} value={} stock={}", snapId, quantity, value, stock);
        }
        redisService.increase(key, -Math.abs(quantity));
    }

    private int effectiveClick(double clickRate) {
        List<WeightRandom.WeightObj<Boolean>> weightList = ImmutableList.of(
                new WeightRandom.WeightObj<>(TRUE, clickRate),
                new WeightRandom.WeightObj<>(FALSE, 1 - clickRate)
        );
        return RandomUtil.weightRandom(weightList).next() == TRUE ? BoolIntVal.True : BoolIntVal.False;
    }
}
