package com.starbuds.server.service.finance.lucky;

import cn.hutool.core.collection.CollectionUtil;
import com.starbuds.server.common.pojo.finance.LuckyGiftLotteryConfig;
import com.starbuds.server.service.finance.util.ProbabilityUtil;
import lombok.Data;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.*;

/**
 * 幸运礼物核心组件
 *
 * @author Lay
 * @date 2020/1/11
 */
public abstract class LuckyGiftCore {
    /**
     * 抽奖函数
     * 资金池保底模型 轮次保底模型 概率模型
     *
     * @param giftId
     * @param userId
     * @param giftPriceType
     * @param giftPrice
     * @param times
     * @return
     */
    public LotteryResult lottery(Long giftId, Long userId, Integer giftPriceType, Long giftPrice, Integer times) {
        // 读取抽奖配置
        LuckyGiftLotteryConfig config = getConfig();

        if (config == null) {
            return null;
        }

        long lossAmount = getLossAmount(giftPrice, config.getAnchorRatio(), config.getPlatformRatio());

        // 礼物资金池保底模型（不管多少数量次数都只能抽一次）
        Double giftPoolRatio = 0d;
        if (config.isGiftPoolModeEnable()) {
            // 资金池抽成
            if (config.getGiftPoolRatio() != null) {
                giftPoolRatio = config.getGiftPoolRatio();
                BigDecimal ratio = new BigDecimal(String.valueOf(Math.abs(config.getGiftPoolRatio())));
                long giftPoolAmount = BigDecimal.valueOf(lossAmount).multiply(ratio).setScale(0, RoundingMode.DOWN).longValue();
                lossAmount -= giftPoolAmount;
                addGiftPool(giftId, giftPoolAmount * times);
            }

            // 资金池抢锁
            if (getGiftPoolLock(giftId, userId)) {
                // 判断奖池
                Integer multiple = 1000;
                long giftPoolAmount = getGiftPoolAmount(giftId);
                if (giftPoolAmount >= giftPrice * multiple) {
                    List<Integer> multiples = Arrays.asList(multiple);
                    // 清奖池
                    addGiftPool(giftId, -giftPrice * multiple);
                    // 解锁
                    releaseGiftPoolLock(giftId);
                    // 直接中奖
                    return handleComplete(giftId, userId, giftPriceType, giftPrice, multiples, -giftPrice * multiple, 0);
                }
            }
        }

        // 数据准备阶段，必须预读取，不然次数多了网络耗时线性增长
        // 读取用户历史输的钱
        long luckyGiftLoss = getLuckyGiftLoss(giftId, userId);
        // 本次抽奖输的钱
        long userLoss = luckyGiftLoss;
        // 读取用户未中奖轮次
        int missedRounds = getLuckyGiftMiss(giftId, userId);

        // 资金池保底模型
        boolean poolGuaranteedModeEnable = false;
        // 资金池保底倍数
        List<Integer> poolMultipleList = Collections.EMPTY_LIST;
        if (config.isPoolGuaranteedModeEnable()) {
            // 从配置读取资金池保底倍数
            poolMultipleList = config.getPoolMultipleList();
            if (CollectionUtil.isNotEmpty(poolMultipleList)) {
                // 开启模式
                poolGuaranteedModeEnable = true;
                // 倒序排序
                Collections.sort(poolMultipleList, Collections.reverseOrder());
            }
        }

        // 轮次保底模型
        boolean roundGuaranteedModeEnable = false;
        // 轮次保底倍数
        List<Integer> roundMultipleList = Collections.EMPTY_LIST;
        if (config.isRoundGuaranteedModeEnable()) {
            // 从配置读取轮次保底倍数
            roundMultipleList = config.getRoundMultipleList();
            if (CollectionUtil.isNotEmpty(roundMultipleList)) {
                roundGuaranteedModeEnable = true;
                Collections.sort(roundMultipleList, Collections.reverseOrder());
            }
        }

        // 纯概率模型
        boolean randomModeEnable = false;
        // 概率列表
        List<LuckyGiftLotteryConfig.MultipleProbability> multipleProbabilityList = Collections.EMPTY_LIST;
        if (config.isRandomModeEnable()) {
            multipleProbabilityList = config.getProbabilityList();
            if (CollectionUtil.isNotEmpty(multipleProbabilityList)) {
                randomModeEnable = true;
            }
        }

        // 用户中奖倍数缓存
        List<Integer> winMultipleList = new ArrayList<>();
        // 抽奖循环
        LOTTERY:
        for (int t = 0; t < times; t++) {
            // 个人资金池保底模型
            if (poolGuaranteedModeEnable) {
                int index = poolMultipleList.size();
                for (Integer multiple : poolMultipleList) {
                    long amount = multiple * giftPrice;
                    if (userLoss > amount) {
                        double[] probArr = new double[]{ProbabilityUtil.getGuaranteedProbability(config.getPoolFirstRoundProb(), index, config.getPoolGuaranteedRound())};
                        if (ProbabilityUtil.getRandomColumnIndex(probArr) == 0) {
                            // 中奖了 结束这一轮
                            winMultipleList.add(multiple);
                            userLoss -= giftPrice * multiple;
                            missedRounds = 0;
                            continue LOTTERY;
                        }

                        // 没中奖
                        userLoss += lossAmount;
                        missedRounds++;
                        continue LOTTERY;
                    }
                    index--;
                }
            }

            // 轮次保底模型
            if (roundGuaranteedModeEnable) {
                int size = roundMultipleList.size();
                int index = size;
                for (Iterator<Integer> iterator = roundMultipleList.iterator(); iterator.hasNext(); ) {
                    Integer multiple = iterator.next();
                    // 获取保底轮次
                    int guaranteedRound = getGuaranteedRound(multiple, config.getAnchorRatio(), config.getPlatformRatio(), giftPoolRatio);
                    // 加个保护 如果超过最大的保底轮次，则认为是 miss 了最大轮次
                    if (index == size && missedRounds > guaranteedRound) {
                        missedRounds = guaranteedRound;
                    }
                    if (guaranteedRound == missedRounds) {
                        // 保底概率
                        double[] probArr = new double[]{ProbabilityUtil.getGuaranteedProbability(config.getRoundFirstRoundProb(), index, size)};
                        if (ProbabilityUtil.getRandomColumnIndex(probArr) == 0) {
                            // 中奖了
                            winMultipleList.add(multiple);
                            userLoss -= giftPrice * multiple;
                            missedRounds = 0;
                            continue LOTTERY;
                        }
                        // 没中奖
                        userLoss += lossAmount;
                        missedRounds++;
                        continue LOTTERY;
                    }
                    index--;
                }
            }

            // 纯概率模型
            if (randomModeEnable) {
                double[] probArr = new double[multipleProbabilityList.size()];
                Integer[] multipleArr = new Integer[probArr.length];
                int index = 0;
                for (LuckyGiftLotteryConfig.MultipleProbability probability : multipleProbabilityList) {
                    probArr[index] = probability.getProbability();
                    multipleArr[index] = probability.getMultiple();
                    index++;
                }

                int hitIndex = ProbabilityUtil.getRandomColumnIndex(probArr);
                if (hitIndex > -1) {
                    // 中奖
                    Integer multiple = multipleArr[hitIndex];
                    winMultipleList.add(multiple);
                    userLoss -= giftPrice * multiple;
                    missedRounds = 0;
                    continue LOTTERY;
                }

                // 没中奖
                userLoss += lossAmount;
                missedRounds++;
                continue LOTTERY;
            }

            // 没中奖
            userLoss += lossAmount;
            missedRounds++;
        }

        return handleComplete(giftId, userId, giftPriceType, giftPrice, winMultipleList, userLoss - luckyGiftLoss, missedRounds);
    }

    /**
     * 获取配置
     *
     * @return
     */
    public abstract LuckyGiftLotteryConfig getConfig();

    /**
     * 获取用户亏的钱
     *
     * @param giftId
     * @param userId
     * @return
     */
    protected abstract long getLuckyGiftLoss(Long giftId, Long userId);

    /**
     * 获取用户未中奖轮次
     *
     * @param giftId
     * @param userId
     * @return
     */
    protected abstract int getLuckyGiftMiss(Long giftId, Long userId);

    /**
     * 完成抽奖
     *
     * @param giftId
     * @param userId
     * @param giftPriceType
     * @param giftPrice
     * @param multiples
     * @param lossAmount
     * @param missedRounds
     * @return
     */
    protected abstract LotteryResult handleComplete(Long giftId, Long userId, Integer giftPriceType, Long giftPrice, List<Integer> multiples, long lossAmount, int missedRounds);

    /**
     * 增加礼物奖池
     *
     * @param giftId
     * @param amount
     */
    protected abstract void addGiftPool(Long giftId, Long amount);

    /**
     * 获取礼物资金池金额
     *
     * @param giftId
     * @return
     */
    protected abstract long getGiftPoolAmount(Long giftId);

    /**
     * 资金池抢锁
     *
     * @param giftId
     * @param userId
     * @return
     */
    protected abstract boolean getGiftPoolLock(Long giftId, Long userId);

    /**
     * 释放资金池锁
     *
     * @param giftId
     */
    protected abstract void releaseGiftPoolLock(Long giftId);

    /**
     * 获取保底轮次
     *
     * @param multiple
     * @param anchorRatio
     * @param platformRatio
     * @return
     */
    private static int getGuaranteedRound(Integer multiple, Double anchorRatio, Double platformRatio, Double giftPoolRatio) {
        BigDecimal round = BigDecimal.valueOf(multiple)
                .divide(BigDecimal.ONE
                        .subtract(new BigDecimal(String.valueOf(anchorRatio)))
                        .subtract(new BigDecimal(String.valueOf(platformRatio)))
                        .subtract(new BigDecimal(String.valueOf(giftPoolRatio))), 2, BigDecimal.ROUND_HALF_UP);
        return round.setScale(0, RoundingMode.UP).intValue();
    }

    /**
     * 计算损失金额
     *
     * @param giftPrice
     * @param anchorRatio
     * @param platformRatio
     * @return
     */
    private static long getLossAmount(Long giftPrice, Double anchorRatio, Double platformRatio) {
        return BigDecimal.valueOf(giftPrice).multiply(BigDecimal.ONE.subtract(new BigDecimal(String.valueOf(anchorRatio))).subtract(new BigDecimal(String.valueOf(platformRatio)))).setScale(0, RoundingMode.DOWN).longValue();
    }

    @Data
    public static class LotteryResult {
        /**
         * 中奖倍数
         */
        private List<Integer> multiples;

        /**
         * 最大中奖倍数
         */
        private Integer maxMultiple;

        /**
         * 中奖金额
         */
        private Long winAmount;
    }
}
