package com.starbuds.server.service.finance.lucky;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.cache.RedisFinanceKeys;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.pojo.db.LuckyGiftConfigDO;
import com.starbuds.server.common.pojo.finance.LuckyGiftLotteryConfig;
import com.starbuds.server.common.pojo.mq.QueueGiftPaidMsg;
import com.starbuds.server.common.service.core.CacheService;
import com.starbuds.server.service.finance.dao.LuckyGiftConfigDAO;
import com.starbuds.server.service.finance.provider.IWalletProvider;
import com.starbuds.server.service.finance.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;

/**
 * 幸运礼物
 *
 * @author Lay
 * @date 2020/1/11
 */
@Component
@Slf4j
public class LuckyGiftProvider extends LuckyGiftCore {

    @Reference(version = "1.0.0")
    CacheService cacheService;

    @Resource
    LuckyGiftConfigDAO luckyGiftConfigDAO;

    @Resource
    IWalletProvider walletProvider;

    @Resource
    RocketProducer rocketProducer;

    /**
     * 获取配置
     *
     * @return
     */
    @Override
    public LuckyGiftLotteryConfig getConfig() {
        // 先从 Redis 中读取
        Object objConfig = cacheService.get(RedisFinanceKeys.LUCKY_GIFT_LOTTERY_CONFIG);
        if (objConfig != null) {
            return JSON.parseObject(objConfig.toString(), LuckyGiftLotteryConfig.class);
        }

        // 从数据库读取并缓存
        LuckyGiftConfigDO luckyGiftConfigDO = luckyGiftConfigDAO.selectLastEnabled();
        if (luckyGiftConfigDO != null) {
            return new LuckyGiftLotteryConfig(luckyGiftConfigDO);
        }

        return null;
    }

    /**
     * 获取用户亏的钱
     *
     * @param giftId
     * @param userId
     * @return
     */
    @Override
    protected long getLuckyGiftLoss(Long giftId, Long userId) {
        return Optional.ofNullable(cacheService.hashGet(RedisFinanceKeys.getLuckyGiftLoss(giftId), String.valueOf(userId))).map(t -> Convert.toLong(t)).orElse(0L);
    }

    /**
     * 获取用户未中奖轮次
     *
     * @param giftId
     * @param userId
     * @return
     */
    @Override
    protected int getLuckyGiftMiss(Long giftId, Long userId) {
        return Optional.ofNullable(cacheService.hashGet(RedisFinanceKeys.getLuckyGiftMiss(giftId), String.valueOf(userId))).map(t -> Convert.toInt(t)).orElse(0);
    }

    /**
     * 完成抽奖
     *
     * @param giftId
     * @param userId
     * @param giftPriceType
     * @param giftPrice
     * @param multiples
     * @param lossAmount
     * @param missedRounds
     * @return
     */
    @Override
    protected LotteryResult handleComplete(Long giftId, Long userId, Integer giftPriceType, Long giftPrice, List<Integer> multiples, long lossAmount, int missedRounds) {
        try {
            // 用户未中奖轮次设置
            cacheService.hashPut(RedisFinanceKeys.getLuckyGiftMiss(giftId), String.valueOf(userId), String.valueOf(missedRounds));

            // 用户单礼物损失金额
            cacheService.hashIncreaseLong(RedisFinanceKeys.getLuckyGiftLoss(giftId), String.valueOf(userId), lossAmount);

            // 用户总损失金额
            cacheService.hashIncreaseLong(RedisFinanceKeys.LUCKY_GIFT_LOSS_TOTAL, String.valueOf(userId), lossAmount);

            // 计算奖金
            if (CollectionUtil.isNotEmpty(multiples)) {
                // 奖金
                Long winAmount = 0L;
                Integer multipleTotal = 0;
                Integer maxMultiple = 0;
                for (Integer multiple : multiples) {
                    if (multiple > maxMultiple) {
                        maxMultiple = multiple;
                    }
                    winAmount += giftPrice * multiple;
                    multipleTotal += multiple;
                }
                WalletTypeEnum walletType = WalletTypeEnum.toEnum(giftPriceType);
                Long newBalance = walletProvider.addBalance(walletType, userId, winAmount, String.format("%d%s幸运礼物中奖%d倍", giftPrice, walletType.getName(), multipleTotal), null);

                if (newBalance != null) {
                    LotteryResult lotteryResult = new LotteryResult();
                    lotteryResult.setMultiples(multiples);
                    lotteryResult.setMaxMultiple(maxMultiple);
                    lotteryResult.setWinAmount(winAmount);
                    return lotteryResult;
                }
            }

        } catch (Exception e) {
            log.error("抽奖处理失败", e);
        }

        return new LotteryResult();
    }

    /**
     * 增加礼物奖池
     *
     * @param giftId
     * @param amount
     */
    @Override
    protected void addGiftPool(Long giftId, Long amount) {
        cacheService.hashIncreaseLong(RedisFinanceKeys.LUCKY_GIFT_POOL, String.valueOf(giftId), amount);
    }

    /**
     * 获取礼物资金池金额
     *
     * @param giftId
     * @return
     */
    @Override
    protected long getGiftPoolAmount(Long giftId) {
        return Convert.toLong(cacheService.hashGet(RedisFinanceKeys.LUCKY_GIFT_POOL, String.valueOf(giftId)), 0L);
    }

    /**
     * 资金池抢锁
     *
     * @param giftId
     * @param userId
     * @return
     */
    @Override
    protected boolean getGiftPoolLock(Long giftId, Long userId) {
        return cacheService.setIfAbsent(RedisFinanceKeys.getLuckyGiftPoolLock(giftId), String.valueOf(userId), 10, TimeUnit.SECONDS);
    }

    /**
     * 释放资金池锁
     *
     * @param giftId
     */
    @Override
    protected void releaseGiftPoolLock(Long giftId) {
        cacheService.delKey(RedisFinanceKeys.getLuckyGiftPoolLock(giftId));
    }

    @Async
    public void asyncLottery(QueueGiftPaidMsg msg) {
        // 拿锁
        String lockKey = RedisFinanceKeys.getLuckyGiftLotteryLock(msg.getSenderUserId());
        while (true) {
            if (cacheService.setIfAbsent(lockKey, String.valueOf(msg.getOrderId()), 10000L, TimeUnit.MILLISECONDS)) {
                LotteryResult lotteryResult = lottery(msg.getGiftId(), msg.getSenderUserId(), msg.getGiftPriceType(), msg.getGiftPrice(), msg.getQuantity());

                if (lotteryResult != null && CollectionUtil.isNotEmpty(lotteryResult.getMultiples())) {
                    msg.setWinMultipleMax(lotteryResult.getMaxMultiple());
                    msg.setWinAmount(lotteryResult.getWinAmount());
                }
                // 解锁
                cacheService.delKey(lockKey);

                break;
            }
        }

        // 发出礼物消息
        rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.GIFT_PAID, msg.serialize());
    }
}
