package com.starbuds.server.service.finance.provider;

import cn.hutool.core.collection.CollectionUtil;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.pojo.db.AgentWalletDO;
import com.starbuds.server.common.pojo.db.AgentWalletHisDO;
import com.starbuds.server.common.pojo.finance.UserWalletHisVO;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.service.finance.dao.AgentWalletDAO;
import com.starbuds.server.service.finance.dao.AgentWalletHisDAO;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * 工会钱包
 *
 * @author Lay
 * @date 2020/2/26
 */
@Component
@Slf4j
public class AgentWalletProvider {
    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Reference(version = "1.0.0")
    StatPlatformService statPlatformService;

    @Resource
    AgentWalletDAO agentWalletDAO;

    @Resource
    AgentWalletHisDAO agentWalletHisDAO;

    /**
     * 增加余额
     *
     * @param agentId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    public BigDecimal addBalance(long agentId, BigDecimal balanceValue, String hisRemarks, Long orderId) {
        return changeBalance(agentId, balanceValue.abs(), hisRemarks, orderId);
    }

    /**
     * 减少余额
     *
     * @param agentId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    public BigDecimal reduceBalance(long agentId, BigDecimal balanceValue, String hisRemarks, Long orderId) {
        return changeBalance(agentId, balanceValue.abs().negate(), hisRemarks, orderId);
    }

    /**
     * 改变余额
     *
     * @param agentId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    public BigDecimal changeBalance(long agentId, BigDecimal balanceValue, String hisRemarks, Long orderId) {
        // 判断账户是否存在
        AgentWalletDO agentWalletDO = agentWalletDAO.selectById(agentId);
        if (agentWalletDO == null) {
            agentWalletDO = new AgentWalletDO();
            agentWalletDO.setAgentId(agentId);
            agentWalletDO.setBalance(BigDecimal.ZERO);
            agentWalletDO.setVersion(0L);
            agentWalletDO.setUpdateTime(System.currentTimeMillis());
            try {
                agentWalletDAO.insert(agentWalletDO);
            } catch (Exception e) {
                // 重试
                return changeBalance(agentId, balanceValue, hisRemarks, orderId);
            }
        }

        return executeChange(agentWalletDO, agentId, balanceValue, hisRemarks, orderId);
    }

    /**
     * 执行改变
     *
     * @param agentWalletDO
     * @param agentId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public BigDecimal executeChange(AgentWalletDO agentWalletDO, long agentId, BigDecimal balanceValue, String hisRemarks, Long orderId) {

        try {
            // 计算新余额
            BigDecimal balanceOld = agentWalletDO.getBalance();
            Long version = agentWalletDO.getVersion();
            BigDecimal balanceNew = balanceOld.add(balanceValue);

            // 余额不足
            if (balanceNew.compareTo(BigDecimal.ZERO) < 0) {
                return balanceNew;
            }

            long curTime = System.currentTimeMillis();

            // 更新
            try {
                int updated = agentWalletDAO.updateCas(agentId, balanceNew, version);
                if (updated == 0) {
                    // 重试
                    return changeBalance(agentId, balanceValue, hisRemarks, orderId);
                }
            } catch (Exception e) {

            }

            // 写入变更记录
            AgentWalletHisDO userWalletHisDO = new AgentWalletHisDO();
            userWalletHisDO.setHisId(identityService.getId());
            userWalletHisDO.setAgentId(agentId);
            userWalletHisDO.setBalanceOld(balanceOld);
            userWalletHisDO.setBalanceValue(balanceValue);
            userWalletHisDO.setBalanceNew(balanceNew);
            userWalletHisDO.setCreateTime(curTime);
            userWalletHisDO.setHisRemarks(hisRemarks);
            userWalletHisDO.setOrderId(orderId);
            agentWalletHisDAO.insert(userWalletHisDO);

            return balanceNew;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("余额变动失败", e);
        }

        return null;
    }

    /**
     * 获取工会钱包
     *
     * @param agentId
     * @return
     */
    public AgentWalletDO getAgentWallet(Long agentId) {
        return agentWalletDAO.selectById(agentId);
    }

    /**
     * 获取余额
     *
     * @param agentId
     * @return
     */
    public BigDecimal getBalance(Long agentId) {
        AgentWalletDO agentWalletDO = agentWalletDAO.selectById(agentId);
        if (agentWalletDO != null) {
            return agentWalletDO.getBalance();
        }
        return BigDecimal.ZERO;
    }

    /**
     * 获取工会钱包历史记录
     *
     * @param agentId
     * @param lastId
     * @param pageSize
     * @param isIncome
     * @param beginTime
     * @param endTime
     * @return
     */
    public List<UserWalletHisVO> getUserWalletHisList(Long agentId, Long lastId, int pageSize, Integer isIncome, Long beginTime, Long endTime) {
        if (isIncome == null) {
            isIncome = BoolIntVal.Ignore;
        }
        List<AgentWalletHisDO> agentWalletDOList = agentWalletHisDAO.selectAll(agentId, lastId, pageSize, isIncome, beginTime, endTime);
        if (CollectionUtil.isNotEmpty(agentWalletDOList)) {
            List<UserWalletHisVO> userWalletHisVOList = new ArrayList<>();
            for (Iterator<AgentWalletHisDO> iterator = agentWalletDOList.iterator(); iterator.hasNext(); ) {
                UserWalletHisVO userWalletHisVO = new UserWalletHisVO();
                BeanUtils.copyProperties(iterator.next(), userWalletHisVO);
                userWalletHisVOList.add(userWalletHisVO);
            }
            return userWalletHisVOList;
        }

        return Collections.emptyList();
    }

    /**
     * 求和
     *
     * @param agentId
     * @param isIncome
     * @param beginTime
     * @param endTime
     * @return
     */
    public BigDecimal sumByTypeAndTime(Long agentId, Integer isIncome, Long beginTime, Long endTime) {
        if (isIncome == null) {
            isIncome = BoolIntVal.Ignore;
        }
        return agentWalletHisDAO.sumByTypeAndTime(agentId, isIncome, beginTime, endTime);
    }
}
