package com.starbuds.server.service.finance.provider;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.crypto.SecureUtil;
import com.alibaba.fastjson.JSON;
import com.google.common.collect.ImmutableList;
import com.starbuds.server.common.constant.data.PropsTypeEnum;
import com.starbuds.server.common.constant.finance.BagItemTypeEnum;
import com.starbuds.server.common.pojo.data.*;
import com.starbuds.server.common.pojo.db.UserBagChangeHisDO;
import com.starbuds.server.common.pojo.db.UserBagDO;
import com.starbuds.server.common.pojo.db.UserBagHisDO;
import com.starbuds.server.common.pojo.finance.BagItem;
import com.starbuds.server.common.pojo.media.GiftItemVO;
import com.starbuds.server.common.pojo.message.im.activity.WinItem;
import com.starbuds.server.common.pojo.message.im.activity.WinningRecordParam;
import com.starbuds.server.common.pojo.mq.QueueAddItemToBagMsg;
import com.starbuds.server.common.pojo.mq.QueueReduceItemToBagMsg;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.message.biz.LotteryGameImMsgService;
import com.starbuds.server.common.validation.Validator;
import com.starbuds.server.service.finance.dao.UserBagDAO;
import com.starbuds.server.service.finance.dao.UserBagHisDAO;
import com.starbuds.server.service.finance.provider.lottery.LotteryGameStatisticsProvider;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections.CollectionUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.hibernate.validator.HibernateValidator;
import org.jetbrains.annotations.Nullable;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.util.*;
import java.util.stream.Collectors;

import static com.google.common.base.Preconditions.*;
import static com.starbuds.server.service.finance.util.SpringContextUtil.isDev;

/**
 * 背包服务实现
 *
 * @author Lay
 * @date 2020/1/18
 */
@Slf4j
@Component
public class BagProvider implements IBagProvider {

    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Resource
    UserBagDAO userBagDAO;

    @Resource
    UserBagHisDAO userBagHisDAO;

    @Reference(version = "1.0.0")
    private LotteryGameImMsgService lotteryGameImMsgService;

    @Resource
    private UserBagChangeHisProvider userBagChangeHisProvider;

    @Resource
    private LotteryGameStatisticsProvider lotteryGameStatisticsProvider;

    private Validator validator = new Validator(HibernateValidator.class);

    private static String getId(Long userId, Long itemId, Integer itemType, Long expireTime) {
        return SecureUtil.md5(String.format("%d:%d:%d:%d", userId, itemId, itemType, expireTime));
    }

    /**
     * 增加物品数量
     *
     * @param userId
     * @param itemId
     * @param itemType
     * @param quantity
     * @param expireTime
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Override
    @Transactional(rollbackFor = Exception.class)
    public boolean addQuantity(Long userId, Long itemId, Integer itemType, Integer quantity, Long expireTime, String hisRemarks, Long orderId) {
        if (quantity < 1) {
            return false;
        }

        // 小于等于0 或者 null 都表示永久
        expireTime = expireTime == null || expireTime < 1 ? -1 : expireTime;
        String id = getId(userId, itemId, itemType, expireTime);

        Long curTime = System.currentTimeMillis();

        UserBagDO userBagDO = new UserBagDO();
        userBagDO.setId(id);
        userBagDO.setUserId(userId);
        userBagDO.setItemId(itemId);
        userBagDO.setItemType(itemType);
        userBagDO.setQuantity(quantity);
        userBagDO.setCreateTime(curTime);
        userBagDO.setUpdateTime(curTime);
        userBagDO.setVersion(0L);
        userBagDO.setExpireTime(expireTime);
        userBagDAO.insertOrUpdate(userBagDO);

        // 写入变更记录
        UserBagHisDO userBagHisDO = new UserBagHisDO();
        userBagHisDO.setHisId(identityService.getId());
        userBagHisDO.setBagId(id);
        userBagHisDO.setUserId(userId);
        userBagHisDO.setItemId(itemId);
        userBagHisDO.setItemType(itemType);
        userBagHisDO.setQuantity(quantity);
        userBagHisDO.setCreateTime(curTime);
        userBagHisDO.setHisRemarks(hisRemarks);
        userBagHisDO.setOrderId(orderId);
        userBagHisDAO.insert(userBagHisDO);

        // 周年庆活动临时统计
        lotteryGameStatisticsProvider.sendGetForkMessages(userId, itemId, itemType, quantity, userBagHisDO.getHisId());
        return true;
    }

    /**
     * 增加物品数量
     *
     * @param msg
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public boolean addQuantity(QueueAddItemToBagMsg msg) {
        return addQuantity(msg.getUserId(), msg.getItemId(), msg.getItemType().getValue(),
                msg.getQuantity(), msg.getExpireTime(), msg.getRemark(), msg.getOrderId());
    }

    /**
     * 扣减物品数量
     *
     * @param msg
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public boolean reduceQuantity(QueueReduceItemToBagMsg msg) {
        Optional failMsg = validator.getFailMsgOfOne(msg);
        if (failMsg.isPresent()) {
            log.error(String.format("扣减背包物品失败，参数有误:%s msg=%s", failMsg.get(), JSON.toJSONString(msg, true)));
            return false;
        }

        return reduceQuantity(msg.getUserId(), msg.getItemId(), msg.getItemType().getValue(),
                msg.getQuantity(), msg.getRemark(), msg.getOrderId()) != null;
    }

    /**
     * 减少物品数量
     *
     * @param userId
     * @param itemId
     * @param itemType
     * @param quantity
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public Integer reduceQuantity(Long userId, Long itemId, Integer itemType, Integer quantity, String hisRemarks, Long orderId) {
        quantity = Math.abs(quantity);
        if (quantity < 1) {
            return null;
        }

        Long curTime = System.currentTimeMillis();

        // 找出所有同类型的物品 按照过期时间正序排列 优先使用快过期的 -1 永久的放最底下
        List<UserBagDO> list = userBagDAO.selectAllByIdAndType(userId, itemId, itemType, curTime);
        if (CollectionUtil.isEmpty(list)) {
            return -quantity;
        }

        list.sort(((o1, o2) -> {
            if (o1.getExpireTime().equals(o2.getExpireTime())) {
                return 0;
            }

            if (o1.getExpireTime().equals(-1L)) {
                return 1;
            }

            return o1.getExpireTime() < o2.getExpireTime() ? -1 : 1;
        }));

        // 判断总数量是否足够
        Integer totalQuantity = 0;
        Integer needQuantity = quantity;
        Map<String, Integer> reduceMap = new HashMap<>();
        Map<String, Integer> reduceDiffMap = new HashMap<>();
        Map<String, Long> reduceVersion = new HashMap<>();
        for (UserBagDO userBagDO : list) {
            totalQuantity += userBagDO.getQuantity();
            // 当前扣减 需要的数量 - 该背包记录拥有的数量
            Integer diffQuantity = needQuantity - userBagDO.getQuantity();
            if (diffQuantity > 0) {
                // 说明数量不够 加入满额扣减
                reduceMap.put(userBagDO.getId(), 0);
                reduceVersion.put(userBagDO.getId(), userBagDO.getVersion());
                reduceDiffMap.put(userBagDO.getId(), -userBagDO.getQuantity());
                needQuantity = diffQuantity;
            } else {
                // 数量足够 加入部分扣减
                reduceMap.put(userBagDO.getId(), userBagDO.getQuantity() - needQuantity);
                reduceVersion.put(userBagDO.getId(), userBagDO.getVersion());
                reduceDiffMap.put(userBagDO.getId(), -needQuantity);
                break;
            }
        }

        Integer quantityNew = totalQuantity - quantity;
        if (quantityNew < 0) {
            return quantityNew;
        }

        // 开启事务逐次扣减
        try {
            Long[] ids = identityService.getIds(reduceMap.size());
            List<UserBagHisDO> bagHisList = new ArrayList<>();
            int index = 0;
            for (Map.Entry<String, Integer> entry : reduceMap.entrySet()) {
                int updated = userBagDAO.updateQuantity(entry.getKey(), entry.getValue(), reduceVersion.get(entry.getKey()), curTime);
                if (updated == 0) {
                    throw new RuntimeException("重试扣减");
                }

                // 写入变更记录
                UserBagHisDO userBagHisDO = new UserBagHisDO();
                userBagHisDO.setHisId(ids[index]);
                userBagHisDO.setBagId(entry.getKey());
                userBagHisDO.setItemId(itemId);
                userBagHisDO.setUserId(userId);
                userBagHisDO.setItemType(itemType);
                userBagHisDO.setQuantity(reduceDiffMap.get(entry.getKey()));
                userBagHisDO.setCreateTime(curTime);
                userBagHisDO.setHisRemarks(hisRemarks);
                userBagHisDO.setOrderId(orderId);
                bagHisList.add(userBagHisDO);

                index++;
            }

            // 保存记录
            userBagHisDAO.batchInsert(bagHisList);

            return quantityNew;
        } catch (Exception e) {
            if (isDev()) {
                log.error("背包减少物品数量失败", e);
            }
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            // 重试
            reduceQuantity(userId, itemId, itemType, quantity, hisRemarks, orderId);
        }

        return null;
    }

    /**
     * 减少数量
     *
     * @param bagId
     * @param quantity
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public Integer reduceQuantity(String bagId, Integer quantity, String hisRemarks, Long orderId) {
        quantity = Math.abs(quantity);
        if (quantity < 1) {
            return null;
        }

        UserBagDO userBagDO = userBagDAO.selectById(bagId);
        if (userBagDO == null) {
            return null;
        }

        Integer newQuantity = userBagDO.getQuantity() - quantity;
        if (newQuantity < 0) {
            return newQuantity;
        }

        Long curTime = System.currentTimeMillis();
        Long hisId = identityService.getId();
        try {
            int updated = userBagDAO.updateQuantity(bagId, newQuantity, userBagDO.getVersion(), curTime);
            if (updated == 0) {
                throw new Exception("重试扣减");
            }
        } catch (Exception e) {
            // 重试
            reduceQuantity(bagId, quantity, hisRemarks, orderId);
        }


        // 写入变更记录
        UserBagHisDO userBagHisDO = new UserBagHisDO();
        userBagHisDO.setHisId(hisId);
        userBagHisDO.setBagId(bagId);
        userBagHisDO.setItemId(userBagDO.getItemId());
        userBagHisDO.setUserId(userBagDO.getUserId());
        userBagHisDO.setItemType(userBagDO.getItemType());
        userBagHisDO.setQuantity(quantity - userBagDO.getQuantity());
        userBagHisDO.setCreateTime(curTime);
        userBagHisDO.setHisRemarks(hisRemarks);
        userBagHisDO.setOrderId(orderId);
        userBagHisDAO.insert(userBagHisDO);

        return newQuantity;
    }

    /**
     * 获取用户礼物物品
     *
     * @param userId
     * @return
     */
    @Override
    public List<BagItem<GiftItemVO>> getUserGiftItems(Long userId) {
        return userBagDAO.selectGiftItemsByUserId(userId, BagItemTypeEnum.GIFT.getValue(), System.currentTimeMillis());
    }

    /**
     * 获取用户背包道具
     *
     * @param userId
     * @return
     */
    @Override
    public List<BagItem<PropsVO>> getUserPropsItems(Long userId, Integer propsType) {
        return userBagDAO.selectPropsItemsByUserId(userId, BagItemTypeEnum.PROPS.getValue(), propsType, System.currentTimeMillis());
    }

    /**
     * 获取用户背包道具
     *
     * @param bagId
     * @param userId
     * @return
     */
    @Override
    public BagItem<PropsVO> getProps(String bagId, Long userId) {
        List<BagItem<PropsVO>> list = userBagDAO.selectPropsItemsById(bagId, userId, BagItemTypeEnum.PROPS.getValue(), System.currentTimeMillis());
        if (CollectionUtil.isNotEmpty(list)) {
            return list.get(0);
        }
        return null;
    }

    /**
     * 获取用户背包坐骑
     *
     * @param userId
     * @return
     */
    @Override
    public List<BagItem<MountsVO>> getUserMountsItems(Long userId) {
        return userBagDAO.selectMountsItemsByUserId(userId, BagItemTypeEnum.MOUNTS.getValue(), System.currentTimeMillis());
    }

    /**
     * 获取用户背包勋章
     *
     * @param userId
     * @return
     */
    @Override
    public List<BagItem<MedalVO>> getUserMedalItems(Long userId) {
        return userBagDAO.selectMedalItemsByUserId(userId, BagItemTypeEnum.MEDAL.getValue(), System.currentTimeMillis());
    }

    /**
     * 获取用户背包头饰
     *
     * @param userId
     * @return
     */
    @Override
    public List<BagItem<HeadwearVO>> getUserHeadwearItems(Long userId) {
        return userBagDAO.selectHeadwearItemsByUserId(userId, BagItemTypeEnum.HEADWEAR.getValue(), System.currentTimeMillis());
    }

    /**
     * 获取用户背包聊天气泡
     *
     * @param userId
     * @return
     */
    @Override
    public List<BagItem<ChatBubbleVO>> getUserChatBubbleItems(Long userId) {
        return userBagDAO.selectChatBubbleItemsByUserId(userId, BagItemTypeEnum.CHAT_BUBBLE.getValue(), System.currentTimeMillis());
    }

    /**
     * 根据物品 id 获取数量
     *
     * @param userId
     * @param bagItem
     * @param itemId
     * @return
     */
    @Override
    public int getQuantityByItemId(Long userId, BagItemTypeEnum bagItem, Long itemId) {
        return Optional.ofNullable(userBagDAO.sumQuantityByItemId(userId, bagItem.getValue(), itemId, System.currentTimeMillis())).orElse(0);
    }

    /**
     * 根据道具类型获取数量
     *
     * @param userId
     * @param propsType
     * @return
     */
    @Override
    public int getPropsQuantityByType(Long userId, PropsTypeEnum propsType) {
        return Optional.ofNullable(userBagDAO.sumPropsQuantityByType(userId, BagItemTypeEnum.PROPS.getValue(), propsType.getValue(), System.currentTimeMillis())).orElse(0);
    }

    /**
     * 根据类型获取数量
     *
     * @param userId
     * @param bagItemType
     * @return
     */
    @Override
    public int getQuantityByType(Long userId, BagItemTypeEnum bagItemType) {
        return Optional.ofNullable(userBagDAO.sumQuantityByItemType(userId, bagItemType.getValue(), System.currentTimeMillis())).orElse(0);
    }

    /**
     * 获取用户礼物列表
     *
     * @param userId
     * @return
     */
    @Override
    public List<UserBagDO> getUserBagDoList(Long userId) {
        return userBagDAO.getUserBagDoList(userId);
    }

    /**
     * 通过礼物ID获取
     *
     * @param giftIds
     * @param userId
     * @return
     */
    @Override
    public List<UserBagDO> getUserBagByGiftIds(List<Long> giftIds, Long userId) {
        return userBagDAO.getUserBagByGiftIds(giftIds, userId);
    }

    /**
     * 添加背包物品并发送IM系统消息、统计消息
     *
     * @param msg
     */
    @Transactional(rollbackFor = Exception.class)
    public void addItemToBagThenSendImMsgStatQueueMsg(QueueAddItemToBagMsg msg) {
        ImmutableList<QueueAddItemToBagMsg> msgList = ImmutableList.of(msg);
        addItemsToBagThenSendImMsgStatQueueMsg(msgList);
    }

    /**
     * 批量添加背包物品并发送IM系统消息、统计消息
     *
     * @param msgList
     */
    @Transactional(rollbackFor = Exception.class)
    public void addItemsToBagThenSendImMsgStatQueueMsg(Collection<QueueAddItemToBagMsg> msgList) {
        checkArgument(CollectionUtils.isNotEmpty(msgList));
        String msgListStr = JSON.toJSONString(msgList);

        Collection<QueueAddItemToBagMsg> validMsgList = getValidQueueAddItemToBagMsgList(msgList);
        checkArgument(CollectionUtils.isNotEmpty(validMsgList));

        boolean createUserBagChangeHis = userBagChangeHisProvider.createBatch(validMsgList);
        checkState(createUserBagChangeHis, "向背包增加物品数量失败,创建背包变更记录失败 msgList=" + msgListStr);

        // 生产环境是JDK8，使用stream遇到异常时可能不能回滚
        for (QueueAddItemToBagMsg msg : validMsgList) {
            boolean add = addQuantity(msg);
            checkState(add, "向背包增加物品数量失败 msg=" + msg);
        }

        sendAddItemToBagImMsg(validMsgList);
//        lotteryGameStatisticsProvider.sendGetForkMessages(validMsgList);
    }

    /**
     * 获得有效的添加背包物品列表
     *
     * @param msgList
     * @return
     */
    @Nullable
    private Collection<QueueAddItemToBagMsg> getValidQueueAddItemToBagMsgList(Collection<QueueAddItemToBagMsg> msgList) {
        msgList = msgList.stream()
                .map(msg -> {
                    boolean check = checkQueueAddItemToBagMsg(msg);
                    if (check) {
                        return msg;
                    }
                    return null;
                })
                .filter(Objects::nonNull)
                .collect(Collectors.toList());
        return msgList;
    }

    @Transactional(rollbackFor = Exception.class)
    public void reduceItemToBag(QueueReduceItemToBagMsg msg) {
        boolean reduce = reduceQuantity(msg);
        checkState(reduce, "从背包扣减物品数量失败 msg=" + msg);
    }

    /**
     * 获得中奖记录参数
     *
     * @param msg
     * @return
     */
    private WinningRecordParam getWinningRecordParam(QueueAddItemToBagMsg msg) {
        WinningRecordParam param = new WinningRecordParam();
        param.setUserId(msg.getUserId());
        param.setSource(msg.getSource());

        WinItem winItem = new WinItem();
        winItem.setQuantity(msg.getQuantity());
        winItem.setPrizeName(msg.getItemName());
        List<WinItem> winItems = new ArrayList<>();
        winItems.add(winItem);
        param.setWinItems(winItems);
        return param;
    }

    /**
     * 验证参数
     *
     * @param msg
     * @return true:数据正确
     */
    private boolean checkQueueAddItemToBagMsg(QueueAddItemToBagMsg msg) {
        Optional failMsg = validator.getFailMsgOfOne(msg);
        if (failMsg.isPresent()) {
            log.error(String.format("添加背包物品参数有误:%s msg=%s", failMsg.get(), JSON.toJSONString(msg, true)));
            return false;
        }
        return true;
    }

    /**
     * 发送添加背包物品的IM消息
     *
     * @param msgList
     */
    private void sendAddItemToBagImMsg(Collection<QueueAddItemToBagMsg> msgList) {
        msgList.stream()
                .filter(msg -> (msg.getNoSendImMsg() == null) || (!msg.getNoSendImMsg()))
                .forEach(msg -> {
                    WinningRecordParam winningRecordParam = getWinningRecordParam(msg);
                    boolean sendImMsg = lotteryGameImMsgService.sendWinningRecordImMsg(winningRecordParam);
                    if (!sendImMsg) {
                        log.error(String.format("发送向背包增加物品数量IM消息失败 msg=%s winningRecordParam=%s",
                                msg, winningRecordParam));
                    }
                });
    }
}
