package com.starbuds.server.service.finance.provider;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.crypto.SecureUtil;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.pojo.db.ChannelWalletDO;
import com.starbuds.server.common.pojo.db.ChannelWalletHisDO;
import com.starbuds.server.common.pojo.finance.ChannelWalletHisVO;
import com.starbuds.server.common.pojo.finance.ChannelWalletVO;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.service.finance.dao.ChannelWalletDAO;
import com.starbuds.server.service.finance.dao.ChannelWalletHisDAO;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 钱包
 *
 * @author Lay
 * @date 2019/12/20
 */
@Component
@Slf4j
public class ChannelWalletProvider implements IChannelWalletProvider {

    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Reference(version = "1.0.0")
    StatPlatformService statPlatformService;

    @Resource
    ChannelWalletDAO channelWalletDAO;

    @Resource
    ChannelWalletHisDAO channelWalletHisDAO;

    public static String getWalletId(long channelId, int type) {
        return SecureUtil.md5(String.format("%d:%d", channelId, type));
    }

    public static WalletTypeEnum getWalletType(String walletId, long channelId) {
        for (WalletTypeEnum walletType : WalletTypeEnum.values()) {
            if (walletId.equals(getWalletId(channelId, walletType.getValue()))) {
                return walletType;
            }
        }
        return null;
    }

    /**
     * 增加余额
     *
     * @param walletType
     * @param channelId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Override
    public Long addBalance(WalletTypeEnum walletType, long channelId, long balanceValue, String hisRemarks, Long orderId) {
        return addBalance(walletType.getValue(), channelId, balanceValue, hisRemarks, orderId);
    }

    /**
     * 增加余额
     *
     * @param walletType
     * @param channelId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Override
    public Long addBalance(Integer walletType, long channelId, long balanceValue, String hisRemarks, Long orderId) {
        return changeBalance(walletType, channelId, Math.abs(balanceValue), hisRemarks, orderId);
    }

    /**
     * 减少余额
     *
     * @param walletType
     * @param channelId
     * @param balanceValue
     * @param hisRemarks
     * @return
     */
    @Override
    public Long reduceBalance(WalletTypeEnum walletType, long channelId, long balanceValue, String hisRemarks, Long orderId) {
        return reduceBalance(walletType.getValue(), channelId, balanceValue, hisRemarks, orderId);
    }

    /**
     * 减少余额
     *
     * @param walletType
     * @param channelId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Override
    public Long reduceBalance(Integer walletType, long channelId, long balanceValue, String hisRemarks, Long orderId) {
        return changeBalance(walletType, channelId, -Math.abs(balanceValue), hisRemarks, orderId);
    }

    /**
     * 改变余额
     *
     * @param channelId
     * @param walletType
     * @param balanceValue
     * @return
     */
    public Long changeBalance(Integer walletType, long channelId, long balanceValue, String hisRemarks, Long orderId) {
        String walletId = getWalletId(channelId, walletType);

        // 判断账户是否存在
        ChannelWalletDO channelWalletDO = channelWalletDAO.selectById(walletId);
        if (channelWalletDO == null) {
            channelWalletDO = new ChannelWalletDO();
            channelWalletDO.setWalletId(walletId);
            channelWalletDO.setChannelId(channelId);
            channelWalletDO.setType(walletType);
            channelWalletDO.setBalance(0L);
            channelWalletDO.setVersion(0L);
            channelWalletDO.setUpdateTime(System.currentTimeMillis());
            try {
                channelWalletDAO.insert(channelWalletDO);
            } catch (Exception e) {
                // 重试
                return changeBalance(walletType, channelId, balanceValue, hisRemarks, orderId);
            }
        }

        return executeChange(channelWalletDO, walletType, channelId, balanceValue, hisRemarks, orderId);
    }

    /**
     * 执行改变
     *
     * @param channelWalletDO
     * @param walletType
     * @param channelId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public Long executeChange(ChannelWalletDO channelWalletDO, Integer walletType, long channelId, long balanceValue, String hisRemarks, Long orderId) {

        try {
            // 计算新余额
            Long balanceOld = channelWalletDO.getBalance();
            Long version = channelWalletDO.getVersion();
            Long balanceNew = balanceOld + balanceValue;

            // 余额不足
            if (balanceNew < 0) {
                return balanceNew;
            }

            long curTime = System.currentTimeMillis();

            // 更新
            try {
                int updated = channelWalletDAO.updateCas(channelWalletDO.getWalletId(), balanceNew, version);
                if (updated == 0) {
                    channelWalletDO = channelWalletDAO.selectById(channelWalletDO.getWalletId());
                    // 重试
                    return executeChange(channelWalletDO, walletType, channelId, balanceValue, hisRemarks, orderId);
                }
            } catch (Exception e) {
                log.error("{} 更新失败，重试", channelWalletDO.getWalletId());
            }

            // 写入变更记录
            ChannelWalletHisDO channelWalletHisDO = new ChannelWalletHisDO();
            channelWalletHisDO.setHisId(identityService.getId());
            channelWalletHisDO.setWalletId(channelWalletDO.getWalletId());
            channelWalletHisDO.setChannelId(channelId);
            channelWalletHisDO.setType(walletType);
            channelWalletHisDO.setBalanceOld(balanceOld);
            channelWalletHisDO.setBalanceValue(balanceValue);
            channelWalletHisDO.setBalanceNew(balanceNew);
            channelWalletHisDO.setCreateTime(curTime);
            channelWalletHisDO.setHisRemarks(hisRemarks);
            channelWalletHisDO.setOrderId(orderId);
            channelWalletHisDAO.insert(channelWalletHisDO);

            // 异步统计
            try {
                StatPlatformTypeEnum statPlatformType = null;
                boolean isAdd = balanceValue > 0;
                WalletTypeEnum walletTypeEnum = WalletTypeEnum.toEnum(walletType);
                if (walletTypeEnum != null) {
                    switch (walletTypeEnum) {
                        case STAR_DIAMOND:
                            statPlatformType = isAdd ? StatPlatformTypeEnum.STAR_DIAMOND_ADD : StatPlatformTypeEnum.STAR_DIAMOND_REDUCE;
                            break;
                        case STAR_YUAN:
                            statPlatformType = isAdd ? StatPlatformTypeEnum.STAR_YUAN_ADD : StatPlatformTypeEnum.STAR_YUAN_REDUCE;
                            break;
                        case STAR_COIN:
                            statPlatformType = isAdd ? StatPlatformTypeEnum.STAR_COIN_ADD : StatPlatformTypeEnum.STAR_COIN_REDUCE;
                            break;
                        case STAR_DUST:
                            statPlatformType = isAdd ? StatPlatformTypeEnum.STAR_DUST_ADD : StatPlatformTypeEnum.STAR_DUST_REDUCE;
                            break;
                        case GOLD_COIN:
                            statPlatformType = isAdd ? StatPlatformTypeEnum.GOLD_COIN_ADD : StatPlatformTypeEnum.GOLD_COIN_REDUCE;
                            break;
                        default:
                            break;
                    }

                    statPlatformService.addPlatformStat(0L, statPlatformType, BigDecimal.valueOf(Math.abs(balanceValue)));
                }
            } catch (Exception e) {
                log.error("统计失败", e);
            }

            return balanceNew;
        } catch (Exception e) {
            log.error("余额变动失败", e);
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
        }

        return null;
    }

    /**
     * 获取渠道钱包
     *
     * @param channelId
     * @return
     */
    @Override
    public ChannelWalletVO getChannelWallet(Long channelId) {
        List<Integer> walletTypes = Arrays.stream(WalletTypeEnum.values()).map(WalletTypeEnum::getValue).collect(Collectors.toList());
        return getChannelWallet(channelId, walletTypes);
    }

    private ChannelWalletVO getChannelWallet(Long channelId, List<Integer> walletTypes) {
        ChannelWalletVO userWalletVO = new ChannelWalletVO();

        List<String> walletIds = new ArrayList<>();
        List<ChannelWalletVO.Item> items = new ArrayList<>();
        for (Integer walletType : walletTypes) {
            walletIds.add(ChannelWalletProvider.getWalletId(channelId, walletType));
        }

        List<ChannelWalletDO> channelWalletDOList = channelWalletDAO.selectByIds(walletIds);

        if (!CollectionUtil.isEmpty(channelWalletDOList)) {
            for (ChannelWalletDO channelWalletDO : channelWalletDOList) {
                ChannelWalletVO.Item item = new ChannelWalletVO.Item();
                item.setWalletType(channelWalletDO.getType());
                item.setBalance(channelWalletDO.getBalance());
                items.add(item);
                walletTypes.remove(item.getWalletType());
            }
        }

        if (CollectionUtil.isNotEmpty(walletTypes)) {
            // 补回去
            for (Integer walletType : walletTypes) {
                ChannelWalletVO.Item item = new ChannelWalletVO.Item();
                item.setWalletType(walletType);
                item.setBalance(0L);
                items.add(item);
            }
        }

        userWalletVO.setWalletItems(items);

        return userWalletVO;
    }

    /**
     * 获取渠道钱包
     *
     * @param walletId
     * @return
     */
    @Override
    public ChannelWalletDO getChannelWallet(String walletId) {
        return channelWalletDAO.selectById(walletId);
    }

    /**
     * 获取余额
     *
     * @param userId
     * @param walletType
     * @return
     */
    @Override
    public Long getBalance(Long userId, WalletTypeEnum walletType) {
        return getBalance(userId, walletType.getValue());
    }

    /**
     * 获取余额
     *
     * @param userId
     * @param walletType
     * @return
     */
    @Override
    public Long getBalance(Long userId, Integer walletType) {
        ChannelWalletDO channelWalletDO = channelWalletDAO.selectById(getWalletId(userId, walletType));
        if (channelWalletDO != null) {
            return channelWalletDO.getBalance();
        }
        return 0L;
    }

    /**
     * 创建渠道钱包
     *
     * @param userId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public boolean createChannelWallet(Long userId) {
        try {
            ChannelWalletDO channelWalletDO = new ChannelWalletDO();
            channelWalletDO.setChannelId(userId);
            channelWalletDO.setBalance(0L);
            channelWalletDO.setVersion(0L);
            channelWalletDO.setUpdateTime(System.currentTimeMillis());
            for (WalletTypeEnum walletTypeEnum : WalletTypeEnum.values()) {
                channelWalletDO.setWalletId(getWalletId(userId, walletTypeEnum.getValue()));
                channelWalletDO.setType(walletTypeEnum.getValue());
                channelWalletDAO.insert(channelWalletDO);
            }
            return true;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("渠道钱包创建失败", e);
        }
        return false;
    }

    /**
     * 获取渠道钱包历史记录
     *
     * @param userId
     * @param lastId
     * @param pageSize
     * @param walletType
     * @param isIncome
     * @param beginTime
     * @param endTime
     * @return
     */
    @Override
    public List<ChannelWalletHisVO> getChannelWalletHisList(Long userId, Long lastId, int pageSize, Integer walletType, Integer isIncome, Long beginTime, Long endTime) {
        if (isIncome == null) {
            isIncome = BoolIntVal.Ignore;
        }
        List<ChannelWalletHisDO> channelWalletDOList = channelWalletHisDAO.selectAll(getWalletId(userId, walletType), lastId, pageSize, isIncome, beginTime, endTime);
        if (CollectionUtil.isNotEmpty(channelWalletDOList)) {
            List<ChannelWalletHisVO> userWalletHisVOList = new ArrayList<>();
            for (Iterator<ChannelWalletHisDO> iterator = channelWalletDOList.iterator(); iterator.hasNext(); ) {
                ChannelWalletHisVO userWalletHisVO = new ChannelWalletHisVO();
                BeanUtils.copyProperties(iterator.next(), userWalletHisVO);
                userWalletHisVOList.add(userWalletHisVO);
            }
            return userWalletHisVOList;
        }

        return Collections.emptyList();
    }

    /**
     * 求和
     *
     * @param userId
     * @param walletType
     * @param isIncome
     * @param beginTime
     * @param endTime
     * @return
     */
    @Override
    public Long sumByTypeAndTime(Long userId, Integer walletType, Integer isIncome, Long beginTime, Long endTime) {
        if (isIncome == null) {
            isIncome = BoolIntVal.Ignore;
        }
        return channelWalletHisDAO.sumByTypeAndTime(getWalletId(userId, walletType), isIncome, beginTime, endTime);
    }
}
