package com.starbuds.server.service.finance.provider;

import cn.hutool.core.collection.CollectionUtil;
import com.starbuds.server.common.pojo.db.UserChatTimeWalletDO;
import com.starbuds.server.common.pojo.db.UserChatTimeWalletHisDO;
import com.starbuds.server.common.pojo.finance.ChatTimeWalletItem;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.service.finance.dao.UserChatTimeWalletDAO;
import com.starbuds.server.service.finance.dao.UserChatTimeWalletHisDAO;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

/**
 * 时间钱包服务
 *
 * @author Lay
 * @date 2020/4/6
 */
@Component
@Slf4j
public class ChatTimeWalletProvider {

    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Resource
    UserChatTimeWalletDAO userChatTimeWalletDAO;

    @Resource
    UserChatTimeWalletHisDAO userChatTimeWalletHisDAO;

    /**
     * 增加分钟数
     *
     * @param userId
     * @param walletType
     * @param expireTime
     * @param minutes
     * @param hisRemarks
     * @param businessId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public boolean addMinutes(Long userId, Integer walletType, Long expireTime, Integer minutes, String hisRemarks, String businessId) {
        minutes = Math.abs(minutes);
        try {
            Long hisId = identityService.getId();

            Long curTime = System.currentTimeMillis();
            UserChatTimeWalletDO userChatTimeWallet = new UserChatTimeWalletDO();
            userChatTimeWallet.setUserId(userId);
            userChatTimeWallet.setWalletType(walletType);
            userChatTimeWallet.setExpireTime(expireTime);
            userChatTimeWallet.setMinutes(minutes);
            userChatTimeWallet.setVersionNum(1L);
            userChatTimeWallet.setUpdateTime(curTime);
            userChatTimeWalletDAO.insertOrIncrease(userChatTimeWallet);

            UserChatTimeWalletHisDO userChatTimeWalletHis = new UserChatTimeWalletHisDO();
            userChatTimeWalletHis.setHisId(hisId);
            userChatTimeWalletHis.setUserId(userId);
            userChatTimeWalletHis.setWalletType(walletType);
            userChatTimeWalletHis.setExpireTime(expireTime);
            userChatTimeWalletHis.setChangeValue(minutes);
            userChatTimeWalletHis.setHisRemarks(hisRemarks);
            userChatTimeWalletHis.setBusinessId(businessId);
            userChatTimeWalletHis.setCreateTime(curTime);
            userChatTimeWalletHisDAO.insert(userChatTimeWalletHis);

            return true;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("时间钱包增加分钟数失败", e);
        }
        return false;
    }

    /**
     * 扣除分钟数
     *
     * @param userId
     * @param walletType
     * @param expireTime
     * @param minutes
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public Integer reduceMinutes(Long userId, Integer walletType, Long expireTime, Integer minutes, String hisRemarks, String businessId) {
        minutes = Math.abs(minutes);

        Long curTime = System.currentTimeMillis();
        // 如果已过期
        if (expireTime < curTime) {
            return -minutes;
        }

        // 先查询
        UserChatTimeWalletDO userChatTimeWallet = userChatTimeWalletDAO.selectById(userId, walletType, expireTime);
        if (userChatTimeWallet == null) {
            return -minutes;
        }

        Integer newMinutes = userChatTimeWallet.getMinutes() - minutes;
        if (newMinutes < 0) {
            return newMinutes;
        }

        // 执行扣减更新
        long whereVersionNum = userChatTimeWallet.getVersionNum();
        userChatTimeWallet.setMinutes(newMinutes);
        userChatTimeWallet.setUpdateTime(curTime);
        userChatTimeWallet.setVersionNum(whereVersionNum + 1);

        try {
            Long hisId = identityService.getId();
            int updated = userChatTimeWalletDAO.updateByVersion(userChatTimeWallet, whereVersionNum);
            if (updated == 0) {
                return reduceMinutes(userId, walletType, expireTime, minutes, hisRemarks, businessId);
            }

            UserChatTimeWalletHisDO userChatTimeWalletHis = new UserChatTimeWalletHisDO();
            userChatTimeWalletHis.setHisId(hisId);
            userChatTimeWalletHis.setUserId(userId);
            userChatTimeWalletHis.setWalletType(walletType);
            userChatTimeWalletHis.setExpireTime(expireTime);
            userChatTimeWalletHis.setChangeValue(minutes);
            userChatTimeWalletHis.setHisRemarks(hisRemarks);
            userChatTimeWalletHis.setBusinessId(businessId);
            userChatTimeWalletHis.setCreateTime(curTime);
            userChatTimeWalletHisDAO.insert(userChatTimeWalletHis);

            return newMinutes;
        } catch (Exception e) {
            log.error("时间钱包扣减失败");
        }
        return null;
    }

    /**
     * 扣除分钟数
     *
     * @param userId
     * @param walletType
     * @param minutes
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public Integer reduceMinutes(Long userId, Integer walletType, Integer minutes, String hisRemarks, String businessId) {
        minutes = Math.abs(minutes);
        // 找出未过期的逐个扣除
        Long curTime = System.currentTimeMillis();
        List<UserChatTimeWalletDO> list = userChatTimeWalletDAO.selectEffective(userId, walletType, curTime);

        if (CollectionUtil.isNotEmpty(list)) {
            // 排序
            list.sort((o1, o2) -> {
                if (o1.getExpireTime().equals(o2.getExpireTime())) {
                    return 0;
                }

                if (o1.getExpireTime().equals(-1L)) {
                    return 1;
                }

                return o1.getExpireTime() < o2.getExpireTime() ? -1 : 1;
            });

            // 判断总数量是否足够
            Integer totalMinutes = 0;
            Integer needMinutes = minutes;
            Map<UserChatTimeWalletDO, Integer> reduceMap = new HashMap<>();
            for (UserChatTimeWalletDO userChatTimeWallet : list) {
                totalMinutes += userChatTimeWallet.getMinutes();
                // 当前扣减
                Integer diffQuantity = needMinutes - userChatTimeWallet.getMinutes();
                if (diffQuantity > 0) {
                    // 说明数量不够 加入满额扣减
                    reduceMap.put(userChatTimeWallet, 0);
                    needMinutes = diffQuantity;
                } else {
                    // 数量足够 加入部分扣减
                    reduceMap.put(userChatTimeWallet, userChatTimeWallet.getMinutes() - needMinutes);
                    break;
                }
            }

            Integer minutesNew = totalMinutes - minutes;
            if (minutes < 0) {
                return minutes;
            }

            // 开启事务逐次扣减
            Long[] hisIds = identityService.getIds(reduceMap.size());
            int index = 0;
            for (Map.Entry<UserChatTimeWalletDO, Integer> entry : reduceMap.entrySet()) {
                UserChatTimeWalletDO userChatTimeWallet = entry.getKey();
                long whereVersion = userChatTimeWallet.getVersionNum();
                userChatTimeWallet.setMinutes(entry.getValue());
                userChatTimeWallet.setVersionNum(whereVersion + 1);
                userChatTimeWallet.setUpdateTime(curTime);
                int updated = userChatTimeWalletDAO.updateByVersion(userChatTimeWallet, whereVersion);
                if (updated == 0) {
                    TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
                    return reduceMinutes(userId, walletType, minutes, hisRemarks, businessId);
                }

                UserChatTimeWalletHisDO userChatTimeWalletHis = new UserChatTimeWalletHisDO();
                userChatTimeWalletHis.setHisId(hisIds[index]);
                userChatTimeWalletHis.setUserId(userId);
                userChatTimeWalletHis.setWalletType(walletType);
                userChatTimeWalletHis.setExpireTime(userChatTimeWallet.getExpireTime());
                userChatTimeWalletHis.setChangeValue(minutes);
                userChatTimeWalletHis.setHisRemarks(hisRemarks);
                userChatTimeWalletHis.setBusinessId(businessId);
                userChatTimeWalletHis.setCreateTime(curTime);
                userChatTimeWalletHisDAO.insert(userChatTimeWalletHis);

                index++;
            }
            return minutesNew;
        }

        return -minutes;
    }

    /**
     * 获取钱包剩余分钟数
     *
     * @param userId
     * @param walletType
     * @return
     */
    public Integer getWalletMinutes(Long userId, Integer walletType, Long time) {
        return Optional.ofNullable(userChatTimeWalletDAO.sumEffectiveMinutes(userId, walletType, time)).orElse(0);
    }

    /**
     * 获取时间列表
     *
     * @param userId
     * @param walletType
     * @param time
     * @return
     */
    public List<ChatTimeWalletItem> getChatTimeWalletItems(Long userId, Integer walletType, Long time) {
        return userChatTimeWalletDAO.selectEffectiveItems(userId, walletType, time);
    }
}
