package com.starbuds.server.service.finance.provider;

import cn.hutool.core.collection.CollectionUtil;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.pojo.db.FamilyWalletDO;
import com.starbuds.server.common.pojo.db.FamilyWalletHisDO;
import com.starbuds.server.common.pojo.finance.UserWalletHisVO;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.service.finance.dao.FamilyWalletDAO;
import com.starbuds.server.service.finance.dao.FamilyWalletHisDAO;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * 家族钱包
 *
 * @author Lay
 * @date 2020/2/26
 */
@Component
@Slf4j
public class FamilyWalletProvider {
    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Resource
    FamilyWalletDAO familyWalletDAO;

    @Resource
    FamilyWalletHisDAO familyWalletHisDAO;

    /**
     * 增加余额
     *
     * @param familyId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    public BigDecimal addBalance(long familyId, BigDecimal balanceValue, String hisRemarks, Long orderId) {
        return changeBalance(familyId, balanceValue.abs(), hisRemarks, orderId);
    }

    /**
     * 减少余额
     *
     * @param familyId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    public BigDecimal reduceBalance(long familyId, BigDecimal balanceValue, String hisRemarks, Long orderId) {
        return changeBalance(familyId, balanceValue.abs().negate(), hisRemarks, orderId);
    }

    /**
     * 改变余额
     *
     * @param familyId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    public BigDecimal changeBalance(long familyId, BigDecimal balanceValue, String hisRemarks, Long orderId) {
        // 判断账户是否存在
        FamilyWalletDO familyWalletDO = familyWalletDAO.selectById(familyId);
        if (familyWalletDO == null) {
            familyWalletDO = new FamilyWalletDO();
            familyWalletDO.setFamilyId(familyId);
            familyWalletDO.setBalance(BigDecimal.ZERO);
            familyWalletDO.setVersion(0L);
            familyWalletDO.setUpdateTime(System.currentTimeMillis());
            try {
                familyWalletDAO.insert(familyWalletDO);
            } catch (Exception e) {
                // 重试
                return changeBalance(familyId, balanceValue, hisRemarks, orderId);
            }
        }

        return executeChange(familyWalletDO, familyId, balanceValue, hisRemarks, orderId);
    }

    /**
     * 执行改变
     *
     * @param familyWalletDO
     * @param familyId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public BigDecimal executeChange(FamilyWalletDO familyWalletDO, long familyId, BigDecimal balanceValue, String hisRemarks, Long orderId) {

        try {
            // 计算新余额
            BigDecimal balanceOld = familyWalletDO.getBalance();
            Long version = familyWalletDO.getVersion();
            BigDecimal balanceNew = balanceOld.add(balanceValue);

            // 余额不足
            if (balanceNew.compareTo(BigDecimal.ZERO) < 0) {
                return balanceNew;
            }

            long curTime = System.currentTimeMillis();

            // 更新
            try {
                int updated = familyWalletDAO.updateCas(familyId, balanceNew, version);
                if (updated == 0) {
                    // 重试
                    return changeBalance(familyId, balanceValue, hisRemarks, orderId);
                }
            } catch (Exception e) {

            }

            // 写入变更记录
            FamilyWalletHisDO userWalletHisDO = new FamilyWalletHisDO();
            userWalletHisDO.setHisId(identityService.getId());
            userWalletHisDO.setFamilyId(familyId);
            userWalletHisDO.setBalanceOld(balanceOld);
            userWalletHisDO.setBalanceValue(balanceValue);
            userWalletHisDO.setBalanceNew(balanceNew);
            userWalletHisDO.setCreateTime(curTime);
            userWalletHisDO.setHisRemarks(hisRemarks);
            userWalletHisDO.setOrderId(orderId);
            familyWalletHisDAO.insert(userWalletHisDO);

            return balanceNew;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("余额变动失败", e);
        }

        return null;
    }

    /**
     * 获取家族钱包
     *
     * @param familyId
     * @return
     */
    public FamilyWalletDO getFamilyWallet(Long familyId) {
        return familyWalletDAO.selectById(familyId);
    }

    /**
     * 获取余额
     *
     * @param familyId
     * @return
     */
    public BigDecimal getBalance(Long familyId) {
        FamilyWalletDO familyWalletDO = familyWalletDAO.selectById(familyId);
        if (familyWalletDO != null) {
            return familyWalletDO.getBalance();
        }
        return BigDecimal.ZERO;
    }

    /**
     * 获取家族钱包历史记录
     *
     * @param familyId
     * @param lastId
     * @param pageSize
     * @param isIncome
     * @param beginTime
     * @param endTime
     * @return
     */
    public List<UserWalletHisVO> getFamilyWalletHisList(Long familyId, Long lastId, int pageSize, Integer isIncome, Long beginTime, Long endTime) {
        if (isIncome == null) {
            isIncome = BoolIntVal.Ignore;
        }
        List<FamilyWalletHisDO> familyWalletDOList = familyWalletHisDAO.selectAll(familyId, lastId, pageSize, isIncome, beginTime, endTime);
        if (CollectionUtil.isNotEmpty(familyWalletDOList)) {
            List<UserWalletHisVO> userWalletHisVOList = new ArrayList<>();
            for (Iterator<FamilyWalletHisDO> iterator = familyWalletDOList.iterator(); iterator.hasNext(); ) {
                UserWalletHisVO userWalletHisVO = new UserWalletHisVO();
                BeanUtils.copyProperties(iterator.next(), userWalletHisVO);
                userWalletHisVOList.add(userWalletHisVO);
            }
            return userWalletHisVOList;
        }

        return Collections.emptyList();
    }

    /**
     * 求和
     *
     * @param familyId
     * @param isIncome
     * @param beginTime
     * @param endTime
     * @return
     */
    public BigDecimal sumByTypeAndTime(Long familyId, Integer isIncome, Long beginTime, Long endTime) {
        if (isIncome == null) {
            isIncome = BoolIntVal.Ignore;
        }
        return familyWalletHisDAO.sumByTypeAndTime(familyId, isIncome, beginTime, endTime);
    }
}
