package com.starbuds.server.service.finance.provider;

import cn.hutool.core.convert.Convert;
import cn.hutool.core.date.DateUtil;
import com.starbuds.server.common.constant.activity.TriggerEventEnum;
import com.starbuds.server.common.constant.cache.RedisFinanceKeys;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.finance.PayStatusEnum;
import com.starbuds.server.common.constant.finance.PayWayEnum;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.constant.statistics.StatTimeType;
import com.starbuds.server.common.constant.statistics.StatUserTypeEnum;
import com.starbuds.server.common.constant.statistics.UserDataTypeEnum;
import com.starbuds.server.common.pojo.activity.UserRechargeEventParam;
import com.starbuds.server.common.pojo.db.InviteConfigDO;
import com.starbuds.server.common.pojo.db.InviteRewardHisDO;
import com.starbuds.server.common.pojo.mq.QueueUserCreateMsg;
import com.starbuds.server.common.pojo.mq.QueueUserFirstRechargeMsg;
import com.starbuds.server.common.pojo.mq.QueueVoiceChatRejectMsg;
import com.starbuds.server.common.pojo.mq.UserRechargeFinishMsg;
import com.starbuds.server.common.service.activity.TriggerService;
import com.starbuds.server.common.service.core.CacheService;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.data.ReleaseConfigService;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.common.service.statistics.StatUserService;
import com.starbuds.server.common.service.user.ChannelService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.finance.dao.InviteConfigDAO;
import com.starbuds.server.service.finance.dao.InviteRewardHisDAO;
import com.starbuds.server.service.finance.dao.PayOrderDAO;
import com.starbuds.server.service.finance.dao.UserDAO;
import com.starbuds.server.service.finance.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.Optional;

/**
 * 备注
 *
 * @author Lay
 * @date 2020/2/5
 */
@Component
@Slf4j
public class RechargeProvider {

    @Reference(version = "1.0.0")
    TriggerService triggerService;

    @Reference(version = "1.0.0")
    StatPlatformService statPlatformService;

    @Reference(version = "1.0.0")
    CacheService cacheService;

    @Reference(version = "1.0.0")
    UserService userService;

    @Reference(version = "1.0.0")
    StatUserService statUserService;

    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Reference(version = "1.0.0")
    ChannelService channelService;


    @Resource
    UserDAO userDAO;

    @Resource
    PayOrderDAO payOrderDAO;

    @Resource
    InviteConfigDAO inviteConfigDAO;

    @Resource
    InviteRewardHisDAO inviteRewardHisDAO;

    @Resource
    WalletProvider walletProvider;

    @Resource
    RocketProducer rocketProducer;

    /**
     * 中央时间戳
     */
    private static final String LUA_TIME;

    static {
        StringBuilder sb = new StringBuilder();
        sb.append("local t = redis.call('TIME');");
        sb.append("return t[1]");
        LUA_TIME = sb.toString();
    }

    /**
     * 获取 Redis 时间戳
     *
     * @return
     */
    public long getRedisTime() {
        return cacheService.luaLong(LUA_TIME, new ArrayList<>(), new ArrayList<>());
    }

    public Integer getToday() {
        Date date = new Date(getRedisTime() * 1000);
        SimpleDateFormat sdfDay = new SimpleDateFormat("yyyyMMdd");
        return Convert.toInt(sdfDay.format(date));
    }

    public boolean saveHasRecharged(Long channelId, Long userId, BigDecimal rechargeAmount, Integer payWay, Long payId) {

        boolean updated = userDAO.updateHasRecharged(userId, BoolIntVal.True) > 0;

        Double amount = rechargeAmount.setScale(2, RoundingMode.DOWN).doubleValue();

        // 缓存金额
        Double todayAmount = cacheService.zIncrementScore(RedisFinanceKeys.getRechargeDaily(getToday()), String.valueOf(userId), amount);

        // 触发充值事件
        UserRechargeEventParam eventParam = new UserRechargeEventParam();
        eventParam.setUserId(userId);
        eventParam.setPayWay(payWay);
        eventParam.setAmount(amount);
        eventParam.setTodayAmount(todayAmount - amount);
        eventParam.setIsFirst(false);
        eventParam.setIsRegDay(false);
        eventParam.setIsTodayFirst(false);

        // 是否注册当日
        boolean isRegDay = false;
        Long userCreateTime = Optional.ofNullable(userDAO.selectCreateTime(userId)).orElse(0L);
        SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd");
        if (sdf.format(new Date(userCreateTime)).equals(sdf.format(new Date(System.currentTimeMillis())))) {
            // 注册当日
            isRegDay = true;
            eventParam.setIsRegDay(true);
        }

        // 首充用户
        if (updated) {
            eventParam.setIsFirst(true);
            eventParam.setIsTodayFirst(true);

            if (payId != null) {
                payOrderDAO.updateFirstRecharge(payId, BoolIntVal.True);
            }

            if (isRegDay) {
                // 数据统计
                statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.FIRST_RECHARGE_USERS_DAILY, BigDecimal.ONE);
                statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.FIRST_RECHARGE_AMOUNT_DAILY, rechargeAmount);

                // 留存统计 首充人数
                statUserService.addUserKeepStat(channelId, Convert.toLong((DateUtil.format(new Date(userCreateTime), "yyyyMMdd"))), Convert.toLong(DateUtil.format(new Date(), "yyyyMMdd")), UserDataTypeEnum.FIRST_RECHARGE_USERS.getValue(), BigDecimal.ONE);

            }

            // 数据统计
            statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.FIRST_RECHARGE_USERS, BigDecimal.ONE);
            statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.FIRST_RECHARGE_AMOUNT, rechargeAmount);
            // 首充也是充值
            statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.RECHARGE_USERS, BigDecimal.ONE);


        } else {
            // 查询判断是否今日首次充值
            Calendar calendar = Calendar.getInstance();
            calendar.set(Calendar.HOUR_OF_DAY, 0);
            calendar.set(Calendar.MINUTE, 0);
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MILLISECOND, 0);
            long beginTime = calendar.getTimeInMillis();
            calendar.set(Calendar.HOUR_OF_DAY, 23);
            calendar.set(Calendar.MINUTE, 59);
            calendar.set(Calendar.SECOND, 59);
            calendar.set(Calendar.MILLISECOND, 999);
            long endTime = calendar.getTimeInMillis();
            if (payOrderDAO.countByUserAndStatusWithTimeRange(userId, PayStatusEnum.SUCCESSED.getValue(), beginTime, endTime) == 1) {
                // 当日首充
                if (payId != null) {
                    payOrderDAO.updateDailyFirstRecharge(payId, BoolIntVal.True);
                }

                eventParam.setIsTodayFirst(true);
                statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.RECHARGE_USERS, BigDecimal.ONE);

                // 留存统计 充值人数
                statUserService.addUserKeepStat(channelId, Convert.toLong((DateUtil.format(new Date(userCreateTime), "yyyyMMdd"))), Convert.toLong((DateUtil.format(new Date(), "yyyyMMdd"))), UserDataTypeEnum.RECHARGE_USERS.getValue(), BigDecimal.ONE);

            }
        }

        // 触发器
        triggerService.emitEvent(TriggerEventEnum.USER_RECHARGE, eventParam);

        // 数据统计
        statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.RECHARGE_TIMES, BigDecimal.ONE);
        statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.RECHARGE_AMOUNT, rechargeAmount);

        // 数据统计-支付方式
        StatPlatformTypeEnum rechargePayWayEnum = getRechargePayWayEnum(payWay);
        if (rechargePayWayEnum != null) {
            statPlatformService.addPlatformStat(channelId, rechargePayWayEnum.getValue(), rechargeAmount);
        }

        // 数据统计 邀请人
        Long inviterUserId = Optional.ofNullable(userService.getInviterUserId(userId)).orElse(0L);
        if (inviterUserId > 0) {
            // 增加邀请人的充值金额
            statUserService.addUserStat(inviterUserId, StatUserTypeEnum.INVITE_RECHARGE_AMOUNT.getValue(), rechargeAmount);
            if (updated) {
                // 增加平台邀请充值用户数
                statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.INVITE_RECHARGE_NUM, BigDecimal.ONE);
                // 首充用户增加邀请充值用户数
                statUserService.addUserStat(inviterUserId, StatUserTypeEnum.INVITE_RECHARGE_NUM.getValue(), BigDecimal.ONE);
            }
        }

        // 增加自己的充值金额
        statUserService.addUserStat(userId, StatUserTypeEnum.RECHARGE_AMOUNT.getValue(), rechargeAmount);
        // 增加充值次数
        statUserService.addUserStat(userId, StatUserTypeEnum.RECHARGE_TIMES.getValue(), BigDecimal.ONE);

        // 留存统计 充值次数
        statUserService.addUserKeepStat(channelId, Convert.toLong((DateUtil.format(new Date(userCreateTime), "yyyyMMdd"))), Convert.toLong((DateUtil.format(new Date(), "yyyyMMdd"))), UserDataTypeEnum.RECHARGE_TIMES.getValue(), BigDecimal.ONE);

        // 留存统计 充值金额
        statUserService.addUserKeepStat(channelId, Convert.toLong((DateUtil.format(new Date(userCreateTime), "yyyyMMdd"))), Convert.toLong((DateUtil.format(new Date(), "yyyyMMdd"))), UserDataTypeEnum.RECHARGE_AMOUNT.getValue(), rechargeAmount);

        // 首充用户
        if (updated) {
            //发送首充消息
            sendFristRecharge(userId);
        }

        return updated;
    }

    /**
     * 首充消息
     */
    private void sendFristRecharge(Long userId) {
        QueueUserFirstRechargeMsg rejectMsg = new QueueUserFirstRechargeMsg();
        rejectMsg.setUserId(userId);
        rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.USER_FIRST_RECHARGE, rejectMsg.serialize());
    }


    /**
     * 获取统计的支付方式类型
     *
     * @param payWay
     * @return
     */
    private StatPlatformTypeEnum getRechargePayWayEnum(Integer payWay) {
        PayWayEnum payWayEnum = PayWayEnum.toEnum(payWay);
        StatPlatformTypeEnum rechargePayWayEnum = null;
        switch (payWayEnum) {
            case APPLE:
                rechargePayWayEnum = StatPlatformTypeEnum.RECHARGE_PAY_WAY_APPLE;
                break;
            case HUAWEI:
                rechargePayWayEnum = StatPlatformTypeEnum.RECHARGE_PAY_WAY_HUAWEI;
                break;
            case ALI_APP:
                rechargePayWayEnum = StatPlatformTypeEnum.RECHARGE_PAY_WAY_ALI_APP;
                break;
            case ALI_H5:
                rechargePayWayEnum = StatPlatformTypeEnum.RECHARGE_PAY_WAY_ALI_H5;
                break;
            case ALI_PC:
                rechargePayWayEnum = StatPlatformTypeEnum.RECHARGE_PAY_WAY_ALI_PC;
                break;
            case WX_APP:
                rechargePayWayEnum = StatPlatformTypeEnum.RECHARGE_PAY_WAY_WX_APP;
                break;
            case WX_H5:
                rechargePayWayEnum = StatPlatformTypeEnum.RECHARGE_PAY_WAY_WX_H5;
                break;
            default:
                break;
        }
        return rechargePayWayEnum;
    }

    /**
     * 充值邀请奖励
     */
    @Transactional(rollbackFor = Exception.class)
    public void processInviteRewards(Long channelId, Long payId, Long userId, Long rechargeCoin) {
        // 判断是否有邀请人
        Long inviterUserId = Optional.ofNullable(userService.getInviterUserId(userId)).orElse(0L);
        log.info("判断是否有邀请人");
        if (inviterUserId > 0) {
            log.info("判断渠道是否配置 {}", channelId);
            // 判断渠道是否配置
            Integer configGroupId = channelService.getInviteConfigGroupId(channelId);
            if (configGroupId == null) {
                return;
            }

            SimpleDateFormat sdfYear = new SimpleDateFormat("yyyy");
            Long curYear = Convert.toLong(sdfYear.format(new Date()));
            // 获取邀请人数据
            long inviteNum = statUserService.sumStatByTimeRange(inviterUserId, StatUserTypeEnum.INVITE_RECHARGE_NUM.getValue(), StatTimeType.YEAR, 0L, curYear + 1).intValue();
            // 获取邀请充值数据
            BigDecimal inviteAmount = statUserService.sumStatByTimeRange(inviterUserId, StatUserTypeEnum.INVITE_RECHARGE_AMOUNT.getValue(), StatTimeType.YEAR, 0L, curYear + 1);

            log.info("邀请人 {} 邀请充值 {}", inviteNum, inviteAmount);
            // 获取邀请奖励配置
            InviteConfigDO inviteConfig = inviteConfigDAO.selectMaxByNumAndAmount(configGroupId, inviteNum, inviteAmount);
            if (inviteConfig != null) {
                log.info("发放奖励 {}", inviteConfig.getCoinRatio());
                // 发放奖励
                Long rewards = BigDecimal.valueOf(rechargeCoin).multiply(new BigDecimal(String.valueOf(inviteConfig.getCoinRatio()))).setScale(0, RoundingMode.DOWN).longValue();

                Long hisId = identityService.getId();
                walletProvider.addBalance(WalletTypeEnum.STAR_DIAMOND, inviterUserId, rewards, "邀请奖励", hisId);

                InviteRewardHisDO inviteRewardHis = new InviteRewardHisDO();
                inviteRewardHis.setHisId(hisId);
                inviteRewardHis.setUserId(userId);
                inviteRewardHis.setInviterUserId(inviterUserId);
                inviteRewardHis.setRechargeCoin(rechargeCoin);
                inviteRewardHis.setRewardCoin(rewards);
                inviteRewardHis.setCoinRatio(inviteConfig.getCoinRatio());
                inviteRewardHis.setPayId(payId);
                inviteRewardHis.setInviteNum(inviteNum);
                inviteRewardHis.setInviteAmount(inviteAmount);
                inviteRewardHis.setConfigId(inviteConfig.getConfigId());
                inviteRewardHis.setCreateTime(System.currentTimeMillis());
                inviteRewardHisDAO.insert(inviteRewardHis);

                // 统计
                statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.INVITE_STAR_DIAMOND, BigDecimal.valueOf(rewards));
                statUserService.addUserStat(inviterUserId, StatUserTypeEnum.INVITE_REWARDS_STAR_DIAMOND.getValue(), BigDecimal.valueOf(rewards));
            }
        }
    }
}
