package com.starbuds.server.service.finance.provider;

import cn.hutool.core.convert.Convert;
import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.cache.RedisFinanceKeys;
import com.starbuds.server.common.constant.cache.RedisLiveKeys;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.message.im.ImTargetType;
import com.starbuds.server.common.pojo.db.PriceConfigDO;
import com.starbuds.server.common.pojo.db.SingleChatHisDO;
import com.starbuds.server.common.pojo.finance.PriceConfig;
import com.starbuds.server.common.pojo.media.GiftPayItem;
import com.starbuds.server.common.pojo.message.im.live.AnchorIncomeChangeMsg;
import com.starbuds.server.common.pojo.user.UserCardVO;
import com.starbuds.server.common.service.core.CacheService;
import com.starbuds.server.common.service.message.ImService;
import com.starbuds.server.common.service.user.FamilyService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.common.utils.LogUtil;
import com.starbuds.server.service.finance.constant.AsyncThreadConstant;
import com.starbuds.server.service.finance.dao.PriceConfigDAO;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.beans.BeanUtils;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.math.BigDecimal;

/**
 * 结算服务
 *
 * @author Lay
 * @date 2020/1/3
 */
@Component
@Slf4j
public class SettleProvider {
    @Resource
    IWalletProvider walletProvider;

    @Reference(version = "1.0.0")
    UserService userService;

    @Reference(version = "1.0.0")
    ImService imService;

    @Reference(version = "1.0.0")
    CacheService cacheService;

    @Reference(version = "1.0.0")
    FamilyService familyService;

    @Resource
    PriceConfigDAO priceConfigDAO;

    @Resource
    FamilyWalletProvider familyWalletProvider;

    @Async(AsyncThreadConstant.DEFAULT)
    public void settleByGift(GiftPayItem giftPayItem, Long receiverUserId, Long senderUserId, Integer quantity, Long amount, Long orderId) {

        // 赠送者
        UserCardVO sender = userService.getCard(senderUserId);

        // 收入
        String remarks = String.format("[%s](%d)赠送%d个%s", sender.getUserName(), senderUserId, quantity, giftPayItem.getGiftName());

        // 礼物数量
        Integer giftNum = Convert.toInt(cacheService.hashIncreaseLong(RedisLiveKeys.LIVE_ANCHOR_GIFT, String.valueOf(receiverUserId), quantity));

        if (WalletTypeEnum.STAR_DIAMOND.getValue() == giftPayItem.getGiftPriceType()) {
            walletProvider.addBalance(WalletTypeEnum.STAR_COIN, receiverUserId, amount, remarks, orderId);

            // 直播增加收入
            Long coinIncome = cacheService.hashIncreaseLong(RedisLiveKeys.LIVE_ANCHOR_COIN, String.valueOf(receiverUserId), amount);

            // 给主播发送收入变更通知
            sendAnchorIncomeChangeMsg(WalletTypeEnum.STAR_COIN, coinIncome, receiverUserId);
        } else {
            walletProvider.addBalance(giftPayItem.getGiftPriceType(), receiverUserId, amount, remarks, orderId);
        }
    }

    @Async
    public void settleBySingleChat(SingleChatHisDO singleChatHis) {
        // 聊天者
        UserCardVO userCard = userService.getCard(singleChatHis.getUserId());

        // 收入
        String remarks = String.format("与[%s](%d)聊天收入", userCard.getUserName(), userCard.getUserId());
        if (WalletTypeEnum.STAR_DIAMOND.getValue() == singleChatHis.getHisPriceType()) {
            walletProvider.addBalance(WalletTypeEnum.STAR_COIN, singleChatHis.getTargetUserId(), singleChatHis.getHisPrice(), remarks, singleChatHis.getUserId());
        } else {
            walletProvider.addBalance(singleChatHis.getHisPriceType(), singleChatHis.getTargetUserId(), singleChatHis.getHisPrice(), remarks, singleChatHis.getUserId());
        }
    }

    /**
     * 给主播发送收入变更通知
     *
     * @param incomeType
     * @param amount
     * @param anchorUserId
     */
    public void sendAnchorIncomeChangeMsg(WalletTypeEnum incomeType, Long amount, Long anchorUserId) {
        AnchorIncomeChangeMsg anchorIncomeChangeMsg = new AnchorIncomeChangeMsg();
        anchorIncomeChangeMsg.setIncomeType(incomeType.getValue());
        anchorIncomeChangeMsg.setIncomeAmount(amount);
        imService.sendLiveRoomMsg(ImTargetType.USER, null, anchorUserId, anchorIncomeChangeMsg);
    }

    /**
     * 获取基础价格配置
     *
     * @return
     */
    public PriceConfig getPriceConfig() {
        Object cacheConfig = cacheService.get(RedisFinanceKeys.PRICE_CONFIG);
        if (cacheConfig != null) {
            return JSON.parseObject(cacheConfig.toString(), PriceConfig.class);
        }

        // 从数据库读取
        PriceConfigDO priceConfigDO = priceConfigDAO.selectLastEnabled();
        if (priceConfigDO != null) {
            PriceConfig priceConfig = new PriceConfig();
            BeanUtils.copyProperties(priceConfigDO, priceConfig);
            // 缓存
            cacheService.set(RedisFinanceKeys.PRICE_CONFIG, JSON.toJSONString(priceConfig));
            return priceConfig;
        }

        log.error("基础价格未配置");
        return null;
    }

    /**
     * 语音速配异步结算
     *
     * @param hisId
     * @param userId
     * @param targetUserId
     * @param hisPrice
     */
    @Async(AsyncThreadConstant.DEFAULT)
    public void settleByVoiceChat(Long hisId, Long userId, Long targetUserId, Long hisPrice) {
        log.info("hisId {} ,userId {} ,targetUserId {} ,hisPrice {}", hisId, userId, targetUserId, hisPrice);
        // 聊天者
        UserCardVO userCard = userService.getCard(userId);

        // 收入
        String remarks = String.format("与[%s](%d)语音速配收入", userCard.getUserName(), userCard.getUserId());

        // 陪陪
        walletProvider.addBalance(WalletTypeEnum.STAR_DUST, targetUserId, Convert.toLong(hisPrice * 0.7), remarks, userId);

        // 家族
        Long targetFamilyId = familyService.getFamilyId(targetUserId);
        if (targetFamilyId == null) {
            log.info("对方家族异常,hisId {}", hisId);
        } else {
            familyWalletProvider.addBalance(targetFamilyId, Convert.toBigDecimal(hisPrice * 0.1), remarks, targetUserId);
            log.info(LogUtil.devLog("语音速配异步结算-familyIncome-{} {} * 0.1=[{}]"),
                    targetFamilyId, hisPrice, Convert.toBigDecimal(hisPrice * 0.1));

        }
    }
}
