package com.starbuds.server.service.finance.provider;

import cn.hutool.core.convert.Convert;
import com.starbuds.server.common.constant.cache.RedisChatKeys;
import com.starbuds.server.common.constant.data.PlayTopTypeEnum;
import com.starbuds.server.common.constant.finance.PayStatusEnum;
import com.starbuds.server.common.constant.finance.SkillOrderStatusEnum;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.message.im.ContentType;
import com.starbuds.server.common.constant.message.im.ImActionType;
import com.starbuds.server.common.constant.message.push.ModuleTypeEnum;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.pojo.api.Result;
import com.starbuds.server.common.pojo.db.PlayTopDO;
import com.starbuds.server.common.pojo.db.SkillOrderDO;
import com.starbuds.server.common.pojo.message.im.biz.SystemMsg;
import com.starbuds.server.common.pojo.mq.QueueSkillEndCheckMsg;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.data.DataService;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.data.SkillService;
import com.starbuds.server.common.service.message.ImService;
import com.starbuds.server.common.utils.LogUtil;
import com.starbuds.server.service.finance.dao.FamilySkillDAO;
import com.starbuds.server.service.finance.dao.SkillOrderDAO;
import com.starbuds.server.service.finance.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.Optional;
import java.util.UUID;
import java.util.concurrent.TimeUnit;

/**
 * @Author jiangxiaowei
 * @DateTime 2020/10/23 10:53
 * @Description 技能订单
 */
@Slf4j
@Component
public class SkillOrderProvider {

    @Resource
    SkillOrderDAO skillOrderDAO;

    @Resource
    RocketProducer rocketProducer;

    @Resource
    WalletProvider walletProvider;

    @Resource
    FamilyWalletProvider familyWalletProvider;

    @Resource
    FamilySkillDAO familySkillDAO;

    @Reference(version = "1.0.0")
    SkillService skillService;

    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Reference(version = "1.0.0")
    DataService dataService;

    @Reference(version = "1.0.0")
    ImService imService;

    @Reference(version = "1.0.0")
    RedisService redisService;

    @Value("${skill.userIncomeRatio}")
    private double userIncomeRatio;

    @Value("${skill.familyIncomeRatio}")
    private double familyIncomeRatio;

    @PostConstruct
    private void init() {
        log.info("陪陪收入分配：个人 {} 家族 {}", userIncomeRatio, familyIncomeRatio);
    }

    /**
     * 支付
     *
     * @param orderId
     * @param payTime
     * @return
     */
    public boolean doPay(Long orderId, Long payTime) {
        SkillOrderDO updateOrder = new SkillOrderDO();
        updateOrder.setOrderId(orderId);
        updateOrder.setPayStatus(PayStatusEnum.SUCCESSED.getValue());
        updateOrder.setOrderStatus(SkillOrderStatusEnum.WAIT_RECEIVE.getValue());
        updateOrder.setPayTime(payTime);
        updateOrder.setUpdateTime(payTime);
        return skillOrderDAO.updateById(updateOrder, PayStatusEnum.WAITTING.getValue(), SkillOrderStatusEnum.ORDER_CREATE.getValue()) > 0;
    }

    /**
     * 是否等待接单
     *
     * @param orderId
     * @return
     */
    public boolean isWaitReceive(Long orderId) {
        return Optional.ofNullable(skillOrderDAO.selectOrderStatus(orderId)).map(e -> e == SkillOrderStatusEnum.WAIT_RECEIVE.getValue()).orElse(false);
    }


    /**
     * 订单是否拒绝
     *
     * @param orderId
     * @return
     */
    public boolean isOrderReject(Long orderId) {
        return Optional.ofNullable(skillOrderDAO.selectOrderStatus(orderId)).map(e -> e == SkillOrderStatusEnum.REJECT.getValue()).orElse(false);
    }

    /**
     * 订单是否接单
     *
     * @param orderId
     * @return
     */
    public boolean isOrderReceive(Long orderId) {
        return Optional.ofNullable(skillOrderDAO.selectOrderStatus(orderId)).map(e -> e == SkillOrderStatusEnum.WAIT_SERVICE.getValue()).orElse(false);
    }

    /**
     * 订单是否开始服务
     *
     * @param orderId
     * @return
     */
    public boolean isStartService(Long orderId) {
        return Optional.ofNullable(skillOrderDAO.selectOrderStatus(orderId)).map(e -> e == SkillOrderStatusEnum.WAIT_SERVICE.getValue()).orElse(true);
    }

    /**
     * 订单是否结束
     *
     * @param orderId
     * @return
     */
    public boolean isOrderEnd(Long orderId) {
        return Optional.ofNullable(skillOrderDAO.selectOrderStatus(orderId)).map(e -> e >= SkillOrderStatusEnum.COMPLETE.getValue()).orElse(false);
    }

    /**
     * 订单是否已取消
     *
     * @param orderId
     * @return
     */
    public boolean isOrderCancel(Long orderId) {
        return Optional.ofNullable(skillOrderDAO.selectOrderStatus(orderId)).map(e -> e == SkillOrderStatusEnum.CANCEL.getValue()).orElse(false);
    }

    /**
     * 订单是否已开始
     *
     * @param orderId
     * @return
     */
    public boolean isOrderCanStart(Long orderId) {
        return Optional.ofNullable(skillOrderDAO.selectOrderStatus(orderId)).map(e -> e == SkillOrderStatusEnum.WAIT_SERVICE.getValue()).orElse(false);
    }

    /**
     * 订单是否等待支付
     *
     * @param orderId
     * @return
     */
    public boolean isWaitPay(Long orderId) {
        return Optional.ofNullable(skillOrderDAO.selectPayStatus(orderId)).map(e -> e == PayStatusEnum.WAITTING.getValue()).orElse(false);
    }

    /**
     * 开始服务
     *
     * @param orderId
     * @return
     */
    public boolean doStartService(Long orderId, Long startTime) {
        if (!isOrderCanStart(orderId)) {
            log.warn("订单不支持开始服务");
            return false;
        }
        log.info("订单{}开始服务", orderId);
        SkillOrderDO orderDO = new SkillOrderDO();
        orderDO.setOrderId(orderId);
        orderDO.setOrderStatus(SkillOrderStatusEnum.IN_SERVICE.getValue());
        orderDO.setStartTime(startTime);
        orderDO.setUpdateTime(startTime);
        if (skillOrderDAO.updateById(orderDO, PayStatusEnum.SUCCESSED.getValue(), SkillOrderStatusEnum.WAIT_SERVICE.getValue()) > 0) {
            String key = RedisChatKeys.getSkillOrderStartKey(orderId);
            String value = UUID.randomUUID().toString();
            try {
                boolean lock = redisService.tryLock(key, value, 10, TimeUnit.SECONDS);
                if (lock) {
                    Long deliverTime = startTime + 30 * 60 * 1000L;
                    QueueSkillEndCheckMsg checkMsg = new QueueSkillEndCheckMsg();
                    checkMsg.setOrderId(orderId);
                    checkMsg.setEndTime(deliverTime);
                    checkMsg.setUserId(orderDO.getUserId());
                    checkMsg.setReceiveUserId(orderDO.getReceiveUserId());
                    rocketProducer.sendMsg(RocketTopic.TPC_DELAY, RocketTag.SKILL_SERVICE_END_CHECK, checkMsg.serialize(), deliverTime);
                    log.info("订单{}服务开始成功", orderId);
                    return true;
                }
            } catch (Exception e) {
                log.error("出现错误", e);
                TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            } finally {
                redisService.releaseLock(key, value);
            }
        }
        log.warn("更新状态失败，操作过于频繁");
        return false;
    }

    /**
     * 结束服务
     *
     * @param orderId
     * @param endTime
     * @return
     */
    public Result doConfirmService(Long orderId, Long endTime) {
        Result result = new Result();
        SkillOrderDO skillOrder = skillOrderDAO.selectById(orderId);
        if (skillOrder == null) {
            result.setSuccess(false);
            result.setMsg("订单不存在");
            return result;
        }
        if (skillOrder.getOrderStatus() != SkillOrderStatusEnum.IN_SERVICE.getValue()) {
            result.setSuccess(false);
            result.setMsg("仅服务中才可确认完成");
            return result;
        }
        if (skillOrder.getPayStatus() != PayStatusEnum.SUCCESSED.getValue()) {
            result.setSuccess(false);
            result.setMsg("仅支付成功才可确认完成");
            return result;
        }
        if (!skillService.isSkillEnable(skillOrder.getReceiveFamilyId(), skillOrder.getReceiveUserId(), skillOrder.getSkillId())) {
            result.setSuccess(false);
            result.setMsg("暂不支持完成");
            return result;
        }
        log.info("订单{}正在结束服务", orderId);
        SkillOrderDO orderDO = new SkillOrderDO();
        orderDO.setOrderId(orderId);
        orderDO.setOrderStatus(SkillOrderStatusEnum.COMPLETE.getValue());
        orderDO.setEndTime(endTime);
        orderDO.setUpdateTime(endTime);
        if (skillOrderDAO.updateById(orderDO, PayStatusEnum.SUCCESSED.getValue(), SkillOrderStatusEnum.IN_SERVICE.getValue()) > 0) {
            String key = RedisChatKeys.getSkillOrderConfirmKey(orderId);
            String value = UUID.randomUUID().toString();
            try {
                boolean lock = redisService.tryLock(key, value, 10, TimeUnit.SECONDS);
                if (lock) {
                    SkillOrderDO order = skillOrderDAO.selectById(orderId);
                    Long payAmount = order.getPayAmount();
                    Long receiveUserId = order.getReceiveUserId();
                    Long receiveFamilyId = order.getReceiveFamilyId();
                    BigDecimal userIncome = Convert.toBigDecimal(payAmount).multiply(Convert.toBigDecimal(userIncomeRatio)).setScale(0, RoundingMode.DOWN);
                    BigDecimal familyIncome = Convert.toBigDecimal(payAmount).multiply(Convert.toBigDecimal(familyIncomeRatio)).setScale(0, RoundingMode.DOWN);

                    String hisRemarks = String.format("[%s]接单", skillService.getSkillName(order.getSkillId()));

                    // 陪陪0.7  家族0.2  平台0.1
                    // 陪陪、家族剩余部分是平台收入
                    log.info("陪陪订单金额：{} 收入分配：陪陪 [{}, {}] 家族 [{}, {}] 平台 {}",
                            payAmount, userIncome, userIncomeRatio, familyIncome, familyIncomeRatio,
                            (payAmount - userIncome.longValue() - familyIncome.longValue()));
                    walletProvider.addBalance(WalletTypeEnum.STAR_DUST, receiveUserId, Convert.toLong(userIncome), hisRemarks, orderId);

                    // 家族
                    familyWalletProvider.addBalance(receiveFamilyId, familyIncome, hisRemarks, orderId);
                    log.info(LogUtil.devLog("技能结束服务-familyIncome-{} {} * {}=[{}]"),
                            receiveFamilyId, payAmount, familyIncomeRatio, familyIncome);

                    // 添加陪玩头条
                    PlayTopDO topDO = new PlayTopDO();
                    topDO.setTopId(identityService.getId());
                    topDO.setTopType(PlayTopTypeEnum.ORDER_COMPLETED.getCode());
                    topDO.setUserId(receiveUserId);
                    topDO.setIncome(Convert.toLong(userIncome));
                    topDO.setOrderId(orderId);
                    topDO.setSkillId(order.getSkillId());
                    topDO.setCreateTime(endTime);
                    dataService.addPlayTop(topDO);

                    // 系统消息
                    SystemMsg systemMsg = new SystemMsg();
                    systemMsg.setContentType(ContentType.TEXT);
                    systemMsg.setTitle("订单已完成");
                    String body = String.format("你有一条【%s*%d】的订单已完成，可以再来一局！", skillService.getSkillName(orderDO.getSkillId()), orderDO.getQuantity());
                    systemMsg.setBody(body);
                    systemMsg.setActionType(ImActionType.JUMP_NATIVE.getValue());
                    systemMsg.setModuleType(ModuleTypeEnum.ORDER_DETAILS.getDesc());
                    systemMsg.setModuleId(Convert.toStr(orderId));
                    imService.sendCustomImMsg(null, orderDO.getUserId(), systemMsg, null);
                    log.info("订单{}结束服务成功", orderId);
                    result.setSuccess(true);
                    return result;
                }
            } catch (Exception e) {
                log.error("出现错误", e);
                TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            } finally {
                redisService.releaseLock(key, value);
            }
        }
        result.setSuccess(false);
        result.setMsg("更新状态失败，或操作过于频繁");
        return result;
    }

    /**
     * 取消订单
     *
     * @param orderId
     * @return
     */
    public boolean cancelOrder(Long orderId) {
        if (isOrderCancel(orderId)) {
            log.warn("订单已取消");
            return false;
        }
        log.info("订单{}正在被取消", orderId);
        long curTime = System.currentTimeMillis();
        SkillOrderDO orderDO = new SkillOrderDO();
        orderDO.setOrderId(orderId);
        orderDO.setPayStatus(PayStatusEnum.CANCELED.getValue());
        orderDO.setOrderStatus(SkillOrderStatusEnum.CANCEL.getValue());
        orderDO.setCancelTime(curTime);
        orderDO.setUpdateTime(curTime);
        log.info("订单{}取消成功", orderId);
        return skillOrderDAO.updateById(orderDO, PayStatusEnum.WAITTING.getValue(), SkillOrderStatusEnum.ORDER_CREATE.getValue()) > 0;
    }

    /**
     * 终止订单
     *
     * @param orderId
     * @return
     */
    public boolean stopOrder(Long orderId, boolean needRefund) {
        if (isOrderReject(orderId)) {
            log.warn("重复终止");
            return false;
        }
        log.info("订单{}正在被终止", orderId);
        long curTime = System.currentTimeMillis();
        SkillOrderDO orderDO = new SkillOrderDO();
        orderDO.setOrderId(orderId);
        orderDO.setOrderStatus(SkillOrderStatusEnum.REJECT.getValue());
        orderDO.setRejectTime(curTime);
        orderDO.setStopTime(curTime);
        orderDO.setUpdateTime(curTime);
        orderDO.setStopReason("超时未接单");
        if (skillOrderDAO.updateById(orderDO, PayStatusEnum.SUCCESSED.getValue(), SkillOrderStatusEnum.WAIT_RECEIVE.getValue()) > 0) {
            // 退款
            if (needRefund) {
                SkillOrderDO order = skillOrderDAO.selectById(orderId);
                String hisRemarks = String.format("[%s]退款", skillService.getSkillName(order.getSkillId()));
                walletProvider.addBalance(WalletTypeEnum.STAR_DIAMOND, order.getUserId(), order.getPayAmount(), hisRemarks, orderId);
            }
            log.info("订单{}终止成功", orderId);
            return true;
        }
        return false;
    }

    /**
     * 拒绝订单
     *
     * @param orderId
     * @param rejectReason
     * @return
     */
    public boolean doRejectOrder(Long orderId, String rejectReason) {
        log.info("订单{}被拒绝", orderId);
        if (isOrderReject(orderId)) {
            log.warn("重复拒单");
            return false;
        }

        long curTime = System.currentTimeMillis();
        SkillOrderDO orderDO = new SkillOrderDO();
        orderDO.setOrderId(orderId);
        orderDO.setOrderStatus(SkillOrderStatusEnum.REJECT.getValue());
        orderDO.setUpdateTime(curTime);
        orderDO.setRejectTime(curTime);
        orderDO.setRejectReason(rejectReason);
        if (skillOrderDAO.updateById(orderDO, PayStatusEnum.SUCCESSED.getValue(), SkillOrderStatusEnum.WAIT_RECEIVE.getValue()) > 0) {
            String key = RedisChatKeys.getSkillOrderRejectKey(orderId);
            String value = UUID.randomUUID().toString();
            try {
                boolean lock = redisService.tryLock(key, value, 10, TimeUnit.SECONDS);
                if (lock) {
                    // 退款
                    SkillOrderDO order = skillOrderDAO.selectById(orderId);
                    String hisRemarks = String.format("[%s]退款", skillService.getSkillName(order.getSkillId()));
                    walletProvider.addBalance(WalletTypeEnum.STAR_DIAMOND, order.getUserId(), order.getPayAmount(), hisRemarks, orderId);
                    return true;
                }
            } catch (Exception e) {
                log.error("出现错误", e);
                TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            } finally {
                redisService.releaseLock(key, value);
            }
        }
        log.warn("更新状态失败，或操作过于频繁");
        return false;
    }

    /**
     * 接收订单
     *
     * @param orderId
     * @return
     */
    public boolean doReceiveOrder(Long orderId) {
        log.info("订单{}已接单", orderId);
        long curTime = System.currentTimeMillis();
        if (isOrderReceive(orderId)) {
            log.warn("重复接单");
            return false;
        }

        SkillOrderDO orderUpdate = new SkillOrderDO();
        orderUpdate.setOrderId(orderId);
        orderUpdate.setOrderStatus(SkillOrderStatusEnum.WAIT_SERVICE.getValue());
        orderUpdate.setUpdateTime(curTime);
        orderUpdate.setReceiveTime(curTime);
        if (skillOrderDAO.updateById(orderUpdate, PayStatusEnum.SUCCESSED.getValue(), SkillOrderStatusEnum.WAIT_RECEIVE.getValue()) > 0) {
            String key = RedisChatKeys.getSkillOrderReceiveKey(orderId);
            String value = UUID.randomUUID().toString();
            try {
                boolean lock = redisService.tryLock(key, value, 10, TimeUnit.SECONDS);
                if (lock) {
                    // 增加接单量
                    SkillOrderDO orderDO = skillOrderDAO.selectById(orderId);
                    familySkillDAO.increaseOrderCount(orderDO.getReceiveFamilyId(), orderDO.getReceiveUserId(), orderDO.getSkillId(), orderDO.getQuantity());
                    return true;
                }
            } catch (Exception e) {
                log.error("出现错误", e);
                TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            } finally {
                redisService.releaseLock(key, value);
            }
        }
        log.warn("更新状态失败，或操作过于频繁");
        return false;
    }

}
