package com.starbuds.server.service.finance.provider;

import cn.hutool.core.convert.Convert;
import com.starbuds.server.common.constant.activity.TriggerEventEnum;
import com.starbuds.server.common.constant.common.CommonStatus;
import com.starbuds.server.common.constant.common.TimeConstants;
import com.starbuds.server.common.constant.finance.ChatTimeWalletType;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.pojo.activity.VipNewEventParam;
import com.starbuds.server.common.pojo.activity.VipRenewEventParam;
import com.starbuds.server.common.pojo.db.UserVipDO;
import com.starbuds.server.common.pojo.db.VipFreeMinutesDayHisDO;
import com.starbuds.server.common.pojo.db.VipOrderDO;
import com.starbuds.server.common.service.activity.TriggerService;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.service.finance.dao.UserVipDAO;
import com.starbuds.server.service.finance.dao.VipFreeMinutesDayHisDAO;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;

/**
 * 会员服务
 *
 * @author Lay
 * @date 2020/4/4
 */
@Component
@Slf4j
public class VipProvider {

    @Reference(version = "1.0.0")
    TriggerService triggerService;

    @Reference(version = "1.0.0")
    StatPlatformService statPlatformService;

    @Resource
    ChatTimeWalletProvider chatTimeWalletProvider;

    @Resource
    UserVipDAO userVipDAO;

    @Resource
    VipFreeMinutesDayHisDAO vipFreeMinutesDayHisDAO;

    /**
     * 获取用户 vip
     *
     * @param userId
     * @return
     */
    public UserVipDO getUserVip(Long userId) {
        return userVipDAO.selectById(userId);
    }

    /**
     * 支付逻辑
     *
     * @param userId
     * @param vipOrder
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public boolean handlePay(Long userId, VipOrderDO vipOrder) {
        UserVipDO userVip = userVipDAO.selectById(userId);
        boolean isNew = true;
        if (userVip != null) {
            if (userVip.getVipExpireTime() == -1 || userVip.getVipExpireTime() > System.currentTimeMillis()) {
                isNew = false;
            }
        }

        if (isNew) {
            // 计算到期时间
            Long expireTime = -1L;
            if (vipOrder.getVipDays() > 0) {
                Calendar calendar = Calendar.getInstance();
                calendar.add(Calendar.DATE, vipOrder.getVipDays());
                calendar.set(Calendar.HOUR_OF_DAY, 23);
                calendar.set(Calendar.MINUTE, 59);
                calendar.set(Calendar.SECOND, 59);
                calendar.set(Calendar.MILLISECOND, 999);
                expireTime = calendar.getTimeInMillis();
            }

            userVip = new UserVipDO();
            userVip.setUserId(userId);
            userVip.setVipCreateTime(System.currentTimeMillis());
            userVip.setVipExpireTime(expireTime);
            userVip.setVipLevel(vipOrder.getVipLevel());
            userVip.setStatus(CommonStatus.Normal);
            userVip.setSingleChatMinutesPerDay(vipOrder.getSingleChatMinutesPerDay());
            userVipDAO.insertOrUpdate(userVip);

            // 触发开通会员事件
            VipNewEventParam vipNewEventParam = new VipNewEventParam();
            vipNewEventParam.setUserId(userId);
            vipNewEventParam.setVipDays(vipOrder.getVipDays());
            vipNewEventParam.setVipLevel(vipOrder.getVipLevel());
            triggerService.emitEvent(TriggerEventEnum.VIP_NEW, vipNewEventParam);

            // 赠送时长
            chatTimeWalletProvider.addMinutes(userId, ChatTimeWalletType.SINGLE_CHAT, expireTime, vipOrder.getSingleChatMinutes(), "开通会员赠送时长", String.valueOf(vipOrder.getOrderId()));

            // 赠送每日时长
            getDayFreeMinutes(userId);

            // 统计
            statPlatformService.addPlatformStat(vipOrder.getChannelId(), StatPlatformTypeEnum.STAT_VIP_NEW, BigDecimal.ONE);
            statPlatformService.addPlatformStat(vipOrder.getChannelId(), StatPlatformTypeEnum.STAT_VIP_NEW_AMOUNT, vipOrder.getVipPrice());
        } else {
            Long expireTime = -1L;
            if (vipOrder.getVipDays() == -1) {
                userVip.setVipExpireTime(-1L);
            } else {
                // 原来不是永久会员才需要加时间
                if (userVip.getVipExpireTime() > 0) {
                    expireTime = userVip.getVipExpireTime() + vipOrder.getVipDays() * TimeConstants.DAY_MILLS;
                    userVip.setVipExpireTime(expireTime);
                }
            }
            userVip.setVipLevel(Math.max(userVip.getVipLevel(), vipOrder.getVipLevel()));
            userVip.setStatus(CommonStatus.Normal);
            userVipDAO.insertOrUpdate(userVip);

            // 赠送时长
            chatTimeWalletProvider.addMinutes(userId, ChatTimeWalletType.SINGLE_CHAT, expireTime, vipOrder.getSingleChatMinutes(), "开通会员赠送时长", String.valueOf(vipOrder.getOrderId()));

            // 触发续费会员事件
            VipRenewEventParam vipRenewEventParam = new VipRenewEventParam();
            vipRenewEventParam.setUserId(userId);
            vipRenewEventParam.setVipDays(vipOrder.getVipDays());
            vipRenewEventParam.setVipLevel(vipOrder.getVipLevel());
            triggerService.emitEvent(TriggerEventEnum.VIP_RENEW, vipRenewEventParam);

            // 统计
            statPlatformService.addPlatformStat(vipOrder.getChannelId(), StatPlatformTypeEnum.STAT_VIP_RENEW, BigDecimal.ONE);
            statPlatformService.addPlatformStat(vipOrder.getChannelId(), StatPlatformTypeEnum.STAT_VIP_RENEW, vipOrder.getVipPrice());
        }

        return true;
    }

    @Transactional(rollbackFor = Exception.class)
    public boolean getDayFreeMinutes(Long userId) {
        // 领取每日免费分钟数
        UserVipDO userVipDO = userVipDAO.selectById(userId);
        if (userVipDO == null) {
            return false;
        }

        // 判断是否发放过
        SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd");
        Integer date = Convert.toInt(sdf.format(new Date()));
        if (vipFreeMinutesDayHisDAO.countById(date, userId) > 0) {
            return false;
        }

        // 尝试发放
        try {
            Calendar calendar = Calendar.getInstance();
            calendar.set(Calendar.HOUR_OF_DAY, 23);
            calendar.set(Calendar.MINUTE, 59);
            calendar.set(Calendar.SECOND, 59);
            calendar.set(Calendar.MILLISECOND, 999);

            // 发放记录
            VipFreeMinutesDayHisDO vipFreeMinutesDayHis = new VipFreeMinutesDayHisDO();
            vipFreeMinutesDayHis.setUserId(userId);
            vipFreeMinutesDayHis.setHisDate(date);
            vipFreeMinutesDayHis.setMinutes(userVipDO.getSingleChatMinutesPerDay());
            vipFreeMinutesDayHis.setHisCreateTime(System.currentTimeMillis());
            vipFreeMinutesDayHisDAO.insert(vipFreeMinutesDayHis);

            // 增加时长
            chatTimeWalletProvider.addMinutes(userId, ChatTimeWalletType.SINGLE_CHAT, calendar.getTimeInMillis(), userVipDO.getSingleChatMinutesPerDay(), "会员每日免费时长发放", String.valueOf(vipFreeMinutesDayHis.getHisDate()));
            return true;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("免费时长发放失败", e);
        }
        return false;
    }
}
