package com.starbuds.server.service.finance.provider;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.crypto.SecureUtil;
import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.pojo.db.UserWalletDO;
import com.starbuds.server.common.pojo.db.UserWalletHisDO;
import com.starbuds.server.common.pojo.db.VirtualCoinDO;
import com.starbuds.server.common.pojo.finance.UserWalletHisVO;
import com.starbuds.server.common.pojo.finance.UserWalletVO;
import com.starbuds.server.common.pojo.mq.QueueWalletBalanceReduceMsg;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.common.utils.LogUtil;
import com.starbuds.server.service.finance.dao.UserWalletDAO;
import com.starbuds.server.service.finance.dao.UserWalletHisDAO;
import com.starbuds.server.service.finance.dao.VirtualCoinDAO;
import com.starbuds.server.service.finance.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.*;
import java.util.stream.Collectors;

import static com.starbuds.server.service.finance.util.SpringContextUtil.isDev;

/**
 * 钱包
 *
 * @author Lay
 * @date 2019/12/20
 */
@Component
@Slf4j
public class WalletProvider implements IWalletProvider {

    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Reference(version = "1.0.0")
    StatPlatformService statPlatformService;

    @Resource
    UserWalletDAO userWalletDAO;

    @Resource
    UserWalletHisDAO userWalletHisDAO;

    @Resource
    VirtualCoinDAO virtualCoinDAO;

    @Resource
    RocketProducer rocketProducer;

    public static String getWalletId(long userId, int type) {
        return SecureUtil.md5(String.format("%d:%d", userId, type));
    }

    public static WalletTypeEnum getWalletType(String walletId, long userId) {
        for (WalletTypeEnum walletType : WalletTypeEnum.values()) {
            if (walletId.equals(getWalletId(userId, walletType.getValue()))) {
                return walletType;
            }
        }
        return null;
    }

    /**
     * 增加余额
     *
     * @param walletType
     * @param userId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Override
    public Long addBalance(WalletTypeEnum walletType, long userId, long balanceValue, String hisRemarks, Long orderId) {
        return addBalance(walletType.getValue(), userId, balanceValue, hisRemarks, orderId);
    }

    /**
     * 增加余额
     *
     * @param walletType
     * @param userId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Override
    public Long addBalance(Integer walletType, long userId, long balanceValue, String hisRemarks, Long orderId) {
        return changeBalance(walletType, userId, Math.abs(balanceValue), hisRemarks, orderId);
    }

    /**
     * 减少余额
     *
     * @param walletType
     * @param userId
     * @param balanceValue
     * @param hisRemarks
     * @return
     */
    @Override
    public Long reduceBalance(WalletTypeEnum walletType, long userId, long balanceValue, String hisRemarks, Long orderId) {
        return reduceBalance(walletType.getValue(), userId, balanceValue, hisRemarks, orderId);
    }

    /**
     * 减少余额
     *
     * @param walletType
     * @param userId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Override
    public Long reduceBalance(Integer walletType, long userId, long balanceValue, String hisRemarks, Long orderId) {
        return changeBalance(walletType, userId, -Math.abs(balanceValue), hisRemarks, orderId);
    }

    /**
     * 改变余额
     *
     * @param userId
     * @param walletType
     * @param balanceValue
     * @return
     */
    public Long changeBalance(Integer walletType, long userId, long balanceValue, String hisRemarks, Long orderId) {
        String walletId = getWalletId(userId, walletType);

        // 判断账户是否存在
        UserWalletDO userWalletDO = userWalletDAO.selectById(walletId);
        if (userWalletDO == null) {
            userWalletDO = new UserWalletDO();
            userWalletDO.setWalletId(walletId);
            userWalletDO.setUserId(userId);
            userWalletDO.setType(walletType);
            userWalletDO.setBalance(0L);
            userWalletDO.setVersion(0L);
            userWalletDO.setUpdateTime(System.currentTimeMillis());
            try {
                userWalletDAO.insert(userWalletDO);
            } catch (Exception e) {
                // 重试
                return changeBalance(walletType, userId, balanceValue, hisRemarks, orderId);
            }
        }

        return executeChange(userWalletDO, walletType, userId, balanceValue, hisRemarks, orderId);
    }

    /**
     * 执行改变
     *
     * @param userWalletDO
     * @param walletType
     * @param userId
     * @param balanceValue
     * @param hisRemarks
     * @param orderId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public Long executeChange(UserWalletDO userWalletDO, Integer walletType, long userId, long balanceValue, String hisRemarks, Long orderId) {
        Long balanceNew;

        try {
            // 计算新余额
            Long balanceOld = userWalletDO.getBalance();
            Long version = userWalletDO.getVersion();
            balanceNew = balanceOld + balanceValue;

            // 余额不足
            if (balanceNew < 0) {
                return balanceNew;
            }

            long curTime = System.currentTimeMillis();

            // 更新
            try {
                int updated = userWalletDAO.updateCas(userWalletDO.getWalletId(), balanceNew, version);
                if (updated == 0) {
                    userWalletDO = userWalletDAO.selectById(userWalletDO.getWalletId());
                    // 重试
                    return executeChange(userWalletDO, walletType, userId, balanceValue, hisRemarks, orderId);
                }
            } catch (Exception e) {
                log.error("{} 更新失败，重试", userWalletDO.getWalletId());
            }

            // 写入变更记录
            UserWalletHisDO userWalletHisDO = new UserWalletHisDO();
            userWalletHisDO.setHisId(identityService.getId());
            userWalletHisDO.setWalletId(userWalletDO.getWalletId());
            userWalletHisDO.setUserId(userId);
            userWalletHisDO.setType(walletType);
            userWalletHisDO.setBalanceOld(balanceOld);
            userWalletHisDO.setBalanceValue(balanceValue);
            userWalletHisDO.setBalanceNew(balanceNew);
            userWalletHisDO.setCreateTime(curTime);
            userWalletHisDO.setHisRemarks(hisRemarks);
            userWalletHisDO.setOrderId(orderId);
            int insert = userWalletHisDAO.insert(userWalletHisDO);
            if (isDev()) {
                log.info("写入用户钱包变更记录 insert={} his={}", insert, LogUtil.obj2JsonObj(userWalletHisDO));
            }
        } catch (Exception e) {
            log.error("余额变动失败", e);
            balanceNew = null;
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
        }

        if (balanceNew != null) {
            // 异步统计
            try {
                StatPlatformTypeEnum statPlatformType = null;
                boolean isAdd = balanceValue > 0;
                WalletTypeEnum walletTypeEnum = WalletTypeEnum.toEnum(walletType);
                if (walletTypeEnum != null) {
                    switch (walletTypeEnum) {
                        case STAR_DIAMOND:
                            statPlatformType = isAdd ? StatPlatformTypeEnum.STAR_DIAMOND_ADD : StatPlatformTypeEnum.STAR_DIAMOND_REDUCE;
                            break;
                        case STAR_YUAN:
                            statPlatformType = isAdd ? StatPlatformTypeEnum.STAR_YUAN_ADD : StatPlatformTypeEnum.STAR_YUAN_REDUCE;
                            break;
                        case STAR_COIN:
                            statPlatformType = isAdd ? StatPlatformTypeEnum.STAR_COIN_ADD : StatPlatformTypeEnum.STAR_COIN_REDUCE;
                            break;
                        case STAR_DUST:
                            statPlatformType = isAdd ? StatPlatformTypeEnum.STAR_DUST_ADD : StatPlatformTypeEnum.STAR_DUST_REDUCE;
                            break;
                        case PINK_DIAMOND:
                            statPlatformType = isAdd ? StatPlatformTypeEnum.PINK_DIAMOND_ADD : StatPlatformTypeEnum.PINK_DIAMOND_REDUCE;
                            break;
                        case GOLD_COIN:
                            statPlatformType = isAdd ? StatPlatformTypeEnum.GOLD_COIN_ADD : StatPlatformTypeEnum.GOLD_COIN_REDUCE;
                            break;
                        default:
                            break;
                    }

                    statPlatformService.addPlatformStat(0L, statPlatformType, BigDecimal.valueOf(Math.abs(balanceValue)));
                }
            } catch (Exception e) {
                log.error("统计失败", e);
            }
        }

        return balanceNew;
    }

    /**
     * 获取用户钱包
     *
     * @param userId
     * @return
     */
    @Override
    public UserWalletVO getUserWallet(Long userId) {
        List<Integer> walletTypes = Arrays.stream(WalletTypeEnum.values()).map(WalletTypeEnum::getValue).collect(Collectors.toList());
        return getUserWallet(userId, walletTypes);
    }

    /**
     * 获取用户虚拟币钱包
     *
     * @param userId
     * @return
     */
    @Override
    public UserWalletVO getUserVirtualWallet(Long userId) {
        List<VirtualCoinDO> list = virtualCoinDAO.selectAll();
        if (CollectionUtil.isNotEmpty(list)) {
            return getUserWallet(userId, list.stream().map(VirtualCoinDO::getCoinId).collect(Collectors.toList()));
        }
        UserWalletVO userWallet = new UserWalletVO();
        userWallet.setWalletItems(Collections.emptyList());
        return userWallet;
    }

    private UserWalletVO getUserWallet(Long userId, List<Integer> walletTypes) {
        UserWalletVO userWalletVO = new UserWalletVO();

        List<String> walletIds = new ArrayList<>();
        List<UserWalletVO.Item> items = new ArrayList<>();
        for (Integer walletType : walletTypes) {
            if (WalletTypeEnum.PINK_DIAMOND_TOTAL.getValue() - walletType == 0) {
                // 粉钻总量（实时计算的数据，没有存储；粉钻-不可提现 + 粉钻-可提现）
                continue;
            }
            walletIds.add(WalletProvider.getWalletId(userId, walletType));
        }

        List<UserWalletDO> userWalletDOList = userWalletDAO.selectByIds(walletIds);

        Long dustAmount = 0L;
        Long pinkAmount = 0L;
        if (!CollectionUtil.isEmpty(userWalletDOList)) {
            for (UserWalletDO userWalletDO : userWalletDOList) {
                UserWalletVO.Item item = new UserWalletVO.Item();
                item.setWalletType(userWalletDO.getType());
                item.setBalance(userWalletDO.getBalance());
                items.add(item);
                walletTypes.remove(item.getWalletType());

                if (WalletTypeEnum.STAR_DUST.getValue() - userWalletDO.getType() == 0) {
                    dustAmount = userWalletDO.getBalance();
                }
                if (WalletTypeEnum.PINK_DIAMOND.getValue() - userWalletDO.getType() == 0) {
                    pinkAmount = userWalletDO.getBalance();
                }
            }
        }

        if (CollectionUtil.isNotEmpty(walletTypes)) {
            // 补回去
            for (Integer walletType : walletTypes) {
                if (WalletTypeEnum.PINK_DIAMOND_TOTAL.getValue() - walletType == 0) {
                    // 粉钻总量（实时计算的数据，没有存储；粉钻-不可提现 + 粉钻-可提现）
                    continue;
                }
                UserWalletVO.Item item = new UserWalletVO.Item();
                item.setWalletType(walletType);
                item.setBalance(0L);
                items.add(item);
            }
        }

        // 粉钻总量（实时计算的数据，没有存储；粉钻-不可提现 + 粉钻-可提现）
        if (!CollectionUtil.isEmpty(items)) {
            dustAmount = dustAmount == null ? 0 : dustAmount;
            pinkAmount = pinkAmount == null ? 0 : pinkAmount;

            Long pinkTotal = dustAmount + pinkAmount;
            UserWalletVO.Item item = new UserWalletVO.Item();
            item.setWalletType(WalletTypeEnum.PINK_DIAMOND_TOTAL.getValue());
            item.setBalance(Math.abs(pinkTotal));
            items.add(item);
        }

        userWalletVO.setWalletItems(items);

        return userWalletVO;
    }

    /**
     * 获取用户钱包
     *
     * @param walletId
     * @return
     */
    @Override
    public UserWalletDO getUserWallet(String walletId) {
        return userWalletDAO.selectById(walletId);
    }

    /**
     * 获取余额
     *
     * @param userId
     * @param walletType
     * @return
     */
    @Override
    public Long getBalance(Long userId, WalletTypeEnum walletType) {
        return getBalance(userId, walletType.getValue());
    }

    /**
     * 获取余额
     *
     * @param userId
     * @param walletType
     * @return
     */
    @Override
    public Long getBalance(Long userId, Integer walletType) {
        UserWalletDO userWalletDO = userWalletDAO.selectById(getWalletId(userId, walletType));
        if (userWalletDO != null) {
            return userWalletDO.getBalance();
        }
        return 0L;
    }

    /**
     * 创建用户钱包
     *
     * @param userId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public boolean createUserWallet(Long userId) {
        try {
            UserWalletDO userWalletDO = new UserWalletDO();
            userWalletDO.setUserId(userId);
            userWalletDO.setBalance(0L);
            userWalletDO.setVersion(0L);
            userWalletDO.setUpdateTime(System.currentTimeMillis());
            for (WalletTypeEnum walletTypeEnum : WalletTypeEnum.values()) {
                userWalletDO.setWalletId(getWalletId(userId, walletTypeEnum.getValue()));
                userWalletDO.setType(walletTypeEnum.getValue());
                userWalletDAO.insert(userWalletDO);
            }
            return true;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("用户钱包创建失败", e);
        }
        return false;
    }

    /**
     * 获取用户钱包历史记录
     *
     * @param userId
     * @param lastId
     * @param pageSize
     * @param walletType
     * @param isIncome
     * @param beginTime
     * @param endTime
     * @return
     */
    @Override
    public List<UserWalletHisVO> getUserWalletHisList(Long userId, Long lastId, int pageSize, Integer walletType, Integer isIncome, Long beginTime, Long endTime) {
        if (isIncome == null) {
            isIncome = BoolIntVal.Ignore;
        }

        Set<String> walletIds = new HashSet<>();
        if (walletType == WalletTypeEnum.PINK_DIAMOND_TOTAL.getValue()) {
            walletIds.add(WalletProvider.getWalletId(userId, WalletTypeEnum.STAR_DUST.getValue()));
            walletIds.add(WalletProvider.getWalletId(userId, WalletTypeEnum.PINK_DIAMOND.getValue()));
        } else {
            walletIds.add(getWalletId(userId, walletType));
        }

        List<UserWalletHisDO> userWalletDOList = userWalletHisDAO.selectAll(walletIds, lastId, pageSize, isIncome, beginTime, endTime);
        if (CollectionUtil.isNotEmpty(userWalletDOList)) {
            List<UserWalletHisVO> userWalletHisVOList = new ArrayList<>();
            for (Iterator<UserWalletHisDO> iterator = userWalletDOList.iterator(); iterator.hasNext(); ) {
                UserWalletHisVO userWalletHisVO = new UserWalletHisVO();
                BeanUtils.copyProperties(iterator.next(), userWalletHisVO);
                userWalletHisVOList.add(userWalletHisVO);
            }
            return userWalletHisVOList;
        }

        return Collections.emptyList();
    }

    /**
     * 求和
     *
     * @param userId
     * @param walletType
     * @param isIncome
     * @param beginTime
     * @param endTime
     * @return
     */
    @Override
    public Long sumByTypeAndTime(Long userId, Integer walletType, Integer isIncome, Long beginTime, Long endTime) {
        WalletTypeEnum walletTypeEnum = WalletTypeEnum.toEnum(walletType);
        // 粉钻总量（实时计算的数据，没有存储；粉钻-不可提现 + 粉钻-可提现）
        if (WalletTypeEnum.PINK_DIAMOND_TOTAL == walletTypeEnum) {
            UserWalletVO userWallet = getUserWallet(userId);
            if (userWallet == null) {
                return 0L;
            }
            return userWallet.getWalletItems()
                    .stream()
                    .filter(Objects::nonNull)
                    .filter(item -> item.getWalletType() != null)
                    .filter(item -> WalletTypeEnum.PINK_DIAMOND_TOTAL.getValue() - item.getWalletType() == 0)
                    .map(item -> item.getBalance())
                    .findFirst()
                    .orElse(0L);
        }

        if (isIncome == null) {
            isIncome = BoolIntVal.Ignore;
        }
        return userWalletHisDAO.sumByTypeAndTime(getWalletId(userId, walletType), isIncome, beginTime, endTime);
    }
}
