package com.starbuds.server.service.finance.provider.lottery;

import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.activity.LotteryGamePlayTypeEnum;
import com.starbuds.server.common.constant.activity.LotteryGameRewardTypeEnum;
import com.starbuds.server.common.constant.cache.RedisActivityKeys;
import com.starbuds.server.common.constant.finance.AddItemToBagSourceEnum;
import com.starbuds.server.common.constant.finance.BagItemTypeEnum;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.pojo.mq.LotteryGameStatisticMsg;
import com.starbuds.server.common.pojo.mq.QueueAddItemToBagMsg;
import com.starbuds.server.common.service.core.CacheService;
import com.starbuds.server.common.service.message.biz.LotteryGameImMsgService;
import com.starbuds.server.common.utils.LogUtil;
import com.starbuds.server.service.finance.pojo.bo.lottery.LotteryGameAwardItemBO;
import com.starbuds.server.service.finance.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections.CollectionUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.util.*;
import java.util.stream.Collectors;

import static com.starbuds.server.common.constant.activity.LotteryGamePlayTypeEnum.DIVIDE_CAKE;
import static com.starbuds.server.common.constant.activity.LotteryGamePlayTypeEnum.RED_BAG;
import static com.starbuds.server.common.constant.activity.LotteryGameRewardTypeEnum.GIFT;
import static com.starbuds.server.common.constant.activity.LotteryGameRewardTypeEnum.PROPS;
import static com.starbuds.server.service.finance.util.SpringContextUtil.isDev;

/**
 * 抽奖游戏
 *
 * @author aoe
 * @date 2021/11/29
 */
@Slf4j
@Component
public class LotteryGameProvider {

    @Resource(name = "redBagRainWinResultProvider")
    private LotteryGameWinResult redBagRainWinResultProvider;

    @Resource(name = "divideCakeWinResultProvider")
    private LotteryGameWinResult divideCakeWinResultProvider;

    @Resource(name = "lotteryGameAwardGiftProvider")
    private LotteryGameAward lotteryGameAwardGiftProvider;

    @Resource(name = "lotteryGameAwardPropsProvider")
    private LotteryGameAward lotteryGameAwardPropsProvider;

    @Resource
    private RocketProducer rocketProducer;

    @Reference(version = "1.0.0")
    private LotteryGameImMsgService lotteryGameImMsgService;

    @Reference(version = "1.0.0")
    private CacheService cacheService;

    private Map<LotteryGamePlayTypeEnum, LotteryGameWinResult> winResultProviderMap;

    @PostConstruct
    private void init() {
        winResultProviderMap = new HashMap<>();
        winResultProviderMap.put(RED_BAG, redBagRainWinResultProvider);
        winResultProviderMap.put(DIVIDE_CAKE, divideCakeWinResultProvider);
        log.info("winResultProviderMap={}", winResultProviderMap);
        log.info("lotteryGameAwardGiftProvider={}", lotteryGameAwardGiftProvider);
        log.info("lotteryGameAwardPropsProvider={}", lotteryGameAwardPropsProvider);
    }

    /**
     * 先抽奖后发奖
     *
     * @param snapId
     * @return
     */
    @Deprecated
    @Transactional(rollbackFor = Exception.class)
    public boolean lotteryThenAward(long snapId, LotteryGamePlayTypeEnum playType) {
        if (isDev()) {
            log.info("先抽奖后发奖 snapId={} playType={}", snapId, playType);
        }

        LotteryGameWinResult winResult = winResultProviderMap.get(playType);
        // 抽奖
        List<LotteryGameAwardItemBO> awardItems = winResult.doLottery(snapId);
        if (isDev()) {
            log.info("抽奖结果 {}", LogUtil.obj2JsonObj(awardItems));
        }
        // 发放奖品
        int winGiftCount = lotteryGameAwardGiftProvider.awardItems(getAwardItemsByRewardType(awardItems, GIFT), playType);
        int winPropsCount = lotteryGameAwardPropsProvider.awardItems(getAwardItemsByRewardType(awardItems, PROPS), playType);
        log.info("抽奖游戏中奖信息 winGiftCount={} winPropsCount={}", winGiftCount, winPropsCount);

        cleanCache(snapId);

        // 发送抽奖游戏IM消息-开奖结果
        boolean sendImMsg = lotteryGameImMsgService.sendLotteryResultImMsg(snapId, playType);
        if (!sendImMsg) {
            log.error("发送抽奖游戏IM消息-开奖结果失败 snapId=" + snapId, " playType=" + playType.desc);
        }
        log.info("发送抽奖游戏IM消息-开奖结果成功 snapId={} playType={}", snapId, playType.desc);

        // 发送抽奖游戏统计消息
        return sendStatMsg(snapId, awardItems);
    }

    private void cleanCache(long snapId) {
        String key = RedisActivityKeys.getLotteryGameWinItemsMap(snapId);
        cacheService.delKey(key);
    }

    public Collection<QueueAddItemToBagMsg> toQueueAddItemToBagMsgList(final Collection<LotteryGameAwardItemBO> items,
                                                                       final LotteryGamePlayTypeEnum playType,
                                                                       long expireTime) {
        if (CollectionUtils.isEmpty(items)) {
            return Collections.emptyList();
        }

        return items.stream()
                .map(item -> {
                    LotteryGameRewardTypeEnum rewardTypeEnum = LotteryGameRewardTypeEnum.toEnum(item.getRewardType());
                    if (rewardTypeEnum == null || rewardTypeEnum == LotteryGameRewardTypeEnum.NONE) {
                        log.error("无效的抽奖活动奖励物品类型 item=" + LogUtil.obj2JsonObj(item));
                        return null;
                    }
                    QueueAddItemToBagMsg msg = new QueueAddItemToBagMsg();
                    msg.setOrderId(item.getOrderId());
                    msg.setSource(toAddItemToBagSourceEnum(playType));
                    msg.setItemId(item.getRewardId());
                    msg.setItemName(item.getRewardName());
                    BagItemTypeEnum itemType = BagItemTypeEnum.PROPS;
                    if (GIFT == rewardTypeEnum) {
                        itemType = BagItemTypeEnum.GIFT;
                    }
                    msg.setItemType(itemType);
                    Integer quantity = item.getRewardNum();
                    msg.setQuantity(quantity);
                    msg.setExpireTime(expireTime);
                    msg.setUserId(item.getUserId());
                    msg.setRemark(String.format("%s获得%s个%s", playType.desc, quantity, item.getRewardName()));
                    return msg;
                })
                .filter(Objects::nonNull)
                .collect(Collectors.toList());
    }

    private AddItemToBagSourceEnum toAddItemToBagSourceEnum(final LotteryGamePlayTypeEnum playType) {
        if (LotteryGamePlayTypeEnum.RED_BAG == playType) {
            return AddItemToBagSourceEnum.RED_BAG;
        }
        return AddItemToBagSourceEnum.DIVIDE_CAKE;
    }

    private List<LotteryGameAwardItemBO> getAwardItemsByRewardType(List<LotteryGameAwardItemBO> awardItems,
                                                                   LotteryGameRewardTypeEnum type) {
        if (CollectionUtils.isEmpty(awardItems)) {
            return Collections.emptyList();
        }
        return awardItems.stream()
                .filter(Objects::nonNull)
                .filter(item -> LotteryGameRewardTypeEnum.toEnum(item.getRewardType()) == type)
                .collect(Collectors.toList());
    }

    /**
     * 发送抽奖游戏统计消息
     *
     * @param snapId
     * @param awardItems
     */
    private boolean sendStatMsg(long snapId, List<LotteryGameAwardItemBO> awardItems) {
        if (CollectionUtils.isEmpty(awardItems)) {
            log.info("没有人中奖，不发送抽奖游戏统计消息 snapId={}", snapId);
            return true;
        }
        int rewardNum = (int) awardItems.stream().map(LotteryGameAwardItemBO::getUserId).distinct().count();
        Long giveRewardTotal = awardItems.stream().map(LotteryGameAwardItemBO::getRewardTotalPrice)
                .reduce(Long::sum).orElse(0L);

        LotteryGameStatisticMsg msg = new LotteryGameStatisticMsg();
        msg.setSnapId(snapId);
        msg.setRewardNum(rewardNum);
        msg.setGiveRewardTotal(giveRewardTotal);

        boolean sendLotteryGameEndMsg = rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.LOTTERY_GAME_STATISTIC, msg.serialize());
        if (!sendLotteryGameEndMsg) {
            log.error("发送抽奖游戏统计消息失败 LOTTERY_GAME_STATISTIC snapId=" + snapId + " msg=" + JSON.toJSONString(msg));
            return false;
        }
        log.info("发送抽奖游戏统计消息成功 snapId={} msg={}", snapId, JSON.toJSONString(msg));
        return true;
    }
}
