package com.starbuds.server.service.finance.provider.ratio;

import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.rtc.RtcBusinessTypeEnum;
import com.starbuds.server.common.pojo.rtc.bo.FamilyRatio;
import com.starbuds.server.common.pojo.rtc.bo.RtcRoomRatio;
import com.starbuds.server.service.finance.exception.RtcRatioException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import java.math.BigDecimal;
import java.util.Arrays;

/**
 * 语聊房收益率
 *
 * @author aoe
 * @date 2021/11/4
 */
@Component
@Slf4j
public class RtcRatioProvider {

    @Value("${gift.noFamilyUserRatio}")
    private double giftNoFamilyUserRatio;

    @Value("${gift.familyUserRatio}")
    private double giftFamilyUserRatio;

    @Value("${gift.familyRatio}")
    private double giftFamilyRatio;

    /**
     * 获得房间内用户收益比率
     *
     * @param userFamilyId 用户的家族id
     * @param rtcRoomRatio
     * @return
     */
    public BigDecimal getUserRatio(Long userFamilyId, RtcRoomRatio rtcRoomRatio) {
        BigDecimal noFamilyRatio = BigDecimal.valueOf(giftNoFamilyUserRatio);
        // 语聊房信息不存在
        if (rtcRoomRatio == null) {
            return noFamilyRatio;
        }
        BigDecimal categoryUserRatio = rtcRoomRatio.getCategoryUserRatio();
        if (categoryUserRatio != null) {
            return categoryUserRatio;
        }
        if (userFamilyId == null) {
            return noFamilyRatio;
        }

        FamilyRatio familyRatio = rtcRoomRatio.getFamilyRatio();
        if (familyRatio == null) {
            return noFamilyRatio;
        }
        if (!userFamilyId.equals(familyRatio.getFamilyId())) {
            return noFamilyRatio;
        }

        return rtcRoomRatio.getBusinessType() == RtcBusinessTypeEnum.RTC_NORMAL
                ? familyRatio.getUserRatio() : familyRatio.getUserSingleRtcRatio();
    }

    /**
     * 获得房间内家族收益比率
     *
     * @param rtcRoomRatio
     * @return
     */
    public BigDecimal getFamilyRatio(RtcRoomRatio rtcRoomRatio) {
        // 家族不存在
        if (rtcRoomRatio == null) {
            return BigDecimal.ZERO;
        }
        FamilyRatio familyRatio = rtcRoomRatio.getFamilyRatio();
        if (familyRatio == null) {
            return BigDecimal.ZERO;
        }
        return rtcRoomRatio.getBusinessType() == RtcBusinessTypeEnum.RTC_NORMAL
                ? familyRatio.getFamilyRatio() : familyRatio.getFamilySingleRtcRatio();
    }

    /**
     * 获得家族id
     *
     * @param rtcRoomRatio
     * @return
     */
    public Long getFamilyId(RtcRoomRatio rtcRoomRatio) {
        if (rtcRoomRatio == null) {
            return null;
        }
        FamilyRatio familyRatio = rtcRoomRatio.getFamilyRatio();
        if (familyRatio == null) {
            return null;
        }
        return familyRatio.getFamilyId();
    }

    public void checkRatio(BigDecimal... ratios) throws RtcRatioException {
        BigDecimal total = Arrays.stream(ratios).reduce(BigDecimal::add).get();
        if (total.compareTo(BigDecimal.ONE) < 1) {
            // 收益不能超过 100%
            return;
        }
        throw new RtcRatioException(String.format("语聊房收益比率异常，totalRatio=%s ratios=%s", total, JSON.toJSONString(ratios)));
    }
}
