package com.starbuds.server.service.finance.provider;

import com.google.common.collect.ImmutableList;
import com.starbuds.server.common.constant.finance.AddItemToBagSourceEnum;
import com.starbuds.server.common.constant.finance.BagItemTypeEnum;
import com.starbuds.server.common.pojo.mq.QueueAddItemToBagMsg;
import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.NoArgsConstructor;
import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;
import org.springframework.dao.DuplicateKeyException;
import org.springframework.test.annotation.Rollback;
import org.springframework.transaction.annotation.Transactional;
import test.BaseTest;

import javax.annotation.Resource;
import java.util.*;
import java.util.concurrent.atomic.AtomicLong;
import java.util.function.Function;
import java.util.stream.Collectors;

class BagProviderTest extends BaseTest {

    @Resource
    private BagProvider bagProvider;

    private AtomicLong orderId = new AtomicLong(System.currentTimeMillis());

    @Test
    @Transactional
    @Rollback
    void addQuantityByQueueAddItemToBagMsg() {
        Long userId = 141075591200768L;
        BagItemTypeEnum itemType = BagItemTypeEnum.GIFT;
        Long itemId = 35951954886761L;
        Integer quantity = 1;
        Long expireTime = -1L;
        int quantityByBag = bagProvider.getQuantityByItemId(userId, itemType, itemId);
        AddItemToBagParam param = AddItemToBagParam.of(orderId.incrementAndGet(), itemId, itemType, quantity, expireTime, userId);
        QueueAddItemToBagMsg msg = toRedBagQueueAddItemToBagMsg(param);
        boolean addQuantity = bagProvider.addQuantity(msg);
        Assertions.assertTrue(addQuantity);
        Assertions.assertEquals(quantityByBag + quantity, bagProvider.getQuantityByItemId(userId, itemType, itemId));
    }

    @Test
    @Transactional
    @Rollback
    void addItemsToBagThenSendImMsgStatQueueMsg() {
        List<QueueAddItemToBagMsg> msgList = toMsgList(initParams(), (params) -> toRedBagQueueAddItemToBagMsg(params));
        addItemsToBagThenSendImMsgStatQueueMsg(msgList);
    }

    @Test
    @Transactional
    @Rollback
    void addItemsToBagThenSendImMsgStatQueueMsgSameOrderIdSuccess() {
        final Long giftOrderId = this.orderId.incrementAndGet();
        final Long propsOrderId = this.orderId.incrementAndGet();
        List<AddItemToBagParam> redBagParams = ImmutableList.of(
                AddItemToBagParam.of(giftOrderId, 35951954886761L, BagItemTypeEnum.GIFT, 1, -1L, 141075591200768L),
                AddItemToBagParam.of(propsOrderId, 69313465679872L, BagItemTypeEnum.PROPS, 1, -1L, 141075591200768L)
        );
        List<QueueAddItemToBagMsg> redBagMsgList = toMsgList(redBagParams, (params) -> toRedBagQueueAddItemToBagMsg(params));

        List<AddItemToBagParam> divideCakeParams = ImmutableList.of(
                AddItemToBagParam.of(giftOrderId, 35951954886761L, BagItemTypeEnum.GIFT, 1, -1L, 141075591200768L),
                AddItemToBagParam.of(propsOrderId, 69313465679872L, BagItemTypeEnum.PROPS, 1, -1L, 141075591200768L)
        );
        List<QueueAddItemToBagMsg> divideCakeMsgList = toMsgList(divideCakeParams, (params) -> toDivideCakeQueueAddItemToBagMsg(params));

        List<QueueAddItemToBagMsg> msgList = new ArrayList<>();
        msgList.addAll(redBagMsgList);
        msgList.addAll(divideCakeMsgList);
        addItemsToBagThenSendImMsgStatQueueMsg(msgList);
    }

    @Test
    @Transactional
    @Rollback
    void addItemsToBagThenSendImMsgStatQueueMsgSameOrderIdFail() {
        final Long orderId = this.orderId.incrementAndGet();
        List<AddItemToBagParam> redBagParams = ImmutableList.of(
                AddItemToBagParam.of(orderId, 35951954886761L, BagItemTypeEnum.GIFT, 1, -1L, 141075591200768L),
                AddItemToBagParam.of(orderId, 69313465679872L, BagItemTypeEnum.PROPS, 1, -1L, 141075591200768L)
        );
        List<QueueAddItemToBagMsg> msgList = toMsgList(redBagParams, (params) -> toRedBagQueueAddItemToBagMsg(params));
        // 发送前背包物品数量
        Map<String, Integer> quantityMapBeforeSendMap = quantityByBagMap(msgList);

        Assertions.assertThrows(DuplicateKeyException.class, () -> {
            addItemsToBagThenSendImMsgStatQueueMsg(msgList);
            // 发送后背包物品数量
            Map<String, Integer> quantityMapAfterSendMap = quantityByBagMap(msgList);
            quantityKeys(msgList).stream()
                    .forEach(key -> {
                        Integer before = quantityMapBeforeSendMap.get(key);
                        Integer after = quantityMapAfterSendMap.get(key);
                        Assertions.assertEquals(before, after);
                    });
        });
    }

    /**
     * 背包中物品的数量
     *
     * @param msgList
     * @return
     */
    private Map<String, Integer> quantityByBagMap(List<QueueAddItemToBagMsg> msgList) {
        Map<String, Integer> map = new HashMap<>();
        quantityKeys(msgList).stream()
                .forEach(key -> {
                    String[] params = StringUtils.split(key, "_");
                    Long userId = Long.parseLong(params[0]);
                    Long itemId = Long.parseLong(params[1]);
                    BagItemTypeEnum itemType = BagItemTypeEnum.toEnum(Integer.parseInt(params[2]));
                    int quantityByBag = bagProvider.getQuantityByItemId(userId, itemType, itemId);
                    map.put(key, quantityByBag);
                });
        return map;
    }

    /**
     * 总共发送物品数量Map
     *
     * @param msgList
     * @return
     */
    private Map<String, Integer> quantityTotalSendMap(List<QueueAddItemToBagMsg> msgList) {
        Map<String, Integer> map = new HashMap<>();
        msgList.stream()
                .forEach(msg -> {
                    String key = quantityKey(msg.getUserId(), msg.getItemId(), msg.getItemType());
                    Integer total = Optional.ofNullable(map.get(key)).orElse(0);
                    total += msg.getQuantity();
                    map.put(key, total);
                });
        return map;
    }

    private List<String> quantityKeys(List<QueueAddItemToBagMsg> msgList) {
        return msgList.stream()
                .map(msg -> quantityKey(msg.getUserId(), msg.getItemId(), msg.getItemType()))
                .distinct()
                .collect(Collectors.toList());
    }

    private String quantityKey(Long userId, Long itemId, BagItemTypeEnum itemType) {
        return String.format("%s_%s_%s", userId, itemId, itemType.getValue());
    }

    private List<QueueAddItemToBagMsg> toMsgList(List<AddItemToBagParam> params,
                                                 Function<AddItemToBagParam, QueueAddItemToBagMsg> function) {
        return params.stream()
                .map(param -> function.apply(param))
                .collect(Collectors.toList());
    }

    private List<AddItemToBagParam> initParams() {
        Long expireTime = System.currentTimeMillis() + 36000;
        return ImmutableList.of(
                AddItemToBagParam.of(orderId.incrementAndGet(), 35951954886761L, BagItemTypeEnum.GIFT, 1, -1L, 141075591200768L),
                AddItemToBagParam.of(orderId.incrementAndGet(), 35951954886761L, BagItemTypeEnum.GIFT, 99, expireTime, 141075591200768L),
                AddItemToBagParam.of(orderId.incrementAndGet(), 69313465679872L, BagItemTypeEnum.PROPS, 1, -1L, 141075591200768L),
                AddItemToBagParam.of(orderId.incrementAndGet(), 69313465679872L, BagItemTypeEnum.PROPS, 99, expireTime, 141075591200768L)
        );
    }

    void addItemsToBagThenSendImMsgStatQueueMsg(List<QueueAddItemToBagMsg> msgList) {
        // 发送前背包物品数量
        Map<String, Integer> quantityMapBeforeSendMap = quantityByBagMap(msgList);
        // 总共发送物品数量Map
        Map<String, Integer> quantityTotalSendMap = quantityTotalSendMap(msgList);

        bagProvider.addItemsToBagThenSendImMsgStatQueueMsg(msgList);

        // 发送后背包物品数量
        Map<String, Integer> quantityMapAfterSendMap = quantityByBagMap(msgList);
        quantityKeys(msgList).stream()
                .forEach(key -> {
                    Integer before = quantityMapBeforeSendMap.get(key);
                    Integer totalSend = quantityTotalSendMap.get(key);
                    Integer after = quantityMapAfterSendMap.get(key);
                    Assertions.assertEquals(before + totalSend, after);
                });
    }

    @Builder
    @NoArgsConstructor
    @AllArgsConstructor(staticName = "of")
    static class AddItemToBagParam {
        Long orderId;
        Long itemId;
        BagItemTypeEnum itemType;
        Integer quantity;
        Long expireTime;
        Long userId;
    }

    private QueueAddItemToBagMsg toRedBagQueueAddItemToBagMsg(AddItemToBagParam param) {
        QueueAddItemToBagMsg msg = new QueueAddItemToBagMsg();
        msg.setOrderId(param.orderId);
        msg.setSource(AddItemToBagSourceEnum.RED_BAG);
        msg.setItemId(param.itemId);
        msg.setItemType(param.itemType);
        msg.setItemName("测试礼物");
        msg.setQuantity(param.quantity);
        msg.setExpireTime(param.expireTime);
        msg.setUserId(param.userId);
        msg.setRemark("单元测试添加背包礼物" + param.itemId);
        return msg;
    }

    private QueueAddItemToBagMsg toDivideCakeQueueAddItemToBagMsg(AddItemToBagParam param) {
        QueueAddItemToBagMsg msg = new QueueAddItemToBagMsg();
        msg.setOrderId(param.orderId);
        msg.setSource(AddItemToBagSourceEnum.DIVIDE_CAKE);
        msg.setItemId(param.itemId);
        msg.setItemType(param.itemType);
        msg.setItemName("测试礼物");
        msg.setQuantity(param.quantity);
        msg.setExpireTime(param.expireTime);
        msg.setUserId(param.userId);
        msg.setRemark("单元测试添加背包礼物" + param.itemId);
        return msg;
    }
}